﻿#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals 		= 3
#pragma IgorVersion		= 8.0
#pragma ModuleName		= SpectraGraphTools
#pragma version			= 2.25
#pragma DefaultTab		= {3,20,4}		// set default tab width in Igor Pro 9 and later

static Constant kVersion = 2.25
static StrConstant kVersionDate = "02/2025"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21562																// the project node on IgorExchange
static StrConstant ksShortTitle = "Spectra Graph Tools"											// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//	Original code & idea for the right-click graph menus by Tony Withers - https://www.wavemetrics.com/user/tony
//
//	This is a collection of functions for modifying traces and graphs.
//	Contents:
//		- Basic Graph Style and Label Macros
//		- Quick Normalization :		Normalize to unit area or height / scale the traces in a graph
//		- SpectraWaterfall :		Generates a waterfall plot out of 2D data or a folder
//		- BuildTraceOffsetPanel :	A small panel to offset and color traces
//		- BuildGraphSizePanel :		A small panel to set the graph size and margins
//	
//	keyboard shortcuts:
//
//	Styles:
//	• hold ctrl while applying ‘BoxStyle’				left axis will have no ticks (blank axis)
//	• hold ctrl, shift or alt while using ‘Colorize’	apply color in reverse order
//
//	Normalize & Align:
//	• hold shift while selecting ‘Normalize’ 			y-offset = 0
//	• hold alt while selecting ‘Normalize’ 				y-offset stays the same
//	• hold ctrl while selecting ‘Align Traces Here’		align to y = 0
//	• hold ctrl while selecting ‘y-Order Trace’			sort trace in descending order
//
//	Trace Offset Panel:
//	• ← → arrow keys									shift data left / right
//	• ↑↓ arrow keys										shift data up / down
//	• ctrl + ← → arrow keys								scale data in x up / down
//	• ctrl + ↑↓ arrow keys								scale data in y up / down
//	• shift + above arrow controls						10x larger changes per step
//	• alt + arrow keys									moves cursor instead of traces
//	• page up, page down keys							set cursor to next / previous trace in the list
//	• h / l (small L) keys								set cursor to highest / lowest intensity of current trace
//	• a, b, c, d keys									select this cursor (if available on the graph)
//	• hold ctrl while scaling offsets per trace			absolute mode: also scales the offset of the first trace
//
//	Graph Size Panel:
//	• hold shift while applying the ‘font’				setting	set the axis font instead of the graph font
//	• hold shift while pressing Copy Style				copies style to the clipboard
//
//
// 2021-03-01 - ver. 1.00:	Initial public release.
// 2021-03-07 - ver. 1.10:	Quick normalization did not work properly when the data included NaNs.
//							Graph Size panel: Axes margins did not copy correctly from one graph to the other.
//							The Graph Size panel threw an error on Mac because of a wrong ModifyControl command.
//							Trace Offset panel: Added keyboard shortcuts for selecting and moving the cursor.
//							Trace Offset panel: It is now possible to move and scale selected traces with the cursor keys.
// 2021-03-11 - ver. 1.11:	A cursor set on an image triggered an error message for the TraceOffset panel.
// 2021-03-16 - ver. 1.20:	Added global constant for the multiplier step and made keyboard control more precise.
//							Added Offset Traces functionality to the graph and traces right-click menus.
// 2021-03-17 - ver. 1.21:	Now it is possible to choose whether offsets are set or added.
//							Fixed bug: Closing a graph did not properly update active window.
//							Now the total shift is also displayed for single traces.
// 2021-03-28 - ver. 1.30:	Graph Size panel: The last selected size unit setting is now saved in the current experiment.
//							Trace offset panel: Fixed a few error messages triggered by very old versions of the panel.
//							Trace offset panel: Trace scaling increment steps now get smaller and bigger depending on the
//							order of magnitude of the current scaling (helps to work with very big or small scaling factors).
//							Trace offset panel: Added a button to quickly set a new cursor onto the graph.
//							The currently selected trace name is now displayed in the panel title.
// 2021-05-14 - ver. 1.40:	Slightly adjusted the axis-label graph macros.
//							Graph size panel: Now switching the axes keeps the axis label visible.
//							Offset menu: Works now for 1-column 2D waves as well.
//							Graph Size panel: Hold shift while changing the font size to change the global axes font instead.
//							Graph Size panel: Added print checkbox to enable print-to-history function for most controls.
// 2021-08-14 - ver. 1.50:	Fixed bug: Commands got printed anyway when an old panel without the print checkbox was used.
//							Graph Size panel: Overhaul of the panel controls.
//							Graph Size panel: Added tick label margin controls.
//							Trace offset panel: Moved controls into tabs for better organization.
//							Trace offset panel: The Prev and Next cursor placement controls honor trace offsets now.
//							Trace offset panel: New cursors get set in the center of the current bottom-axis range.
// 2021-12-01 - ver. 1.60:	Added support for sub-graphs (also in layouts and panels) in all tools.
//							Trace offset panel: Fixed bug - cursors on shared image plots were not caught properly and led to errors.
//							Graph Size panel: Doubled the increment for size changes for faster adjustments.
//							Graph Size panel: Added controls for label margin.
//							Graph Size panel: Tick and label margins will be copied & pasted between graphs as well.
// 2022-03-23 - ver. 1.65:	Normalize to cursor now honors x-offsets of traces.
//							Added support for keyboard-shortcut (and other future) settings via external text file.
//							Added quick-scale tool to modify data directly from a graph.
//							Trace tool now shows color of currently selected trace.
// 2022-04-05 - ver. 1.70:	New right-click shortcut to add legend with live trace offset information.
//							Moved all graph related menu entries into Offset Spectra sub-menu.
//							Fixed bug: Quick-Scale panel did not react when trace together with cursor was removed.
//							Made the Quick-Scale panel x-wave aware.
// 2022-07-15 - ver. 1.80:	Quick-offset legend now ignores traces from ksTraceIgnoreList.
//							New keyboard shortcut for ScaleAllTraces() => hold ctrl to keep current offsets.
//							Added alphanumerical trace ordering.
//							Added new Normalize functions to graph's right-click menus and reorganized menu entries.
//							Live-offset legend: Support for trace instances and fixed some bugs.
//							Fixed Bug: Reverse sorting was not re-applying offsets correctly.
// 2022-10-19 - ver. 1.85:	Optimized code for ScaleAllTraces and added smoothing option to the right-click menu.
//							Added mean around cursor to ScaleAllTraces.
//							Added mouse cursor as a target for quick normalization.
//							Added normalization to the right-click menu of the graph marquee.
//							Added smoothing option for normalization.
//							Modifying traces now works for vertical graphs and non-standard axes.
// 2022-10-27 - ver. 1.90:	Trace offset panel: Filter now accepts lists and got a right-click menu to add traces to the list.
//							Trace offset panel: Fixed error when there is only one trace on the graph.
//							Graph Size panel: Added 'copy style to / paste style from clipboard' functionality.
//							Quick normalization can now be done to an arbitrary y value.
//							Fixed bug: Normalization to small values did not work properly.
//							All actions now work for vertical graphs and non-standard axes.
// 2023-07-25 - ver. 2.00:	Fixed bug: Quick-scaling failed sometimes when data had NaNs.
//							Fixed bug: Quick-scaling was not working for the zero-offset setting.
//							The Y Reset and X Reset buttons honor filter settings now.
//							Upgraded code to Igor 8.
//							Alphanumerical sorting now works for plots of 2D data.
//							Improved XY-plot support for scaling/offsetting and the Trace Tool.
//							Fixed various small inconsistencies and edge cases.
//							Fixed bug: Copy Style did not work properly with swapped axes and a fixed aspect ratio.
//							Graph Size and Trace Offset panels are now displayed next to the top graph and otherwise centered on the screen.
//							Overhauled the Trace Normalize menu to be less confusing.
//							Added 'Export Offsets' option to normalize menu.
// 2024-06-19 - ver. 2.05:	Added 'Plot as ...' menu entries to the Data Browser menu (Igor 9).
//							Quick Trace Offset and Quick Legend calls are now printed to the history.
//							QuickOffsetLegend() anchor codes are now strings instead of numbers.
//							Added delete-from-cursor function to the Quick-Scale panel.
//							Fixed bug: Axis shift was not working correctly for images.
//							Redesigned the Quick-Scale panel for improved usability.
//							Fixed bug with the Quick-Scale panel: Photon energies were always overwritten by the automatic fetch function.
//							Fixed sudden magnitude jump when scaling traces with the Trace Tool.
// 2024-12-03 - ver. 2.20:	Fixed bug with live offsets: Providing the wrong graph name to GT_liveOff() gave an error.
//							Trace Offset: Improved granularity for individual trace movement with the keyboard.
//							Trace Offset: Keyboard movement of individual traces can now be accelerated with the shift key.
//							Trace Offset: Scale all trace offsets - holding the ctrl /cmd key now scales the offsets including the baseline.
//							Trace Offset: Added keyboard movements on a per-tab basis (multiple or single traces).
//							Trace Offset: Now traces can also be modified with the graph in focus.
//							Fixed error when AddTraceProfile is missing while the 'Profileleft' axis is still available.
//							Graph Size: Removed custom graph macros and margin changes from tick label toggle.
//							Graph Size: Improved aspect-ratio code and added custom aspect ratio input.
//							Graph Size: Added support for modifying custom axes.
//							Graph Size: Axis label position is now aware of positioning mode.
//							Graph Size: Increased the overall size of the Panel.
//							Graph Size: Now it is possible to set a width aspect ratio.
//							Graph Size: Axis range can now be omitted when applying a graph style.
//							Optimized color selection for ColorAllTraces().
//							QuickOffsetLegend() now supports setting of frame and background via optional parameters.
//							The legend created by QuickOffsetLegend() is now generic (unnamed).
//							Graph Size: Added feature for changing and adding graph legends to the top graph.
//							Graph Size: More settings are now saved in the new V_GTStyleOptions variable.
//							Graph Size: Axis label and tick settings can now be toggled to different parameters.
//							Added the function QuickLegendStyle() to change the style and layout of existing legends.
//							Changed GT_liveOff() into GT_liveInfo() which allows to define the number of significant digits for each entry.
//							Streamlined the code for QuickOffsetLegend(), since live information can now be added via QuickLegendStyle().
//							Graph Size: Merged the size and style sub-sections into one.
//							Graph Size: Fixed bug with reading size values when a sub-graph was present.
//							Graph Size: Improved sizes for preset formats and made them axis-aware.
//							Graph Size: Added plot and window color selectors.
// 2025-02-18 - ver. 2.25:	Annotations for images now contain the image (wave) name and can also be switched to a TextBox (displaying only the name).
//							Fixed bug: QuickLegendStyle() had issues with instanced traces.
//							Removed legacy hook functions for old panels.
//							Fixed Trace Offset bug: Page up / down keys were not caught when end of the trace list was reached.
//							Fixed Graph Size bug: Error when using the graph size controls with the custom axes tab open.
//							Fixed QuickLegendStyle(): Changes were not properly applied to multi-column legends with space / tab characters.
//							Fixed bug: Options for ScaleAllTraces() were not honored in the main menu.
//							Optimized the code for ScaleAllTraces() to be more readable.
//							Fixed bug: The search range for ScaleAllTraces() was not adjusted for the x-axis scaling.
//							Fixed bug: ScaleAllTraces() threw an error if designated position was out-of-bounds of the x wave.
//							ColorAllTraces() backs up trace colors to the clipboard, which are reset for the same graph via a second call of the function.
//________________________________________________________________________________________________

static Constant kOffsetIncrStep = 0.005															// fraction of the total range to increase / decrease values each step
static Constant kMultipIncrStep = 0.1															// by how much the multiplier is incremented (keyboard presses are 1/5 of this value)
static Constant kScale_defRange = 0.03															// percentage used for scaling: default is +-3 % of the total wave's x range
static Constant QuickMenu_Print = 1																// prints quick-menu commands

static StrConstant ksTraceIgnoreList	= "w_base;MQP_*;"										// list of traces to ignore, wildcards okay
static StrConstant ksGraphSizeSetting	= "V_GTStyleOptions"									// global variable name which saves the Graph Size Panel settings
static StrConstant ksMenuOptionsSave	= "V_GTMenuOptions"										// global variable name for right-click menu options
static StrConstant ksMenuOptionsDefault	= "offsetMode:0;cursor:Mouse;cursorRange:3;normSmooth:0;waveExport:0;"	// default settings for right-click menus
static StrConstant kUsedCursorList		= "A;B;C;D;E;F;G;H;I;J;"
static StrConstant kSettingsFileName	= "Spectra Tools settings.dat"							// just for keyboard shortcuts for now.

//-------------------------------- functions for user code ---------------------------------------
static function TraceTool_FetchXShift(wave input, string &unit)									// extract predetermined X offset shift values from data
	variable xShift = NaN																		// the shift value which will be preset in the panel
	// ++++++++++++++++++++++ get the photon energy as shift value +++++++++++++++++++++++++++++++
	variable neg = 1
	string notes = ReplaceString(" =",ReplaceString("= ",note(input),"="),"=")
	string energy = StringByKey("Xray energy", notes, "=","\r")									// find the photon energy
	if (!strlen(energy))
		energy = StringByKey("photon energy", notes, "=","\r")									// try a different key
	endif
	if (!strlen(energy))
		energy = StringByKey("Excitation Energy", notes, "=","\r")								// try yet a different key
		neg = -1																				// this data requires flipping
	endif
	unit = "eV"
	xShift = str2num(energy)*neg
	if (xShift == 0)
		xShift = NaN
	endif
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	return xShift 																				// return NaN for no change and return negative values for flipping the trace
end

static function TraceTool_FetchYShift(wave input, string &unit)									// extract predetermined Y offset shift values
	return NaN	 																				// return NaN for no change and return negative values for flipping the trace
end
//------------------------------------------------------------------------------------------------

//################################################################################################

Menu "Spectra Tools"
	"-"
	SubMenu "Normalize Traces"
		SpectraGraphTools#MenuEntry("norm_to1"),			/Q, SpectraGraphTools#ScaleMainMenuWrapper("One")		// hold shift for zero offset; hold alt to skip offset alignment
		SpectraGraphTools#MenuEntry("norm_toCur"),			/Q, SpectraGraphTools#ScaleMainMenuWrapper("Cursor")
		SpectraGraphTools#MenuEntry("norm_toArea"),			/Q, SpectraGraphTools#ScaleMainMenuWrapper("Area")
		"-"
		"Smoothing Factor ...",								/Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	"-"
	SpectraGraphTools#MenuEntry("waterfall"),	/Q, CreateBrowser; SpectraWaterfall()
	SpectraGraphTools#MenuEntry("quick_color"),	/Q, ColorAllTraces()
	SpectraGraphTools#MenuEntry("graph_GUI"),	/Q, BuildGraphSizePanel()
	SpectraGraphTools#MenuEntry("trace_GUI"),	/Q, BuildTraceOffsetPanel()
	SpectraGraphTools#MenuEntry("quick_scale"),	/Q, AttachQuickScalePanel()
End

Menu "GraphPopup"
	"-"
	SubMenu "Traces: Normalize"
		SubMenu "to One"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleClickWrapper("One")
		End
		SubMenu "to Global Maximum"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleClickWrapper("Global")
		End
			SpectraGraphTools#DisplayNormMenu(atCursor=1),	/Q, SpectraGraphTools#ScaleClickWrapper("Cursor")
		"-"
		SubMenu "Select Cursor"
			SpectraGraphTools#DisplayCursorOptions(),		/Q, SpectraGraphTools#SetCursorOption()
		End
		"Smoothing Factor ...",								/Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	Submenu "Traces: Offset and Sort"
		"y-Order Traces Alphanumerically",	/Q, SpectraGraphTools#OrderTracesAlphabetical()		// holding ctrl / command sorts in descending order
		"-"
		"Spread Traces Across y Range",		/Q, SpectraGraphTools#SpreadToFillRange()
		"Spread Traces: Equal Offset",		/Q, SpectraGraphTools#SpreadSpectra(0)
		"Spread Traces: Equal Gap",			/Q, SpectraGraphTools#SpreadSpectra(1)
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	"-"
	"Add Legend with Live Offset ...",	/Q, SpectraGraphTools#QuickOffsetLegendWrapper()
	"Left Axis: Autoscale Only Visible", 	/Q, SetAxis/A=2/E=0 left
	"Left Axis: Autoscale From Zero",	 	/Q, SetAxis/A=2/E=1 left
End

Menu "TracePopup"//, dynamic
	"-"
	SubMenu "Traces: Normalize"
		SubMenu "to One"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleClickWrapper("One")
		End
		SubMenu "to Global Maximum"
			SpectraGraphTools#DisplayNormMenu(), 			/Q, SpectraGraphTools#ScaleClickWrapper("Global")
		End
			SpectraGraphTools#DisplayNormMenu(atCursor=1),	/Q, SpectraGraphTools#ScaleClickWrapper("Cursor")
		"-"
		SubMenu "Select Cursor"
			SpectraGraphTools#DisplayCursorOptions(),		/Q, SpectraGraphTools#SetCursorOption()
		End
		"Smoothing Factor ...",								/Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	Submenu "Traces: Offset and Sort"
		SpectraGraphTools#MenuWithinPlotrange("y-Offset: Align Traces Here"),	/Q, SpectraGraphTools#AlignTracesAtXPos(0)		// holding ctrl / command aligns to zero
		SpectraGraphTools#MenuWithinPlotrange("y-Scale: Match Traces Here"),	/Q, SpectraGraphTools#AlignTracesAtXPos(1)		// holding ctrl / command scales to one
		SpectraGraphTools#MenuWithinPlotrange("y-Order Traces Here"),			/Q, SpectraGraphTools#OrderTracesAtXPos()		// holding ctrl / command sorts in descending order
		"y-Order Traces Alphanumerically",										/Q, SpectraGraphTools#OrderTracesAlphabetical()	// holding ctrl / command sorts in descending order
		"-"
		"Spread Traces Across y Range",		/Q, SpectraGraphTools#SpreadToFillRange()
		"Spread Traces: Equal Offset", 		/Q, SpectraGraphTools#SpreadSpectra(0)
		"Spread Traces: Equal Gap", 		/Q, SpectraGraphTools#SpreadSpectra(1)
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
	"-"
	"Add Legend with Live Offset ...",		/Q, SpectraGraphTools#QuickOffsetLegendWrapper()
	"Left Axis: Autoscale Only Visible",	/Q, SetAxis/A=2/E=0 left
	"Left Axis: Autoscale From Zero",		/Q, SetAxis/A=2/E=1 left
End

Menu "GraphMarquee", dynamic
	SubMenu SpectraGraphTools#TraceMarqueeMenu("Traces: Normalize")
		SpectraGraphTools#DisplayMarqueeMenu(),				/Q, SpectraGraphTools#ScaleMarqueeWrapper()
		"-"
		"Smoothing Factor ...",								/Q, SpectraGraphTools#SetSmoothOption()
		SpectraGraphTools#DisplayMenuOptions("zeroOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("zeroOffset")
		SpectraGraphTools#DisplayMenuOptions("keepOffset"),	/Q, SpectraGraphTools#ToggleMenuOptions("keepOffset")
		SpectraGraphTools#DisplayMenuOptions("waveExport"),	/Q, SpectraGraphTools#ToggleMenuOptions("waveExport")
		"-"
		"Undo All Scaling",	/Q, ModifyGraph muloffset={0,0}
		"Undo All Offsets",	/Q, ModifyGraph offset={0,0}
	End
End

//################################################################################################

static function/S MenuEntry(string which)
	string menuStr = "", searchStr = ""
	StrSwitch (which)
		case "norm_to1":
			menuStr = "to One"
			searchStr = "normalize traces to one"
		break
		case "norm_toCur":
			menuStr = "to One at Cursor A"
			searchStr = "normalize traces to cursor"
		break
		case "norm_toArea":
			menuStr = "to Area"
			searchStr = "normalize traces to area"
		break
		case "waterfall":
			menuStr = "Waterfall Plot of 2D Data or Folder"
			searchStr = "create waterfall plot"
		break
		case "quick_color":
			menuStr = "Quick Colorize Traces"
			searchStr = "quick colorize traces"
		break
		case "graph_GUI":
			menuStr = "Graph Size and Style Panel ..."
			searchStr = "start graph size panel"
		break
		case "trace_GUI":
			menuStr = "Trace Offset Panel ..."
			searchStr = "start trace offset panel"
		break
		case "quick_scale":
			menuStr = "Attach Quick-Scale Panel ..."
			searchStr = "start quick-scale panel"
		break
		default:
			return ""
	EndSwitch
	
	int fileID; string read = ""
	Open/Z/R fileID as ParseFilePath(1, FunctionPath(""), ":", 1, 0)+kSettingsFileName
	if (!V_flag)
		do
			FReadLine fileID, read
		while (strlen(read) > 0 && !StringMatch(read, searchStr+"*"))
		read = cleanStr(read,"\t| |")
		read = StringByKey(cleanStr(searchStr,"\t| |"),read,"=","\r")
		Close fileID
	endif
	return menuStr+SelectString(strlen(read),"","/"+read)
end

//------------------------------------------------------------------------------------------------

static function/S TraceMarqueeMenu(string s_menu)
	return SelectString(ItemsInList(TraceNameList("",";",1)),"(","")+s_menu
end

static function ToggleMenuOptions(string which)
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	StrSwitch(which)
		case "waveExport":
			settings = ReplaceNumberByKey("waveExport" ,settings ,(NumberByKey("waveExport"	,settings) == 1 ? 0 : 1))
		break
		default:
			int mode = WhichListItem(which, "zeroOffset;keepOffset;")+1
			if (mode >= 1)
				settings = ReplaceNumberByKey("offsetMode" ,settings ,(NumberByKey("offsetMode"	,settings) == mode ? 0 : mode))
			endif
		break
	EndSwitch
	string/G root:$ksMenuOptionsSave = settings
	BuildMenu "GraphPopup"
	BuildMenu "TracePopup"
	return 0
end

static function/S DisplayCursorOptions()
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	string csr = StringByKey("cursor" ,settings); int mouse = CmpStr(csr,"Mouse")==0
	string cursorList = ReplaceString(csr,kUsedCursorList,"!"+num2char(18)+csr)
	return SelectString(mouse,"","!"+num2char(18))+"Mouse;"+cursorList
end

static function/S DisplayMenuOptions(string which)
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	StrSwitch(which)
		case "zeroOffset":
			return SelectString(NumberByKey("offsetMode" ,settings) == 1,"","!"+num2char(18)) + "Keep y-Offset at Zero"
		break
		case "keepOffset":
			return SelectString(NumberByKey("offsetMode" ,settings) == 2,"","!"+num2char(18)) + "Keep Current y-Offset"
		break
		case "waveExport":
			return SelectString(NumberByKey("waveExport" ,settings) == 1,"","!"+num2char(18)) + "Export Offsets to Wave"
		break
	EndSwitch
end

static function SetCursorOption()
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	GetLastUserMenuInfo
	string/G root:$ksMenuOptionsSave = ReplaceStringByKey("cursor",settings,S_Value)
	BuildMenu "GraphPopup"
	BuildMenu "TracePopup"
	return 0
end

static function SetSmoothOption()
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	variable smth = NumberByKey("normSmooth",settings)
	Prompt smth, "Set Smoothing Factor for Normalization (0 = off):"
	DoPrompt "Smoothing for Trace Normalization", smth
	if (!V_flag)
		string/G root:$ksMenuOptionsSave = ReplaceNumberByKey("normSmooth",settings,smth)
	endif
	return 0
end

static function/S DisplayNormMenu([int atCursor])
	string settings = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)
	string csr = StringByKey("cursor" ,settings); int mouse = CmpStr(csr,"Mouse")==0
	csr = SelectString(mouse,"Cursor ","") + StringByKey("cursor" ,settings) + SelectString(mouse,""," Cursor")
	if (atCursor)
		return "to "+csr+" Position"
	else
		return "Maxima;Area;at "+csr+" x Position;Maxima around "+csr+";Average around "+csr+";"
	endif
end

//################################################################################################

#if (IgorVersion() >= 9.0)
Menu "DataBrowserObjectsPopup", dynamic
	"-"
	SpectraGraphTools#DB_DisplayMenu("Plot as Waterfall",	0),	/Q, SpectraGraphTools#DB_PlotData(0)
	SpectraGraphTools#DB_DisplayMenu("Plot as XY Pair",		1),	/Q, SpectraGraphTools#DB_PlotData(1)
	SpectraGraphTools#DB_DisplayMenu("Plot as Image",		2),	/Q, SpectraGraphTools#DB_PlotData(2)
	"-"
End

static function/S DB_DisplayMenu(string menuStr, int mode)
	string list = DB_selectionList(mode) + SelectString(mode==0,"",DB_selectionList(2))			// add 2D waves for waterfall mode
	return SelectString(ItemsInList(list)>(mode==1),"",menuStr)									// XY pair needs two waves
end

static function DB_PlotData(int mode)
	string list = DB_selectionList(mode); int i
	Switch (mode)
		case 0:	// waterfall
			if (!strlen(list))
				SpectraWaterfall(path=StringFromList(0,DB_selectionList(2)))
			else
				string folderPath = StringFromList(0,list)
				string waveInPath = PossiblyQuoteName(StringFromList(0,WaveList("*", ";", "DIMS:1", $folderPath)))
				SpectraWaterfall(path=folderPath+waveInPath)
			endif
		break
		case 1:	// XY pair
			if (ItemsInList(list)<2)
				return 0
			endif
			string xData = StringFromList(1,list)
			Prompt xData,"Which one is the x wave?",popup,list
			DoPrompt "Select the x Data",xData
			if (V_flag)
				break
			endif
			list = RemoveFromList(xData,list)
			Display/K=1
			for (i=0; i<ItemsInList(list); i++)
				AppendToGraph/Q $StringFromList(i,list) vs $xData
			endfor
			Execute/Q/Z "BoxStyle()"
		break
		case 2:	// image
			for (i=0; i<ItemsInList(list); i++)
				Display/K=1; AppendImage $StringFromList(i,list)
				Execute/Q/Z "BoxStyle()"
				ColorSpectrumFromTable(0, "Terrain")
			endfor
		break
	EndSwitch
	return 0
end

static function/S DB_selectionList(int mode)													// mode: [0] DF; [1-3] Waves with dim 1-3
	if (!strlen(GetBrowserSelection(-1)))														// Data Browser is not open
		return ""
	endif
	string list = ""; int i = 0
	do
		string entry = GetBrowserSelection(i++)
		int isXY = (Exists(entry) == 1 && mode == 1 && WaveDims($entry) == 1)
		int is2D = (Exists(entry) == 1 && mode == 2	&& WaveDims($entry) > 1)
		int isDF = (Exists(entry) != 1 && mode == 0	&& DataFolderExists(entry) && strlen(entry))
		if (isXY || is2D || isDF)
			list += entry + ";"
		endif
	while(strlen(entry) > 0 && i < 100)
	return Sortlist(list, ";", 16)
end
#endif

//________________________________________________________________________________________________
//
//	Default Graph Macro
//________________________________________________________________________________________________

Proc BoxStyle() : GraphStyle
	PauseUpdate; Silent 1
	GetAxis/Q right
	if (V_flag == 1)
		ModifyGraph/Z mirror(left)=2
	endif
	GetAxis/Q top
	if (V_flag == 1)
		ModifyGraph/Z mirror(bottom)=2
	endif
	ModifyGraph noLabel=0
	ModifyGraph standoff=0
	ModifyGraph ZisZ=1
	ModifyGraph tick=0
	ModifyGraph notation(left)=1
	ModifyGraph axisOnTop=1
	ModifyGraph minor=1
	if ((GetKeyState(0) & 2^0) != 0)															// ctrl / cmd = paper style (no left axis notation)
		ModifyGraph tick(left)=3
		ModifyGraph noLabel(left)=1
	endif
	ResumeUpdate
EndMacro

//________________________________________________________________________________________________
//
//	Change the color of up to 13 traces with preset colors
//________________________________________________________________________________________________

function ColorAllTraces()																		// colors traces with 12 different colors (hard-coded)
	string traces, gName = GetTopGraph()
	variable trNum = FetchAllTraces(gName,traces)
	if (trNum == 0)
		return 0
	endif
	variable i, color, leap = 1
	
	string header = "["+gName+" color backup]"
	string backup = GetScrapText()
	if (!CmpStr(StringfromList(0,backup,"\r"),header))											// load backup on second button press
		for (i = 1; i < ItemsInList(backup,"\r"); i++)
			Execute/Q StringfromList(i,backup,"\r")
		endfor
		PutScrapText ""
		return 0
	else
		backup = header + "\r" + GrepList(WinRecreation(gName,1),"(rgb)",0,"\r")				// write backup on first button press
		PutScrapText backup
	endif
	
	if (trNum < 6)
		Make/FREE colors = {{65280,0,0},{10,16020,65535},{2,39321,1},{29524,1,58982},{0,0,0}}
	elseif (trNum < 8)
		Make/FREE colors = {{65280,0,0},{65280,43520,0},{10,16020,65535},{2,39321,1},{29524,1,58982},{0,0,0},{0,56540,56540}}
	else
		Make/FREE colors = {{65280,0,0},{65280,43520,0},{0,65280,0},{0,52224,0},{2,39321,1},{0,56540,56540},{0,43520,65280},{0,15872,65280},{26411,1,52428},{44253,29492,58982},{65535,29555,59110},{26112,26112,26112},{0,0,0}}
		leap = trNum<DimSize(colors,1)/2 ? trunc((DimSize(colors,1)-1)/(trNum-1)) : 1			// introduce a leap value to choose more distant colors for only a few traces
	endif
	for (i = 0, color = 0; i < trNum; i++, color+=leap)
		color = color>=DimSize(colors,1) ? 0 : color											// loop colors around
		ModifyGraph/W=$gName rgb($StringFromList(i,traces))=(colors[0][color],colors[1][color],colors[2][color])
	endfor
end

//________________________________________________________________________________________________
//
//						Add Legend with Trace Offset and Scaling Infos
//________________________________________________________________________________________________

static function InsertIntoMainProc(string newCode)
	string currScrap = GetScrapText()															// copy current scrap text
	GetWindow Procedure hide; int wasHidden = V_Value != 0
	DisplayProcedure/W=Procedure
	DoIgorMenu "Edit" "Select All"
	DoIgorMenu "Edit" "Copy"
	PutScrapText GetScrapText()+newCode															// modify procedure code
	DoIgorMenu "Edit" "Paste"
	PutScrapText currScrap			 															// put previous scrap text back
	Execute/P/Q/Z "COMPILEPROCEDURES "															// recompile all
	if (wasHidden)
		HideProcedures																			// hide all procedure windows
	endif
end
//------------------------------------------------------------------------------------------------
static function/S OffsetFuncCode()																// inserts function for a live display of trace offset
	string newCode = "\r// Below function is part of the live-info legend of the Graph Tools package. May be deleted if no such annotations are in use.\r"
	newCode += "function/S GT_liveInfo(string gName, string trace, variable which, int digits)\r"
	newCode += "\tif (which<0 || which>3 || !WinType(gName))\t\t// which modes: [0] = x offset, [1] = y offset, [2] = x multiplier, [3] = y multiplier"
	newCode += "\r\t\treturn \"\"\r\tendif\r\tvariable mode = which>1, xval, yval;\tstring result\r"
	newCode += "\tsscanf StringByKey(SelectString(mode,\"offset(x)\",\"muloffset(x)\"), TraceInfo(gName,trace,0), \"=\"), \"{%f,%f}\", xval, yval\r"
	newCode += "\tsprintf result, \"%.*g\", limit(digits,0,16), (which-2*mode ? yval+(yval==0)*mode : xval)\r\treturn result\rend"
	return newCode
end

//################################################################################################

static StrConstant kAnchorList = "LB;MB;RB;LC;MC;RC;LT;MT;RT;"

static function QuickOffsetLegendWrapper()
	variable mode = 1, add = 1, pos = 1, rnd = 6
	string hlpStr = "This tool adds a function GT_liveInfo(graphName, traceName, mode) to the current experiment to display live offset information of traces."
	Prompt mode,"1st Information:",popup,"x Offsets;y Offsets;x Multiplier;y Multiplier;"
	Prompt add, "2nd Information:",popup,"None;x Offsets;y Offsets;x Multiplier;y Multiplier;"
	Prompt pos, "Legend Position:",popup,"Default;LB: left bottom;MB: mid bottom;RB: right bottom;LC: left center;MC: mid center;RC: right center;LT: left top;MT: mid top;RT: right top;"
	Prompt rnd, "No. of Significant Digits (Rounding):",popup,"0;1;2;3;4;5;6;7;8;9;10;"
	DoPrompt/HELP=hlpStr "Which information should be displayed in the legend of "+GetTopGraph()+"?",mode, add, rnd, pos
	if (V_flag)
		return -1
	endif
	string cmd = "QuickOffsetLegend("+num2str(mode-1)
	cmd += SelectString((add!=1),"", ", addinfo="+num2str(add-2))
	cmd += SelectString((pos!=1),"", ", anchor=\""+StringFromList(pos-2, kAnchorList)+"\"")
	cmd += SelectString((rnd!=6),"", ", digits="+num2str(rnd-1)) + ")"
	Execute/Z/Q cmd
	if (QuickMenu_Print)
		Print "•"+cmd
	endif
end

//------------------------------------------------------------------------------------------------

// ++++++ parameter explanation
//	which		= modes: [0] x offset, [1] y offset, [2] x multiplier, [3] y multiplier
//	graphName	= graph to act on (will be top main graph if omitted)
//	annoName	= annotation (legend) name used for creation - needs to be unique
//	addinfo		= 2nd info to display (values are the same as which above)
//	anchor		= anchor position on graph
//	digits		= the number of significant digits for displaying values
//	frame		= type of legend frame: [0] none, [1] underline, [2] box
//	background	= type of background used for the legend: [0] opaque, [1] transparent, [2] graph color, [3] window color

function QuickOffsetLegend(int which, [, string graphName, string annoName, int addinfo, string anchor, int digits, int frame, int background])
	graphName = SelectString(ParamIsDefault(graphName) || !strlen(graphName), graphName, GetTopGraph())
	annoName  = SelectString(ParamIsDefault(annoName), annoName, "")
	anchor	  = SelectString(ParamIsDefault(anchor) || WhichListItem(anchor, kAnchorList,";",0,0)<0, anchor, "")
	string list = ""
	if (!strlen(graphName) || WinType(graphName)!=1 || which < 0 || which > 3 || FetchAllTraces(graphName,list)<1)
		return -1
	endif
	frame		= ParamIsDefault(frame)		?  0 : limit(frame,0,2)
	background	= ParamIsDefault(background)?  1 : limit(background,0,3)
	addinfo		= ParamIsDefault(addinfo)	? -1 : limit(addinfo,-1,3)							// default is no info
	digits		= ParamIsDefault(digits)	?  5 : limit(digits,0,16)							// default is 5 significant digits
	if (!strlen(annoName) || WhichListItem(annoName,AnnotationList(graphName))<0)				// does not exist yet
		Execute/Q/Z "Legend/W="+graphName+SelectString(strlen(annoName),"","/N="+annoName)+"/F="+num2str(frame)+"/B="+num2str(background)+SelectString(strlen(anchor),"","/A="+anchor)
		if (V_flag)
			return 0
		endif
	endif
	if (!strlen(annoName))
		list = AnnotationList(graphName)
		annoName = StringFromList(ItemsInList(list)-1,list)
	endif
	QuickLegendStyle(graphName=graphName, annoName=annoName, liveInfo=buildLiveInfoFormat(which, addinfo, digits, digits))
	return 0
end

//------------------------------------------------------------------------------------------------

static function/S buildLiveInfoFormat(int mode1, int mode2, int digit1, int digit2)
	string cmd = "("+StringFromList(mode1,"Δx = ;Δy = ;x * ;y * ;")+"%i"+num2str(mode1)+"."+num2str(digit1)
	if (mode2>-1)
		cmd += ", " +StringFromList(mode2,"Δx = ;Δy = ;x * ;y * ;")+"%i"+num2str(mode2)+"."+num2str(digit2)
	endif
	return SelectString(mode1<0, cmd+")", "")
end

//################################################################################################

// ++++++ parameter explanation
//	graphName	= graph to act on (will be top main graph if omitted)
//	annoName	= annotation(legend) to act on (will be top annotation if omitted)
//	color		= color mode: [0] no change, [1] black, [2] trace color, [3] set RGB color
//	columns		= number of trace columns in the legend
//	fontStyle	= bit-wise font style option analogous to the \fXX escape sequence where XX is the numerical setting
//	fontZoom	= relative font size in percent
//	tabsize		= distance of tabs in points (no effect if columns < 2)
//	cR,cG,cB	= define a color for color mode 3 in RGB - valid values are 0-65535
//	moveTo		= moves legend to default anchor position
//	liveInfo	= special string to add live trace information
//				  format:  (x ... ) or (y ... ) or (Δx ... ) or (Δy ... ) with the desired info as %iX.Y, where X is the info mode and Y is the number of significant digits
//				  example: "(Δx = %i0.5, Δy = %i1.4)"

function QuickLegendStyle([string graphName, string annoName, int color, int columns, int fontStyle, int fontZoom, int tabSize, int cR, int cG, int cB, string moveTo, string liveInfo])		// styles an existing legend
	graphName = SelectString(ParamIsDefault(graphName) || !strlen(graphName), graphName, GetTopGraph())
	annoName  = SelectString(ParamIsDefault(annoName), annoName, "")
	moveTo	  = SelectString(ParamIsDefault(moveTo) || WhichListItem(moveTo, kAnchorList,";",0,0)<0, moveTo, "")
	if (WinType(graphName) != 1)
		return 0
	endif
	if (!strlen(annoName))
		annoName = StringFromList(ItemsInList(AnnotationList(graphName))-1,AnnotationList(graphName))
	endif
	if (!strlen(annoName) || WhichListItem(annoName,Annotationlist(graphName))<0)
		return 0
	endif
	
	// +++++++++++++++++++++++++++++++ prepare options and text ++++++++++++++++++++++++++++++++++
	liveInfo	= SelectString(ParamIsDefault(liveInfo), liveInfo, "")
	fontStyle	= ParamIsDefault(fontStyle) ? -1 : limit(fontStyle,0,99)
	fontZoom	= ParamIsDefault(fontZoom)  ? -1 : limit(fontZoom,0,999)
	color		= ParamIsDefault(color) ? 0 : color
	columns		= ParamIsDefault(columns) ? 0 : columns											// number of columns in the layout
	cR = ParamIsDefault(cR) ? 0 : limit(cR,0,65535)
	cG = ParamIsDefault(cG) ? 0 : limit(cG,0,65535)
	cB = ParamIsDefault(cB) ? 0 : limit(cB,0,65535)
	
	string text = StringByKey("TEXT",AnnotationInfo(graphName,annoName)), cmd = ""
	text = ReplaceString(num2char(92)+num2char(34),text,num2char(34))							// replace \" with "
	text = ReplaceString("\\\\",text,"\\")														// replace escaped backslash
	text = ReplaceString("\\r\\n",text,"\r")
	text = ReplaceString("\\t",text,"\t")
	text = ReplaceString("\\r",text,"\r")
	text = ReplaceString("\\n",text,"\r")
	
	string info1, info2
	// +++++++++++++++++++++++++++++++++++ live trace info +++++++++++++++++++++++++++++++++++++++
	if (strlen(liveInfo) && strlen(text))
		string mode1, mode2, digit1, digit2, modeList = "Δx = ;Δy = ;x * ;y * ;"
		SplitString/E=("\(.*?(%i[0-9]\.?[0-9]?)(?:.*?(%i[0-9]\.?[0-9]?))?.*\)") liveInfo, info1, info2
		
		if (strlen(info1) || strlen(info2))
			if (!strlen(FunctionInfo("GT_liveInfo","Procedure")))								// first inject offset macro into experiment
				InsertIntoMainProc(OffsetFuncCode())
			endif
			
			SplitString/E=("%i([0-9])\.?([0-9])?") info1, mode1, digit1
			SplitString/E=("%i([0-9])\.?([0-9])?") info2, mode2, digit2
			variable m1_val = str2num(mode1);  m1_val = numtype(m1_val) ?  0 : m1_val
			variable m2_val = str2num(mode2);  m2_val = numtype(m2_val) ? -1 : m2_val
			
			cmd = "\\{GT_liveInfo(\""+graphName+"\",\"%tName\","+num2str(m1_val)+","+SelectString(strlen(digit1),"5",digit1)+")}"
			liveInfo = ReplaceString(info1,liveInfo,cmd,1,1)
			if (m2_val > -1)
				cmd ="\\{GT_liveInfo(\""+graphName+"\",\"%tName\","+num2str(m2_val)+","+SelectString(strlen(digit2),"5",digit2)+")}"
				liveInfo = ReplaceString(info2,liveInfo,cmd,1,1)
			endif
			
			text = modifyLegendTextTags(text, "", liveInfo, remInText="(\s?\([Δ|x|y].*?\\\\{GT_live.*?}\))", graphName=graphName, fromTrace="insertTrace")
		elseif (!CmpStr(liveInfo,"remove"))
			text = modifyLegendTextTags(text, "", "", remInText="(\s?\([Δ|x|y].*?\\\\{GT_live.*?}\))")		// just remove all info tags
		endif
	endif
	
	// ++++++++++++++++++++++++++++++++++++ text coloring ++++++++++++++++++++++++++++++++++++++++
	if (color>0 && strlen(text))
		Switch (color)
			case 2:
				text = modifyLegendTextTags(text, "\K", "\K(0,0,0)", remInText="(\\\\K\(.*?\))", graphName=graphName, fromTrace="rgb(x)")
			break
			case 3:
			if (cR || cG || cB)
				text = modifyLegendTextTags(text, "\K("+num2str(cR)+","+num2str(cG)+","+num2str(cB)+")", "\K(0,0,0)", remInText="(\\\\K\(.*?\))")
			endif
			break
			default:
				text = modifyLegendTextTags(text,"","", remInText="(\\\\K\(.*?\))")							// just remove color tags
		EndSwitch
	endif
	
	// ++++++++++++++++++++++++++++++++++++ text font size +++++++++++++++++++++++++++++++++++++++
	if (fontZoom>-1 && strlen(text))
		info1 = ""; info2 = ""
		if (fontZoom!=100 && fontZoom>0)
			sprintf info1, "\Zr%.3d", fontZoom;	info2 = "\M"									// additionally exchange all /M tags in trace explanations with the new size
			text = modifyLegendTextTags(text, info1, info2, remInText="(\\\\Zr[0-9]{3})", remEndTag="(\\\\M)", replaceInText="\\M", withThisText="\\M"+info1)
		else
			text = modifyLegendTextTags(text, info1, info2, remInText="(\\\\Zr[0-9]{3})", remEndTag="(\\\\M)")	// just remove font size tags
		endif
	endif
	
	// +++++++++++++++++++++++++++++++++++ text font style +++++++++++++++++++++++++++++++++++++++
	if (fontStyle>-1 && strlen(text))
		info1 = ""; info2 = ""
		if (fontStyle>0)
			sprintf info1, "\f%.2d", fontStyle;	info2= "\f00"
		endif
		text = modifyLegendTextTags(text, info1, info2, remInText="(\\\\f[0-9]{2})")
	endif
	
	// +++++++++++++++++++++++++++++++++++++ text columns ++++++++++++++++++++++++++++++++++++++++
	int i, col = 0
	if (columns>0 && strlen(text))
		text = ReplaceString(" \t", text, "")													// remove all special tabs
		do
			SplitString/E=("(?<!\\r)([^\\r|\r]\\\\s)") text, cmd								// find \s with something before them
			if (strlen(cmd))
				text = ReplaceString(cmd, text, ReplaceString("\\s",cmd,"")+"\r\\s")			// break at \s tabs with are within the text (resets columns)
			endif
		while(strlen(cmd))
		
		wave/T rows = ListToTextWave(text, "\r");	text = ""
		for (i=0; i<DimSize(rows,0)-1; i++)
			text += rows[i]
			if (StringMatch(rows[i],"*\s(*") &&StringMatch(rows[i+1],"*\s(*") && col<columns-1)	// align into tabbed rows until column limit is reached
				text+=" \t";	col += 1
			else
				text+= "\r";	col  = 0
			endif
		endfor
		text += rows[i]
		text = RemoveEnding(RemoveEnding(text," \t"),"\r")
	endif
	
	// +++++++++++++++++++++++++++++++++++ rebuild legend ++++++++++++++++++++++++++++++++++++++++
	cmd  = StringByKey("TYPE",AnnotationInfo(graphName,annoName))
	cmd += "/W="+graphName+"/C/N="+annoName+"/J"+SelectString(!ParamIsDefault(tabsize),"","/T="+num2str((limit(tabsize,0,2000))))
	cmd += SelectString(strlen(moveTo),"","/A="+moveTo+"/X=5.00/Y=5.00")
	Execute/Q cmd + " \"\""																		// first, delete the text
	
	int items = ItemsInList(text,"\r")
	for (i=0; i<items; i++)
		AppendText/W=$graphName/N=$annoName/NOCR StringFromList(i,text,"\r") + SelectString(i<items-1,"","\r")
	endfor
end

//------------------------------------------------------------------------------------------------

static function/S modifyLegendTextTags(string text, string preTag, string postTag[, string graphName, string fromTrace, string remInText, string remEndTag, string replaceInText, string withThisText])
	wave/T rows = ListToTextWave(text, "\s(")													// split along "\s"
	if (numpnts(rows)<2)
		return text
	endif
	rows[1,numpnts(rows)-1] = "\s("+rows[p]														// add "\s" back to the start
	
	graphName = SelectString(ParamIsDefault(graphName),graphName,"")							// graph name for trace info
	fromTrace = SelectString(ParamIsDefault(fromTrace),fromTrace,"")							// what to graph from trace info
	remInText = SelectString(ParamIsDefault(remInText),remInText,"")							// grep remove in main text
	remEndTag = SelectString(ParamIsDefault(remEndTag),remEndTag,remInText)						// remove from end tags
	replaceInText = SelectString(ParamIsDefault(replaceInText),replaceInText,"")				// also replace strings within main text
	withThisText  = SelectString(ParamIsDefault(withThisText),withThisText,"")
	
	string tmpPre = preTag, tmpPost = postTag													// backup for tag replacement
	string preTxt, traceID, tr_text, tempStr, trcName, trcInst; int i
	for (i=1; i<DimSize(rows,0); i++)
		// ++++++++++++++++++++++++++++++++++ split into parts +++++++++++++++++++++++++++++++++++
		int isLiberal = StringMatch(rows[i],"*\s('*")
		SplitString/E=(SelectString(isLiberal,"\\\\s\(([^\)]*)","\\\\s\('([^\)]*)")) rows[i], traceID
		SplitString/E=("(.*?)(?:\\\\s\()") rows[i], preTxt
		SplitString/E="([^#]*)(.*)" traceID, trcName, trcInst
		trcName = RemoveEnding(trcName , "'")	
		traceID = PossiblyQuoteName(trcName)+trcInst											// FIRST: put the trace name back in quotes
		tr_text = ReplaceString(preTxt+"\s("+traceID+")",rows[i],"")							// the remaining string is the description
		
		string omitTxt = "", omitTag = "", tr_info = ""
		if (StringMatch(tr_text,"*\r*"))														// check for additional text unrelated to the trace description
			omitTxt = ReplaceString(StringFromList(0,tr_text,"\r"),tr_text,"")					// modify only up to the first line break
			tr_text = StringFromList(0,tr_text,"\r")
		endif
		do																						// put all end tags into omitTag
			SplitString/E=("(\\\\K\(0,0,0\)|\\\\M|\\\\f00|\\t|\s)$") tr_text, tempStr
			omitTag += tempStr
			tr_text = RemoveEnding(tr_text, tempStr)
		while(strlen(tempStr))
		
		// ++++++++++++++++++++++++++++++++++++ remove strings +++++++++++++++++++++++++++++++++++
		if (strlen(remInText))
			do																					// remove requested strings
				SplitString/E=(remInText) tr_text, tempStr
				tr_text = ReplaceString(tempStr, tr_text, "")
			while(strlen(tempStr))
		endif
		if (strlen(remEndTag))
			do																					// remove requested tags
				SplitString/E=(remEndTag) omitTag, tempStr
				omitTag = ReplaceString(tempStr, omitTag, "")
			while(strlen(tempStr))
		endif
		
		// ++++++++++++++++++++++++++++++++++++ replace strings +++++++++++++++++++++++++++++++++++
		if (strlen(replaceInText))
			tr_text = ReplaceString(replaceInText,tr_text,withThisText)
		endif
		
		// +++++++++++++++++++++++++++++++++ add trace information +++++++++++++++++++++++++++++++
		if (strlen(fromTrace) && WinType(graphName)==1)
			if (!CmpStr(fromTrace,"insertTrace"))												// special keyword to replace %tName with the trace name
				preTag  = ReplaceString("%tName", tmpPre,  traceID)
				postTag = ReplaceString("%tName", tmpPost, traceID)
				if (strlen(postTag) && CmpStr(postTag[0],"\\") && !(!CmpStr(tr_text[strlen(tr_text)-1]," ") || !CmpStr(tr_text[strlen(tr_text)-1],"\t")))
					postTag[0] = " "															// separate text from trace info with a space character
				endif
			else
				tr_info = StringByKey(fromTrace,TraceInfo(graphName,traceID,0),"=")
			endif
		endif
		
		// ++++++++++++++++++++++++++++++++++++ rebuild text +++++++++++++++++++++++++++++++++++++
		rows[i] = preTxt + "\s("+traceID+")" + preTag + tr_info + tr_text + postTag + omitTag + omitTxt
	endfor
	
	wfprintf text, "%s", rows
	return text
end

//________________________________________________________________________________________________
//
//							Scaling Normalization of Graph Traces
//________________________________________________________________________________________________

static Structure ScaleSettingsStruct
	string csr
	variable offMode
	variable xRange
	variable doSmth
	variable wExport
EndStructure

static function readScaleSettings(STRUCT ScaleSettingsStruct &s)
	string opt = StrVarOrDefault("root:"+ksMenuOptionsSave,ksMenuOptionsDefault)				// load options
	int zeroY0 = (NumberByKey("offsetMode",opt) == 1) || ((GetKeyState(0)&4) != 0)				// don't shift background if shift is pressed
	int keepY0 = (NumberByKey("offsetMode",opt) == 2) || ((GetKeyState(0)&1) != 0)				// don't touch offset if ctrl is pressed
	s.offMode = keepY0 ? 2 : (zeroY0 ? 1 : 0)
	s.csr	  = StringByKey("cursor",opt)
	s.xRange  = NumberByKey("cursorRange",opt)/100												// x range around cursor to look for maximum in percent
	s.doSmth  = NumberByKey("normSmooth" ,opt)													// smoothing factor
	s.wExport = NumberByKey("waveExport" ,opt)
	s.wExport = numtype(s.wExport) != 0 ? 0 : s.wExport
end

//------------------------------------------------------------------------------------------------

static function ScaleMainMenuWrapper(string type)												// call ScaleAllTraces from the top menu
	STRUCT ScaleSettingsStruct s;	readScaleSettings(s)
	int nMode = !CmpStr(type,"Cursor") ? 2 : (!CmpStr(type,"Area") ? 3 : 0)
	string cmd = buildScaleCall(smth=s.doSmth,normMode=nMode,offMode=s.offMode,relRange=s.xRange,csr=SelectString(nMode==2,"None","A"),export=s.wExport)
	Execute/Z/Q cmd
	if (QuickMenu_Print)
		Print cmd
	endif
end

//------------------------------------------------------------------------------------------------

static function ScaleClickWrapper(string type)													// call ScaleAllTraces from right-click menu
	GetLastUserMenuInfo
	string gName = S_graphName, trace = S_traceName, selText = S_Value;	int selVal = V_Value
	if (V_flag != 0)
		return 0
	endif
	
	STRUCT ScaleSettingsStruct s;	readScaleSettings(s)
	
	int rMode = limit(selVal-3,0,inf)
	int nArea = StringMatch(selText,"*Area*")
	int nMode = 0
	nMode = !CmpStr(type,"One") ?	 0+3*nArea : nMode
	nMode = !CmpStr(type,"Global") ? 1+3*nArea : nMode
	nMode = !CmpStr(type,"Cursor") ? 2 : nMode
	s.csr = SelectString(limit(selVal-2,0,inf)>0,"None",s.csr)									// cursor related entries start from the third one
	
	variable xpos = NaN, ypos = NaN, axis = NaN
	if (CmpStr(s.csr,"Mouse") == 0)																// get mouse position
		string xAxis = "", yAxis = "",traces = ""
		if (strlen(trace))
			xAxis = StringByKey("XAXIS",TraceInfo(gName,trace,0))
			yAxis = StringByKey("YAXIS",TraceInfo(gName,trace,0))
		else
			if (FetchAllTraces(gName,traces) == 0)												// no traces found
				return -1
			endif
			xAxis = GetMajorAxisFromTraceList(gName, traces, "X")
			yAxis = GetMajorAxisFromTraceList(gName, traces, "Y")
		endif
		axis = FindListItem(StringByKey("AXTYPE",AxisInfo(gName,xAxis)), "top;bottom;") > -1	// which type of axis do we work with: left / right or top / bottom?
		xpos = axis ? AxisValFromPixel(gName, xAxis, V_mouseX) : AxisValFromPixel(gName, xAxis, V_mouseY)
		ypos = axis ? AxisValFromPixel(gName, yAxis, V_mouseY) : AxisValFromPixel(gName, yAxis, V_mouseX)
	endif

	string cmd = buildScaleCall(name=gName,smth=s.doSmth,xpos=xpos,ypos=ypos,normMode=nMode,offMode=s.offMode,rangeMode=rMode,relRange=s.xRange,csr=s.csr,export=s.wExport)
	Execute/Z/Q cmd
	if (QuickMenu_Print)
		Print cmd
	endif
	
	return 0
end

//------------------------------------------------------------------------------------------------

static StrConstant kMarqueeMenu = "Maxima to One;Maxima to Global Maximum;Average to One;Average to Global Average;Area to One;Area to Global Maximum;"

static function/S DisplayMarqueeMenu()
	return kMarqueeMenu
end

static function ScaleMarqueeWrapper()															// call ScaleAllTraces from graph marquee
	GetLastUserMenuInfo
	string gName = S_graphName, traces = "", selText = S_Value
	if (FetchAllTraces(gName,traces) == 0)
		return -1
	endif
	
	int rMode = StringMatch(selText,"*Average*") ? 2 : 1
	int nArea = StringMatch(selText,"*Area*")
	int nMode = 0
	nMode = StringMatch(selText,"*One*") ?    0+3*nArea : nMode
	nMode = StringMatch(selText,"*Global*") ? 1+3*nArea : nMode
	
	STRUCT ScaleSettingsStruct s;	readScaleSettings(s)
	
	GetMarquee/K/W=$gName/Z $GetMajorAxisFromTraceList(gName, traces, "X")
	if (!V_flag)
		return -1
	endif
	
	variable x_center = NaN
	if (V_top == 0 && V_bottom == 0)															// horizontal axis
		x_center = (V_left+V_right)/2
		s.xRange = abs(V_right-V_left)
	else																						// vertical axis
		x_center = (V_top+V_bottom)/2
		s.xRange = abs(V_bottom-V_top)
	endif
	
	string cmd = buildScaleCall(name=gName,smth=s.doSmth,xpos=x_center,normMode=nMode,offMode=s.offMode,rangeMode=rMode,xRange=s.xRange,export=s.wExport)
	Execute/Z/Q cmd
	if (QuickMenu_Print)
		Print cmd
	endif
	
	return 0
end

//------------------------------------------------------------------------------------------------

static function/S buildScaleCall([string name, variable smth, variable xpos, variable ypos, int normMode, int offMode, int rangeMode, variable xRange, variable relRange, string csr, int export])
	string cmd = ""
	if (!ParamIsDefault(name) && strlen(name))
		cmd += ", name=\""+name+"\""
	endif
	if (!ParamIsDefault(csr)  && strlen(csr))
		if (WhichListItem(csr, kUsedCursorList,";", 0, 0) > -1)
			cmd += ", csr=\""+csr+"\""
		endif
	endif
	if (!ParamIsDefault(smth) && smth>0)
		cmd += ", smth="	+num2str(smth)
	endif
	if (!ParamIsDefault(relRange) && !numtype(relRange) && relRange != kScale_defRange)
		cmd += ", relRange="+num2str(relRange)
	endif
	if (!ParamIsDefault(xpos)     && !numtype(xpos))
		cmd += ", xpos="	+num2str(xpos)
	endif
	if (!ParamIsDefault(ypos)     && !numtype(ypos))
		cmd += ", ypos="	+num2str(ypos)
	endif
	if (!ParamIsDefault(xRange)   && !numtype(xRange))
		cmd += ", xRange="	+num2str(xRange)
	endif
	if (!ParamIsDefault(normMode) && normMode !=0)
		cmd += ", normMode="+num2str(normMode)
	endif
	if (!ParamIsDefault(offMode)  && offMode  !=0)
		cmd += ", offMode="	+num2str(offMode)
	endif
	if (!ParamIsDefault(rangeMode)&& rangeMode!=0)
		cmd += ", rangeMode="+num2str(rangeMode)
	endif
	if (!ParamIsDefault(export)   && export   !=0)
		cmd += ", export="	+num2str(export)
	endif
	cmd  = "•ScaleAllTraces(" + cmd[2,inf] + ")"
	return cmd
end

//------------------------------------------------------------------------------------------------

function ScaleTo1()																				// backwards compatibility
	ScaleAllTraces()
end

//------------------------------------------------------------------------------------------------

function ScaleAllTraces([string name, variable smth, variable xpos, variable ypos, int normMode, int offMode, int rangeMode, variable xRange, variable relRange, string csr, int export])
	export = (!ParamIsDefault(export) && export != 0)
	// these modes alter the behavior of normalization:
	int nMode = ParamIsDefault(normMode)  ? 0 : normMode		// how to normalize:		[0] = to one,  [1] = to global max,	[2] = to given y-value, [3] = area to one, [4] = area to global area	
	int oMode = ParamIsDefault(offMode)   ? 0 : offMode			// how to handle offsets:	[0] = to zero, [1] = no offset,		[2] = keep offset
	int rMode = ParamIsDefault(rangeMode) ? 0 : rangeMode		// use range around xPos:	[0] = off,     [1] = find maximum,	[2] = calculate average
	
	int useGlobScale = nMode == 1 || nMode == 4													// scale to global maximum (highest intensity of all traces)
	int useYposScale = nMode == 2																// use given y-value
	int useAreaScale = nMode == 3 || nMode == 4													// scale to trace area

	xpos = ParamIsDefault(xpos) ? NaN : xpos													// scale at specific x position
	ypos = ParamIsDefault(ypos) ? NaN : ypos
	relRange = (ParamIsDefault(relRange) || numtype(relRange)) ? kScale_defRange : limit(relRange,0.001,1)		// percentage of the total wave range used for scaling
	
//	Print "normalize:", nMode, "offset:", oMode, "range:", rMode, "position X/Y:", xpos, ypos	// debug
	
	string gName = SelectString(ParamIsDefault(name), name, GetTopGraph())
	if (WinType(gName) != 1)
		Print "No graph"+SelectString(strlen(gName),""," with name "+gName)+" found."
		return 0
	endif
	
	string traces = TraceNameList(gName,";",1), trace
	int i, items  = ItemsInList(Traces)
	if (items == 0)
		return 0
	endif
	
	variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, scale
	if (!ParamIsDefault(csr) && WhichListItem(csr, kUsedCursorList,";", 0, 0) > -1)
		if (!strlen(CsrInfo($csr,gName)))
			Abort "Cursor "+csr+" is not set."
		endif
		GetAllTraceOffsets(gName, StringByKey("TNAME",CsrInfo($csr,gName)), xOff, yOff, xMul, yMul)
		xpos = hcsr($csr, gName)*xMul + xOff													// add trace offset to the cursor position
		ypos = vcsr($csr, gName)*yMul + yOff
	endif
	
	// ++++++++++++++++++++++ extract area and height values from all traces +++++++++++++++++++++
	Make/D/Free/N=(items) minVals = NaN, maxVals = NaN, maxArea = NaN, minArea = NaN
	
	for (i = 0; i < items; i++)
		trace = StringFromList(i,traces)
		wave w = prepareWaveFromTrace(gName, trace)
		wave/Z w_x = XWaveRefFromTrace(gName, trace)
		
		if (!ParamIsDefault(smth) && smth>0)													// optional smoothing before scaling
			Smooth limit(smth,1,2^15-1), w
		endif
		
		minVals[i] = wavemin(w)
		maxVals[i] = wavemax(w)
		if (WaveExists(w_x))
			SetScale/P x, 0, 1, w																// make sure scaling is removed
			maxArea[i] = areaXY(w,w_x)
			minArea[i] = wavemin(w)*abs(wavemax(w_x)-wavemin(w_x))
		else
			maxArea[i] = area(w)
			minArea[i] = wavemin(w)*DimDelta(w,0)*(DimSize(w,0)-1)								// area baseline
		endif
		
		if (numtype(xpos))
			continue
		endif
		
		// ++++++++++++++++++ position based area and height values from all traces ++++++++++++++	
		GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
		variable scaled_xpos = (xpos - xOff) / xMul												// adjust for the clicked / selected trace's offsets
		variable scaled_xrng = (ParamIsDefault(xRange) || numtype(xRange)) ? (abs(DimDelta(w,0)*DimSize(w,0)*relRange)/xMul) : (xRange/xMul/2)
		
		if(WaveExists(w_x))
			FindLevel/P/Q w_x, scaled_xpos-scaled_xrng;	int pMin = !v_flag ? V_LevelX : 0
			FindLevel/P/Q w_x, scaled_xpos+scaled_xrng;	int pMax = !v_flag ? V_LevelX : DimSize(w,0)-1
			WaveStats/Q/R=[pMin, pMax] w
			Switch (rMode)
				case 1:
					maxVals[i] = V_max
				break
				case 2:
					maxVals[i] = V_avg
				break
				default:
					FindLevel/P/Q w_x, scaled_xpos
					maxVals[i] = !v_flag ? w[limit(V_LevelX,0,DimSize(w,0)-1)] :NaN
				break
			EndSwitch
			maxArea[i] = areaXY(w,w_x,w_x[pMin],w_x[pMax])
			minArea[i] = wavemin(w)*abs(w_x[pMax]-w_x[pMin])
		else
			WaveStats/Q/R=(scaled_xpos-scaled_xrng, scaled_xpos+scaled_xrng) w
			Switch (rMode)
				case 1:
					maxVals[i] = V_max
				break
				case 2:
					maxVals[i] = V_avg
				break
				default:
					maxVals[i] = w[limit(x2pnt(w,scaled_xpos),0,DimSize(w,0)-1)]
				break
			EndSwitch
			maxArea[i] = area(w,scaled_xpos-scaled_xrng, scaled_xpos+scaled_xrng)
			minArea[i] = wavemin(w)*scaled_xrng*2
		endif
	endfor
	
	MatrixOP/O maxVals = replaceNaNs(maxVals,1)
	MatrixOP/O minVals = replaceNaNs(minVals,0)
	MatrixOP/O maxArea = replaceNaNs(maxArea,1)
	MatrixOP/O minArea = replaceNaNs(minArea,1)
	WaveStats/Q maxVals;	variable targetMax  = useGlobScale ? V_max-minVals[V_maxLoc]*(!oMode) : 1
	WaveStats/Q maxArea;	variable targetArea = useGlobScale ? V_max-minArea[V_maxLoc]*(!oMode) : 1
	targetMax  = useYposScale ? ypos : targetMax
	targetMax  = numtype(targetMax)  != 0 || targetMax  == 0 ? 1 : targetMax
	targetArea = numtype(targetArea) != 0 || targetArea == 0 ? 1 : targetArea
	
	// ++++++++++++++++++++++++ apply scaling / shift to all traces ++++++++++++++++++++++++++++++	
	Make/Free/D/N=(items,4) outWave=NaN
	SetDimLabel 1, 0, $("y Offset"),	 outWave
	SetDimLabel 1, 1, $("y Multiplier"), outWave
	SetDimLabel 1, 2, $("x Offset"),	 outWave
	SetDimLabel 1, 3, $("x Multiplier"), outWave
	
	for (i = 0; i < items; i++)
		trace = StringFromList(i,traces)
		variable currMin  = !oMode ? minVals[i] : 0
		variable baseArea = !oMode ? minArea[i] : 0
		
		scale = useAreaScale ? (maxArea[i]-baseArea)/targetArea  : (maxVals[i]-currMin)/targetMax	// scale to match global area / maximum or to unit area / one
		scale = scale == 0 ? 1 : scale
		
		string cleanName = cleanStr(trace,"\"|'|,|;|")[0,255]
		SetDimLabel 0, i, $cleanName, outWave
		
		Switch (oMode)
			case 0:		// adjust to Y=0
				ModifyGraph/W=$gName offset($trace)={,-currMin/scale}
				outWave[i][0] = -currMin/scale
			break
			case 1:		// offset = 0
				ModifyGraph/W=$gName offset($trace)={,0}
				outWave[i][0] = 0
			break
			case 2:		// keep offset => do nothing
				outWave[i][0] = GetTraceOffset(gName, trace, 1)
			break
		EndSwitch
		
		outWave[i][1] = 1/scale
		scale = abs(scale-1)<10^-12 ? 0 : 1/scale												// no scaling for maximal trace
		ModifyGraph /W=$gName muloffset($trace)={,scale}
		
		sscanf StringByKey("offset(x)",   TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff	// also record x offsets / scales
		sscanf StringByKey("muloffset(x)",TraceInfo(gName,trace,0), "="), "{%f,%f}", xMul, yMul
		outWave[i][2] = xOff
		outWave[i][3] = xMul == 0 ? 1 : xMul
	endfor
	
	if (export)
		Note outWave, "Scaling and offsets for traces in graph '"+gName+"'\rOutput done on " + Secs2Date(DateTime,-2) + ", " + time()
		Duplicate/O outWave, $("Offsets_"+gName)
	endif
	
	return 0
end
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function/S cleanStr(string str, string list)
	int i
	for (i = 0; i < ItemsInList(list,"|"); i++)
		str = ReplaceString(StringFromList(i,list,"|"), str, "")
	endfor
	return str
end

//________________________________________________________________________________________________
//
//					Creates Waterfall Plot from a 2D wave or a Folder of Waves
//________________________________________________________________________________________________

function SpectraWaterfall([string path])
	if (!ParamIsDefault(path))
		wave/Z inwave = $path
	else
		wave/Z inwave = $GetBrowserSelection(0)													// get the first selected wave
	endif
	if (!WaveExists(inwave))
		return -1
	endif
	DFREF saveDFR = GetDataFolderDFR()
	
	string list	= ""
	int is1D = WaveDims(inwave)==1
	if (is1D)
		SetDataFolder GetWavesDataFolderDFR(inwave)
		list = SortList(WaveList("!Int_*",";","DIMS:1"), ";", 16)
	endif
	
	string wTitle = SelectString(is1D,NameOfWave(inwave),GetWavesDataFolder(inwave,0))			// name of 2D wave or folder
	Display/W=(230,50,650,450)/K=1/N=$UniqueName("WaterfallGraph", 6, 0) as "Waterfall of "+wTitle
	
	WaveStats/Q inwave;	variable initOffset	= V_max
	int i, items = is1D ? ItemsInList(list) : DimSize(inwave,1)
	for (i=items-1; i>-1; i--)																	// go backwards (in reverse order, so that the later traces are BEHIND the earlier ones)
		if (is1D)
			AppendToGraph/Q $StringFromList(i,list)
			WaveStats/Q     $StringFromList(i,list)
			initOffset = initOffset < V_max ? V_max : initOffset
		else
			AppendToGraph/Q inwave[][i]
		endif
	endfor
	SetDataFolder saveDFR
	
	Execute/Q "BoxStyle()"
	ColorSpectrumFromTable(0, "Rainbow")
	
	list = TraceNameList("",";",1)
	for (i = 1; i < items; i++)
		ModifyGraph offset($StringFromList(items-1 - i,list))={,i*initOffset*0.02}				// the initial distance will be set in percent from the maximal value of the selected wave
	endfor
	return 0
end

//________________________________________________________________________________________________
//
//							Right-Click Offset Menus for Graph Traces
//				Adapted from the StackTraces procedures v1.20 by Tony Withers
//________________________________________________________________________________________________

static function/S MenuWithinPlotrange(string s_menu)											// use to create a menu item in a popup when you need to know where the user clicks to invoke the popup
	if (WinType("") != 1)																		// don't create variables if Igor is just rebuilding the menu
		return ""
	endif
	GetMouse/W=kwTopWin
	if (V_left<0 || V_top<0)
		return ""
	endif
	return s_menu
end

//################################################################################################

static function GetYfromWave(string gName, string trace, variable xPos, variable &yVal, variable &yMin)
	wave w = prepareWaveFromTrace(gName, trace)
	wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	yMin = WaveMin(w);	yVal = NaN
	if(WaveExists(w_x))
		FindLevel/P/Q w_x, xPos
		if (!v_flag && v_levelX > 0 && v_levelX < DimSize(w,0)-1)								// make sure to stay within range
			yVal = w[round(v_levelX)]
		endif
	else
		if (x2pnt(w, xPos) > 0 && x2pnt(w, xPos) < DimSize(w,0)-1)
			yVal = w(xPos)
		endif
	endif
	return numtype(yVal)																		// error return
end

//################################################################################################

static function GetMinMaxfromWave(string gName, string trace, variable &yMin, variable &yMax)	// figure out plotted x-range for trace and finds wave's max and min values in this range
	wave w = prepareWaveFromTrace(gName, trace)
	wave/Z w_x = XWaveRefFromTrace(gName, trace)
	
	variable xOff = 0, yOff = 0, xMul = 0, yMul = 0												// x-offset and y-scale aware
	GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
	GetAxis/W=$gName/Q $StringByKey("XAXIS",TraceInfo(gName,trace,0))
	if(WaveExists(w_x))
		FindLevel /Q w_x, v_min;	v_min = !v_flag ? V_LevelX : 0								// no /P flag here - assumes same x-scaling for w and w_x
		FindLevel /Q w_x, v_max;	v_max = !v_flag ? V_LevelX : DimSize(w,0)-1
	endif
	variable fr = (v_min - xOff)/xMul
	variable to = (v_max - xOff)/xMul
	yMax = WaveMax(w, fr, to)*yMul
	yMin = WaveMin(w, fr, to)*yMul
	return 0
end

//################################################################################################

static function/wave prepareWaveFromTrace(string gName, string trace)							// extracts the correct column and removes NaNs
	wave w = TraceNameToWaveRef(gName, trace)
	Duplicate/FREE w temp
	if (WaveDims(w) == 2 && DimSize(w,1) > 1)													// handle 2D waterfall plots => extract column
		string columnStr
		SplitString/E=("\]\[(.*)\]") StringByKey("YRANGE",TraceInfo(gName,trace,0)), columnStr	// find the displayed column from TraceInfo
		variable col = str2num(columnStr)
		if (numtype(col) == 0)
			Duplicate/R=[][col]/FREE w temp
		endif
	endif
	Redimension/N=(-1,1) temp																	// needs 2D data for MatrixFilter
	MatrixFilter NanZapMedian temp																// make sure there are no NaNs
	return temp
end

//################################################################################################

static function AlignTracesAtXPos(int which)													// use trace popup to y-align traces at clicked x position by setting offsets (which = 0) or multipliers (which = 1)
	GetLastUserMenuInfo																			// figure out graph and trace names
	string gName = S_graphName, trace = S_traceName, traces = ""
	int trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	string xAxis = StringByKey("XAXIS",TraceInfo(gName, trace, 0)), yAxis = StringByKey("YAXIS",TraceInfo(gName, trace, 0))
	variable axType = FindListItem(StringByKey("AXTYPE",AxisInfo(gName,xAxis)), "top;bottom;") > -1			// which type of axis do we work with: left / right or top / bottom?
	variable xClick = axType ? AxisValFromPixel(gName, xAxis , V_mouseX) : AxisValFromPixel(gName, xAxis , V_mouseY)
	variable yClick = axType ? AxisValFromPixel(gName, yAxis , V_mouseY) : AxisValFromPixel(gName, yAxis , V_mouseX)
	variable doNorm = ((GetKeyState(0) & 2^0)!=0)												// holding ctrl / cmd aligns to zero or scales to one
	yClick = doNorm && which==0 ? 0 : yClick													// shift to zero?
	
	int i
	variable modVal
	for(i = 0; i < trNum; i++)
		trace = StringFromList(i,traces)
		variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, ywVal, ywMin
		GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
		int err = GetYfromWave(gName, trace, (xClick-xOff)/xMul, ywVal, ywMin)					// extract wave min and Y value at X position
		if (err)
			continue
		endif
		if (which)
			modVal  = doNorm ? 1 : (yClick-(yOff+ywMin)/yMul)									// scale to one?
			modVal /= (ywVal-ywMin/yMul)
			ModifyGraph/W=$gName muloffset($trace)={,modVal}									// scale to match position
			if (doNorm)																			// if normalized scale offsets down as well (because the scale difference can be huge)
				ModifyGraph/W=$gName offset($trace)={,-ywMin*modVal}
			endif
		else
			modVal = yClick-ywVal*yMul
			ModifyGraph/W=$gName offset($trace)={,modVal}										// shift to align position
		endif
	endfor
	return 0
end

//################################################################################################

static function OrderTracesAtXPos()																// reorder traces based on (possibly offset) value at x=xpos
	GetLastUserMenuInfo																			// figure out graph and trace names
	string gName = S_graphName, trace = S_traceName, traces = ""
	int trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	string xAxis = StringByKey("XAXIS",TraceInfo(gName, trace, 0))
	variable axType = FindListItem(StringByKey("AXTYPE",AxisInfo(gName,xAxis)), "top;bottom;") > -1			// which type of axis do we work with: left / right or top / bottom?
	variable xClick = axType ? AxisValFromPixel(gName, xAxis , V_mouseX) : AxisValFromPixel(gName, xAxis , V_mouseY)
	
	Make/Free/T/N=(trNum) w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum) w_order
	
	int i
	for(i = 0 ; i < trNum; i++)
		trace = StringFromList(i,traces)
		variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, ywVal, ywMin
		GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
		int err = GetYfromWave(gName, trace, (xClick-xOff)/xMul, ywVal, ywMin)					// extract wave min and Y value at X position
		w_order[i] = err ? yOff : yOff + ywVal * yMul
	endfor
	
	if (GetKeyState(0) & 2^0)																	// holding ctrl / cmd triggers descending order sorting
		Sort/R w_order, w_order, w_traces
	else
		Sort   w_order, w_order, w_traces
	endif
	
	if (trNum <= 100)																			// this even works for instances (but is limited to 100 traces)
		wfprintf traces, "%s,", w_traces
		Execute/Q "ReorderTraces _front_,{"+RemoveEnding(traces,",")+"}"
	else
		for(i = 0; i < numpnts(w_traces); i++)
			ReorderTraces/W=$gName _front_, {$w_traces[i]}
		endfor
	endif
	return 0
end

//################################################################################################

static function OrderTracesAlphabetical()														// reorder traces based on trace names
	GetLastUserMenuInfo																			// figure out graph and trace names
	string gName = S_graphName, trace = S_traceName, traces = ""
	int trNum = FetchAllTraces(gName, traces)
	if (!trNum)
		return -1
	endif
	variable sortOpt = (GetKeyState(0) & 2^0) + 16												// holding ctrl / cmd triggers descending order sorting
	
	int i, do2Dsort = 0
	string instanceList = GrepList(traces,"#")
	if (ItemsInList(instanceList))
		if (WaveDims(TraceNameToWaveRef(gName, StringFromList(0, instanceList))) == 2)			// special treatment of 2D waterfall plots
			if (trNum <= 100)																	// this method is limited to 100 traces
				do2Dsort = 1
			else
				Print "Error: Cannot sort 2D data larger that 100 traces."
				return 0
			endif
		endif
	endif
	
	if (!do2Dsort)
		traces = SortList(traces, ";", sortOpt)
		for(i = 0; i < ItemsInList(traces); i++)
			ReorderTraces/W=$gName _front_, {$StringFromList(i,traces)}
		endfor
		return 0
	endif
	
	string curtrace, basename, instance, copylist = ""
	for(i = 0; i < ItemsInList(traces); i++)													// add range for sorting to list
		curtrace = StringFromList(i,traces)
		basename = StringFromList(0,curtrace,"#")
		instance = StringFromList(1,curtrace,"#")
		copylist += basename + "@@"+StringByKey("YRANGE",TraceInfo(gName,curtrace,0))+"@@" + instance + ";"
	endfor
	
	copylist = SortList(copylist, ";", sortOpt); traces = ""
	
	for(i = 0; i < ItemsInList(copylist); i++)													// remove range
		SplitString/E=("(.*)@@(.*)@@(.*)") StringFromList(i,copylist), basename, curtrace, instance
		traces += basename + SelectString(strlen(instance),"","#") + instance + ";"
	endfor
	Execute/Q "ReorderTraces _front_,{"+RemoveEnding(ReplaceString(";",traces,","),",")+"}"
	return 0
end

//################################################################################################

static function SpreadToFillRange()																// spread (equally offset) spectra over Y axis range; traces are stacked in the order of plotting
	string traces = "", gName = GetTopGraph()
	variable trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	Make/Free/N=(trNum)/T w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum) w_offsets = GetTraceOffset(gName, w_traces[p], 1)
	variable yMin, yMax, low, high
	GetMinMaxfromWave(gName, w_traces[0], yMin, yMax);		 low  = yMin + w_offsets[0]			// find low from first trace
	GetMinMaxfromWave(gName, w_traces[trNum-1], yMin, yMax); high = yMax + w_offsets[trNum-1]	// find high from last trace
	
	GetAxis/W=$gName/Q $GetMajorAxisFromTraceList(gName, traces, "Y")
	variable bottomGap = low-v_Min, topGap = v_Max-high
	variable i, dOffset = (topGap+bottomGap)/(trNum-1)
	for (i = 0; i < trNum; i++)																	// apply offsets
		ModifyGraph/W=$gName offset($w_traces[i])={,w_offsets[i]- bottomGap + dOffset*i}
	endfor
	return 0
end

//################################################################################################

static function SpreadSpectra(int which)														// which: 0 = equal offset increment, 1 = equal spacing
	string traces = "", gName = GetTopGraph()
	int trNum = FetchAllTraces(gName,traces)
	if (!trNum)
		return -1
	endif
	
	Make/Free/N=(trNum)/T w_traces = StringFromList(p,traces)
	Make/Free/N=(trNum)/wave w_waves = TraceNameToWaveRef(gName,w_traces)
	Make/Free/N=(trNum) w_low, w_high, w_range	
	variable i, yMin, yMax
	for (i = 0; i < trNum; i++)
		GetMinMaxfromWave(gName, w_traces[i], yMin, yMax);  w_low[i] = yMin;  w_high[i] = yMax;
	endfor
	w_range = w_high-w_low
	
	GetAxis/W=$gName/Q $GetMajorAxisFromTraceList(gName, traces, "Y")
	variable yAxMin = v_min, yAxRange = (v_max-v_min), dRange = sum(w_range)
	variable gap = which ? (yAxRange-dRange)/(trNum-1) : (yAxRange-w_range[trNum-1])/(trNum-1)	// gap or offset between each trace (defined as distance between highest value of spectrum n and lowest value of spectrum n+1)
	variable newOff
	for (i = 0; i < trNum; i++)
		if (i == 0)
			newOff = yAxMin - w_low[0]															// add this to bring first trace down to bottom
		else
			newOff = which ? w_high[i-1] - w_low[i] + gap : w_low[i-1] - w_low[i] + gap
		endif
		ModifyGraph/W=$gName offset($w_traces[i])={,newOff}
		w_high[i] += newOff
		w_low[i]  += newOff
	endfor
	return 0
end

//################################################################################################

static function/S GetTopMainGraph()
	return StringFromList(0,WinList("*", ";", "WIN:1"))
end

static function/S GetTopGraph()
	string windows = WinList("*", ";", "")
	string topwin = StringFromList(0, windows)
	int i = 1
	do																							// try finding a valid graph, even if it is not the top one
		int wtype = WinType(topwin)
		if (wtype == 1 || wtype == 3 || wtype == 7)												// top window is graph or a type that can contain a graph
			GetWindow $topwin, activeSW															// for the active sub window (this can also be the top-level, if there are not children, or no child is active)
			if (strlen(S_Value) > 0 && WinType(S_Value) == 1)
				return S_Value																	// this isn't necessarily an immediate child of top window
			elseif (wtype == 1)
				return topwin
			endif
		endif
		topwin = StringFromList(i, windows)
		i++
	while (strlen(topwin) > 0)
	return ""																					// no graph found
end

//________________________________________________________________________________________________
//
//										TRACE OFFSET PANEL
//________________________________________________________________________________________________

function BuildTraceOffsetPanel()
	if (WinType("TraceOffsetPanel"))
		DoWindow/F TraceOffsetPanel
		return 0
	endif
	
	variable pW = 224, pH = 450
	
	variable TFTResX, TFTResY																	// get the display resolution to align window better (center of full screen)
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,",")) - Str2Num(StringFromList(0,TFTInfo,",")))*(72/ScreenResolution)  
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,",")) - Str2Num(StringFromList(1,TFTInfo,",")))*(72/ScreenResolution)  
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	
	NewPanel/K=1/W=((TFTResX-pW)/2,(TFTResY-pH)/2,(TFTResX+pW)/2,(TFTResY+pH)/2)/N=TraceOffsetPanel as "Modify Traces"
	ModifyPanel/W=TraceOffsetPanel ,fixedSize=1
	SetWindow TraceOffsetPanel hook(UpdateHook)=OffsetPanelUpdateFunc
	SetWindow TraceOffsetPanel userdata(procVersion)=num2str(kVersion)							// save procedure version

	TabControl offsetTabs ,pos={5,5} ,size={216,408} ,tabLabel(0)="All Traces" ,tabLabel(1)="Individual Trace" ,proc=TraceTool_TabControl
	TabControl offsetTabs focusring=0
	
	DefineGuide TabL={FL,5}
	DefineGuide TabR={FR,-5}
	DefineGuide TabT={FT,30}
	DefineGuide TabB={FB,-5}
	// +++++++++++++++++++++++++++++++++++ multi traces tab ++++++++++++++++++++++++++++++++++++++
	string pMulti = "TraceOffsetPanel#MultiTab"
	NewPanel/FG=(TabL, TabT, TabR, TabB)/HOST=#/N=MultiTab
	ModifyPanel frameStyle=0, frameInset=0
	ModifyPanel cbRGB=(60000,60000,60000,0)
	
	SetDrawLayer UserBack
	DrawLine 12,280,205,280
	DrawLine 12,345,205,345
	
	TitleBox MultiTargetTitle	,pos={10,3}		,size={195,18}	,title=""	,fixedSize=1		,frame=0
	
	CheckBox cHiddenTraces		,pos={13,25}	,size={60,22}	,title="Include Hidden Traces"	,value = 1		,help={"Include hidden traces when scaling."}
	SetVariable vaSetYOffset	,pos={13,70}	,size={90,16}	,title="y:"
	SetVariable vaScaleYOffset	,pos={13,110}	,size={90,16}	,title="y:"										,help={"Scales all trace offsets by this amount. Hold 'ctrl'/'command' for absolute scaling, i.e., including the baseline."} 
	SetVariable vaShiftYOffset	,pos={13,150}	,size={90,16}	,title="y:"
	SetVariable vaSetXOffset	,pos={113,70}	,size={90,16}	,title="x:"
	SetVariable vaScaleXOffset	,pos={113,110}	,size={90,16}	,title="x:"										,help={"Scales all trace offsets by this amount. Hold 'ctrl'/'command' for absolute scaling, i.e., including the baseline."} 
	SetVariable vaShiftXOffset	,pos={113,150}	,size={90,16}	,title="x:"
	
	CheckBox ChooseSet			,pos={13,50}	,size={60,22}	,title="Set or "				,value=1		,help={"Sets and overrides offsets of traces."}
	CheckBox ChooseAdd			,pos={67,50}	,size={60,22}	,title="Add"					,value=0		,help={"Adds to the current trace offsets."}
	TitleBox SetTitle			,pos={110,50}	,size={160,13}	,title="Offset per Trace"
	TitleBox ScaleTitle			,pos={13,93}	,size={160,13}	,title="Scale Offset per Trace"
	TitleBox ShiftTitle			,pos={13,133}	,size={160,13}	,title="Shift Offset of All Traces"
	Button ResetYButton			,pos={13,175}	,size={90,22}	,title="Y Reset"								,help={"Resets all Y offsets to 0."}
	Button ResetXButton			,pos={113,175}	,size={90,22}	,title="X Reset"								,help={"Resets all X offsets to 0."}
	Button ReverseButton		,pos={13,205}	,size={190,22}	,title="Reverse Sorting"						,help={"Reverses the order in which the traces are sorted in the graph."}
	SetVariable vTraceFilter	,pos={13,235}	,size={190,16}	,title="Filter:"								,help={"Apply offsets only to traces with matching name. You can use wildcard characters."}
	CheckBox cOmitFilter		,pos={47,257}	,size={90,22}	,title="Filter Excludes Traces"	,value = 0		,help={"The filter keyword is used to omit traces instead of including them."}

	PopupMenu ColorSelect		,pos={13,290}	,size={190,22}									,bodyWidth=190
	Button ColorButton			,pos={13,315}	,size={110,22}	,title="Colorize"								,help={"Colorize traces with currently selected color spectrum. To reverse spectrum hold 'ctrl'/'command'."}
	CheckBox cReverseColor		,pos={133,318}	,size={60,22}	,title="in Reverse"				,value = 0		,help={"Applies the color in reverse (also works directly by holding 'ctrl'/'command' while pressing the button.)"}
	PopUpMenu ProfileSelect		,pos={13,353}	,size={190,20}	,title="Profile"				,bodyWidth=152
	
	Button vaApplyAllButton		,pos={5,387}	,size={205,23}	,title="Apply All to Data (overwrite!)"			,help={"Applies all offsets and scales to 1D data. Will overwrite the original waves, so use this with care!"}
	// ++++++++++++++++++++++++++++++++++++++++ settings +++++++++++++++++++++++++++++++++++++++++
	SetVariable vaSetYOffset	,value=_NUM:0	,limits={0,inf,1}					,userdata="0"
	SetVariable vaScaleYOffset	,value=_NUM:1	,limits={0,inf,kMultipIncrStep}		,userdata="1"
	SetVariable vaShiftYOffset	,value=_NUM:0	,limits={-inf,inf,1}
	SetVariable vaSetXOffset	,value=_NUM:0	,limits={-inf,inf,0.1}				,userdata="0"
	SetVariable vaScaleXOffset	,value=_NUM:1	,limits={0,inf,kMultipIncrStep}		,userdata="1"
	SetVariable vaShiftXOffset	,value=_NUM:0	,limits={-inf,inf,1}
	SetVariable vTraceFilter	,value=_STR:"*"								,proc=TraceTool_TraceFilterProc
	
	PopupMenu ColorSelect		,mode=2	,value="*COLORTABLEPOPNONAMES*"							// must be mode first
	PopupMenu ProfileSelect		,mode=1	,value=TraceTool_ProfileWaveList()	,proc=TraceTool_ProfileSelect
	
	ModifyControlList ControlNameList(pMulti,";","Choose*")	,win=$pMulti	,proc=TraceTool_SwitchOffsetMode,mode = 1
	ModifyControlList ControlNameList(pMulti,";","*Button")	,win=$pMulti	,proc=TraceTool_ButtonFunctions
	ModifyControlList ControlNameList(pMulti,";","vaS*") 	,win=$pMulti	,proc=TraceTool_MultiTraceVars	,format="%g"
	ModifyControlList ControlNameList(pMulti,";","*Title")	,win=$pMulti	,frame=0
	ModifyControlList ControlNameList(pMulti,";","vaS*")+"ColorSelect;ProfileSelect;" ,win=$pMulti ,focusRing=0
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pMulti,";","*")	,win=$pMulti	,fsize=10
	endif
	SetActiveSubwindow ##
	
	// ++++++++++++++++++++++++++++++++++++ single trace tab +++++++++++++++++++++++++++++++++++++
	string pSingle = "TraceOffsetPanel#SingleTab"
	NewPanel/FG=(TabL, TabT, TabR, TabB)/HOST=#/N=SingleTab
	ModifyPanel frameStyle=0, frameInset=0
	ModifyPanel cbRGB=(60000,60000,60000,0)
	
	SetDrawLayer UserBack
	DrawLine 12,235,205,235
	DrawLine 12,317,205,317

	TitleBox SingleTargetTitle	,pos={10,3}		,size={195,18}	,title=""			,frame=0	,fixedSize=1

	Button CsrActButton			,pos={10,25}	,size={60,22}	,title="Cursor:"	,fstyle=1					,help={"Sets a new cursor on the graph."}
	CheckBox traceCsrA			,pos={78,28}	,size={60,22}	,title="A"										,help={"Selects cursor A (a key)."}
	CheckBox traceCsrB			,pos={111,28}	,size={60,22}	,title="B"										,help={"Selects cursor B (b key)."}
	CheckBox traceCsrC			,pos={144,28}	,size={60,22}	,title="C"										,help={"Selects cursor C (c key)."}
	CheckBox traceCsrD			,pos={177,28}	,size={60,22}	,title="D"										,help={"Selects cursor D (d key)."}
	
	TitleBox trSetTitle			,pos={13,53}	,size={170,13}	,title="Set Trace Offset"
	TitleBox trScaleTitle		,pos={13,93}	,size={170,13}	,title="Set Trace Multiplier"
	TitleBox trShiftTitle		,pos={13,133}	,size={170,13}	,title="Reposition at Cursor"
	SetVariable trSetYOffset	,pos={13,70}	,size={90,16}	,title="y:"
	SetVariable trScaleYOffset	,pos={13,110}	,size={90,16}	,title="y:"
	SetVariable trShiftYOffset	,pos={13,150}	,size={90,16}	,title="y:"
	SetVariable trSetXOffset	,pos={113,70}	,size={90,16}	,title="x:"
	SetVariable trScaleXOffset	,pos={113,110}	,size={90,16}	,title="x:"
	SetVariable trShiftXOffset	,pos={113,150}	,size={90,16}	,title="x:"
	SetVariable trScaleNorm		,pos={13,178}	,size={140,16}	,title="Normalize:"				,bodyWidth=80	,help={"Adjusts the multiplier to normalize the trace height to this value at the current cursor's position (excluding the Y offset)."}
	Button trNormTo1Button		,pos={163,177}	,size={40,20}	,title="to 1"									,help={"Normalizes trace height to 1 at current cursor position (excluding the Y offset)."}
	Button trResetYButton		,pos={13,205}	,size={90,22}	,title="Y Reset"								,help={"Resets Y offset and multiplier to 0."}
	Button trResetXButton		,pos={113,205}	,size={90,22}	,title="X Reset"								,help={"Resets X offset and multiplier to 0."}
	
	TitleBox trCsrControlTitle	,pos={13,240}	,size={190,13}	,title="Control Cursor Position"
	Button trPrevTraceButton	,pos={13,258}	,size={90,22}	,title="< Prev Trace"							,help={"Moves the cursor to onto the previous trace in the list of traces (page down key)."}
	Button trNextTraceButton	,pos={113,258}	,size={90,22}	,title="Next Trace >"							,help={"Moves the cursor to onto the next trace in the list of traces (page up key)."}
	Button trGoToMaxButton		,pos={13,288}	,size={90,22}	,title="Go to Max"								,help={"Sets the cursor to trace's maximum (h key)."}
	Button trGoToMinButton		,pos={113,288}	,size={90,22}	,title="Go to Min"								,help={"Sets the cursor to trace's minimum (l key)."}
	
	SetVariable trOffxDelta		,pos={58,325}	,size={100,16}	,title="dx:"					,bodyWidth=80	,help={"Negative values will flip trace in X direction."}
	SetVariable trOffyDelta		,pos={58,353}	,size={100,16}	,title="dy:"					,bodyWidth=80	,help={"Negative values will flip trace in Y direction."}
	Button trAddXButton			,pos={168,323}	,size={35,20}	,title="+"			,fstyle=1	,fsize=14		,help={"Add X delta value."}
	Button trRemXButton			,pos={13,323}	,size={35,20}	,title="-"			,fstyle=1	,fsize=14		,help={"Subtract X delta value."}
	Button trAddYButton			,pos={168,351}	,size={35,20}	,title="+"			,fstyle=1	,fsize=14		,help={"Add Y delta value."}
	Button trRemYButton			,pos={13,351}	,size={35,20}	,title="-"			,fstyle=1	,fsize=14		,help={"Subtract Y delta value."}

	Button trApplyCurButton		,pos={5,387}	,size={205,23}	,title="Apply to Current Data (overwrite!)"		,help={"Applies offsets and scales to the currently selected 1D data. Will overwrite the original waves, so use this with care!"}
	// ++++++++++++++++++++++++++++++++++++++++ settings +++++++++++++++++++++++++++++++++++++++++
	SetVariable trSetYOffset	,value=_NUM:0	,limits={-inf,inf,1}
	SetVariable trScaleYOffset	,value=_NUM:1	,limits={-inf,inf,kMultipIncrStep}
	SetVariable trShiftYOffset	,value=_NUM:0	,limits={-inf,inf,0}
	SetVariable trSetXOffset	,value=_NUM:0	,limits={-inf,inf,0.1}
	SetVariable trScaleXOffset	,value=_NUM:1	,limits={-inf,inf,kMultipIncrStep}
	SetVariable trShiftXOffset	,value=_NUM:0	,limits={-inf,inf,0}
	SetVariable trScaleNorm		,value=_NUM:0	,limits={-inf,inf,1}
	
	SetVariable trOffxDelta		,value=_NUM:0	,limits={-inf,inf,0}
	SetVariable trOffyDelta		,value=_NUM:0	,limits={-inf,inf,0}
	
	ModifyControlList ControlNameList(pSingle,";","tr*")		,win=$pSingle	,disable=2		// disable all cursor related controls for now
	ModifyControlList ControlNameList(pSingle,";","traceCsr*")	,win=$pSingle	,proc=TraceTool_CursorSelector	,mode = 1
	ModifyControlList ControlNameList(pSingle,";","*Button")	,win=$pSingle	,proc=TraceTool_ButtonFunctions
	ModifyControlList ControlNameList(pSingle,";","trS*") 		,win=$pSingle	,proc=TraceTool_SingleTraceVars	,format="%g" ,focusRing=0
	ModifyControlList ControlNameList(pSingle,";","*Title")		,win=$pSingle	,frame=0
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pSingle,";","*")		,win=$pSingle	,fsize=10
	endif
	SetActiveSubwindow ##
	
	int csrSet = 0;	string gName = GetTopGraph()
	if (strlen(gName))																			// check whether one of the cursors is set
		csrSet = strlen(CsrInfo(A,gName)+CsrInfo(B,gName)+CsrInfo(C,gName)+CsrInfo(D,gName)) > 0
		AutoPositionWindow/M=0/R=$GetTopMainGraph() TraceOffsetPanel
	endif
	TabControl offsetTabs value=csrSet															// if a cursor is set, display the single trace tab
	TraceTool_ToggleTabs(csrSet)
	FetchLatestTraceValues()																	// set variable values from top window
	return 0
end
//+++++++++++++++++++++++++++++++++++++ popup help functions +++++++++++++++++++++++++++++++++++++
function/S TraceTool_ProfileWaveList()
	return "none;"+WaveList("*",";", "DIMS:1")
end

//################################################################################################

static function GetAllTraceOffsets(string gName, string trace, variable &xOff, variable &yOff, variable &xMul, variable &yMul)
	sscanf StringByKey("offset(x)",   TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
	sscanf StringByKey("muloffset(x)",TraceInfo(gName,trace,0), "="), "{%f,%f}", xMul, yMul
	yMul = yMul == 0? 1 : yMul
	xMul = xMul == 0? 1 : xMul
end

//################################################################################################

static function GetTraceOffset(string gName, string trace, int axis)
	variable xOff = 0, yOff = 0
	sscanf StringByKey("offset(x)", TraceInfo(gName,trace,0), "="), "{%f,%f}", xOff, yOff
	return axis ? yOff : xOff
end

//################################ panel update helper function ##################################

function OffsetPanelUpdateFunc(STRUCT WMWinHookStruct &s)
	int HookTakeover = 0
	StrSwitch (s.EventName)
		case "activate":
			HookTakeover = 1
			if (!CmpStr(s.winName,"OffsetPanel"))												// in case an older panel is open -> rebuild
				KillWindow/Z OffsetPanel
				BuildTraceOffsetPanel()
			endif
			if (WinType("TraceOffsetPanel")==7)
				FetchLatestTraceValues()
			endif
		break
		case "keyboard":
			HookTakeover = TraceKeyboardEvents(GetTopGraph(), s.keycode)
		break
		case "kill":
			KillAllGraphHooks()
		break
	EndSwitch
	return HookTakeover
end

//################################################################################################

function TraceTool_GraphUpdateFunc(STRUCT WMWinHookStruct &s)									// is hooked to target graphs
	StrSwitch (s.EventName)
		case "keyboard":
			return TraceKeyboardEvents(s.winName, s.keycode)
		break
		case "activate":
			FetchLatestTraceValues()															// graph might have been switched
		break
	EndSwitch
	return 0
end

//------------------------------------------------------------------------------------------------

static function KillAllGraphHooks()
	string list = WinList("*",";","WIN:1"); int i
	for (i=0; i<ItemsInList(list); i++)
		SetWindow $StringFromList(i,list,";") hook(GT_TraceUpdateHook)=$""
	endfor
end

//################################################################################################

static function TraceKeyboardEvents(string gName, variable code)
	if (WinType("TraceOffsetPanel")!=7)
		return 0
	endif
	ControlInfo/W=TraceOffsetPanel offsetTabs
	int tab = V_flag == 8 ? V_Value : 0
	
	Switch(code)
		case 11:	// page up
			return TraceButtonEvent(gName, "trNextTraceButton")
		break
		case 12:	// page down
			return TraceButtonEvent(gName, "trPrevTraceButton")
		break
		case 104:	// h
			return TraceButtonEvent(gName, "trGoToMaxButton")
		break
		case 108:	// l
			return TraceButtonEvent(gName, "trGoToMinButton")
		break
	EndSwitch
	
	if (code > 96 && code < 101 && strlen(gName))	// a-d keys
		string doCsr = StringFromList(code-97, "A;B;C;D;")
		if (strlen(CsrInfo($doCsr,gName)))
			return TraceSwitchCursors("traceCsr"+doCsr)
		endif
	endif
	
	if ((code > 27 && code < 32) && ((GetKeyState(0)&2^1) == 0))								// left, right, up, down (alt key passes movement to cursor event)
		if (tab == 0)
			return Keypad_ShiftScaleMulti(gName, code)
		elseif (tab == 1)
			return Keypad_ShiftScaleSingle(gName, code)
		endif
	endif
	return 0
end

//################################################################################################

static function TraceTool_ToggleTabs(int whichTab)
	SetWindow TraceOffsetPanel#MultiTab  hide=whichTab==1
	SetWindow TraceOffsetPanel#SingleTab hide=whichTab==0
	ModifyControl CsrActButton ,win=TraceOffsetPanel#SingleTab ,activate						// move focus away from tab control to prevent interference with cursor controls
end

function TraceTool_TabControl(STRUCT WMTabControlAction &s) : TabControl
	if (s.eventCode == 2)
		TraceTool_ToggleTabs(s.tab)
	endif
end

//################################################################################################

static function Keypad_ShiftScaleSingle(string gName, int key)									// keyboard control of single trace
	if (WinType(gName)!=1 || WinType("TraceOffsetPanel")!=7)
		return 0
	endif
	string csr = GetUserData("TraceOffsetPanel", "", "cursor")									// get currently used cursor from panel user data
	if (strlen(csr) == 0)
		return 0
	endif
	string trace = StringByKey("TNAME", CsrInfo($csr,gName))
	if (strlen(trace) == 0)
		return 0
	endif
	
	string xUnit = "", yUnit = ""
	variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, xRange = 0, yRange = 0
	GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
	GetAxisRangeFromTrace(xRange, yRange, xUnit, yUnit, gName, trace)
	
	variable neg = 1, ctrlKey = (GetKeyState(0) & 2^0)!=0
	Switch(key)
		case 28:	// left
			neg = -1
		case 29:	// right
			if (ctrlKey)
				ModifyGraph/W=$gName muloffset($trace)={calcMultiplier(xMul, neg),}
			else
				ModifyGraph/W=$gName offset($trace)={calcShift(xOff, neg, xRange),}
			endif
		break
		case 31:	// down
			neg = -1
		case 30:	// up
			if (ctrlKey)
				ModifyGraph/W=$gName muloffset($trace)={,calcMultiplier(yMul, neg)}
			else
				ModifyGraph/W=$gName offset($trace)={,calcShift(yOff, neg, yRange)}
			endif
		break
	EndSwitch
	DoUpdate/W=$gName/SPIN=10																	// update the graph but don't seem busy while doing it
	FetchLatestCursorValues(csr)
	AdjustProfileAxisScaling()
	return 1
end

//------------------------------------------------------------------------------------------------

static function calcMultiplier(variable scale, variable neg)
	variable shift = (GetKeyState(0) & 2^2)!=0, delta, output
	delta  = kMultipIncrStep*10^GetValuesOrderOfMagnitude(scale/1.5)/(10-9*shift)
	delta *= abs(scale*neg-delta)<10e-8*abs(scale*neg) ? 2 : 1									// don't scale to exactly zero (which will flip back to 1)
	output = scale+neg*delta
	return ((abs(output-1) < 10e-8) ? 0 : output)												// 1 is transformed to zero
end

static function calcShift(variable offset, variable neg, variable delta)
	variable shift  = (GetKeyState(0) & 2^2)!=0
	variable output = offset+neg*kOffsetIncrStep*delta*(1+9*shift)
	return ((abs(output)/abs(delta) < 10e-8) ? 0 : output)										// round to zero if value is very small
end

//################################################################################################

static function Keypad_ShiftScaleMulti(string gName, int key)									// keyboard control of multiple traces
	string pMulti= "TraceOffsetPanel#MultiTab", traces = ""
	if (WinType(gName)!=1 || WinType("TraceOffsetPanel")!=7)
		return 0
	endif
	if (FetchTraces(gName,traces) == 0)
		return 0
	endif
	
	variable xRange, yRange, newVal; string xUnit = "", yUnit = ""
	GetAxisRangeFromTraceList(xRange, yRange, xUnit, yUnit, gName, traces)
	
	variable neg = 1, ctrlKey = (GetKeyState(0) & 2^0)!=0
	Switch(key)
		case 28:	// left
			neg = -1
		case 29:	// right
			newVal = neg*xRange*kOffsetIncrStep
			if (ctrlKey)
				ControlInfo/W=$pMulti vaSetXOffset
				SetVariable vaSetXOffset   ,win=$pMulti, value=_NUM:V_Value+newVal ,userdata=ExecuteMultiTraceEdits("vaSetXOffset",   S_UserData, V_Value+newVal)
			else
				ControlInfo/W=$pMulti vaShiftXOffset
				SetVariable vaShiftXOffset ,win=$pMulti, value=_NUM:V_Value+newVal ,userdata=ExecuteMultiTraceEdits("vaShiftXOffset", S_UserData, V_Value+newVal)
			endif
		break
		case 31:	// down
			neg = -1
		case 30:	// up
			newVal = neg*yRange*kOffsetIncrStep
			if (ctrlKey)
				ControlInfo/W=$pMulti vaSetYOffset
				SetVariable vaSetYOffset   ,win=$pMulti, value=_NUM:V_Value+newVal ,userdata=ExecuteMultiTraceEdits("vaSetYOffset",   S_UserData, V_Value+newVal)
			else
				ControlInfo/W=$pMulti vaShiftYOffset
				SetVariable vaShiftYOffset ,win=$pMulti, value=_NUM:V_Value+newVal ,userdata=ExecuteMultiTraceEdits("vaShiftYOffset", S_UserData, V_Value+newVal)
			endif
		break
	EndSwitch
	return 1
end

//################################################################################################

function TraceTool_SwitchOffsetMode(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		CheckBox $StringFromList(0,RemoveFromList(s.ctrlName, ControlNameList("TraceOffsetPanel#MultiTab",";","Choose*"))) ,win=TraceOffsetPanel#MultiTab ,value=0		
	endif
	return 0
end

//################################################################################################

function TraceTool_CursorSelector(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		TraceSwitchCursors(s.ctrlName)
	endif
	return 0
end

//################################################################################################

static function TraceSwitchCursors(string which)
	int i;	string ctrl = RemoveFromList(which, ControlNameList("TraceOffsetPanel#SingleTab",";","traceCsr*"))
	for (i = 0; i < ItemsInList(ctrl); i++)
		CheckBox $StringFromList(i,ctrl) ,win=TraceOffsetPanel#SingleTab ,value=0
	endfor
	CheckBox $which ,win=TraceOffsetPanel#SingleTab ,value=1
	FetchLatestCursorValues(ReplaceString("traceCsr", which, ""))
	return 1
end

//################################################################################################

static function FetchLatestTraceValues()														// update both left and right side values (if cursors are set)
	string gName = GetTopGraph()
	string pSingle = "TraceOffsetPanel#SingleTab"
	string pMulti  = "TraceOffsetPanel#MultiTab"
	
	ModifyControlList/Z ControlNameList(pSingle,";","tr*")+"trApplyCurButton;",win=$pSingle ,disable=2	// disable cursor related controls for now
	if (!strlen(gName))
		TitleBox MultiTargetTitle  ,win=$pMulti  ,title="No Graph Available!" ,fColor=(65535,0,0)
		TitleBox SingleTargetTitle ,win=$pSingle ,title="No Graph Available!" ,fColor=(65535,0,0)
		return 0
	endif
	string traces = ""
	variable items = FetchTraces(gName,traces)
	if (items == 0)
		TitleBox MultiTargetTitle  ,win=$pMulti  ,title="No Traces Available!" ,fColor=(65535,0,0)
		TitleBox SingleTargetTitle ,win=$pSingle ,title="No Traces Available!" ,fColor=(65535,0,0)
		return -1
	endif
	
	variable xOff = 0, yOff = 0, xShift = 0,  yShift = 0
	sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-1,traces),0), "="), "{%f,%f}", xShift, yShift
	if (items > 1)																				// calculate shift and offset from first two traces
		sscanf StringByKey("offset(x)", TraceInfo(gName,StringFromList(items-2,traces),0), "="), "{%f,%f}", xOff, yOff
	else
		xOff = xShift; yOff = yShift
	endif
	
	variable xRange, yRange; string xUnit = "", yUnit = ""
	GetAxisRangeFromTraceList(xRange, yRange, xUnit, yUnit, gName, traces)
	
	SetWindow $gName hook(GT_TraceUpdateHook)=TraceTool_GraphUpdateFunc							// set graph hook link
	
	TitleBox MultiTargetTitle 	,win=$pMulti	,title="\f01Modify:\f00 "+gName ,fColor=(0,0,0)
	SetVariable vaSetYOffset	,win=$pMulti	,value=_NUM:yOff-yShift	,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep}	,userdata=num2str(yOff-yShift)
	SetVariable vaShiftYOffset	,win=$pMulti	,value=_NUM:yShift		,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep}
	SetVariable vaSetXOffset	,win=$pMulti	,value=_NUM:xOff-xShift	,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep}	,userdata=num2str(xOff-xShift)
	SetVariable vaShiftXOffset	,win=$pMulti	,value=_NUM:xShift		,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep*10}
	PopupMenu ProfileSelect 	,win=$pMulti 	,mode=1											// update the selected profile or select none
	if (FindListItem("Profileleft", AxisList(gName)) >= 0)
		wave/Z profile = TraceNameToWaveRef(gName, "AddTraceProfile")
		if (WaveExists(profile))
			PopupMenu ProfileSelect	,win=$pMulti	,popmatch=NameOfWave(profile)
		endif
	endif
	
	wave data = TraceNameToWaveRef(gName, StringFromList(0,traces))
	if (!WaveExists(data))																		// probably image data => not useful here
		return 0
	endif
	Button vaApplyAllButton 	,win=$pMulti 	,disable=2*(WaveDims(data) != 1)				// this button only works for 1D traces
	
	//------------------------------------- cursor controls --------------------------------------
	string csr = "A;B;C;D;", currCsr															// fetch available cursor list
	variable i, selCsr = -1,  setCsr = -1, csrActive = 0
	for (i = 0; i < ItemsInList(csr); i++)														// find currently selected cursor
		currCsr = "traceCsr"+StringFromList(i,csr)
		ControlInfo/W=$pSingle $currCsr
		if (V_Value)
			CheckBox $currCsr ,win=$pSingle ,value=0											// deselect for now
			selCsr = i
			break
		endif
	endfor

	for (i = 0; i < ItemsInList(csr); i++)
		currCsr = StringFromList(i,csr)
		if (strlen(CsrInfo($currCsr,gName)))
			if (numtype(zcsr($currCsr,gName)) != 2)												// check for cursors on image plots
				continue
			endif
			if (csrActive == 0 && selCsr == i)													// the currently selected cursor is active and can stay
				csrActive = 1
				setCsr = i
			endif
			if (setCsr == -1)																	// find the first free cursor from left
				setCsr = i
			endif
			CheckBox $("traceCsr"+currCsr) ,win=$pSingle ,disable=0
		endif
	endfor
	
	if (setCsr == -1)
		SetWindow TraceOffsetPanel	,userdata(cursor)=""										// clear cursor user data
		if (selCsr != -1)
			CheckBox $("traceCsr"+StringFromList(selCsr,csr)) ,win=$pSingle ,value=1			// set previous cursor (disabled but keeps cursor selected)
		endif
		TitleBox SingleTargetTitle	,win=$pSingle ,title="No Trace Selected!" ,fColor=(65535,0,0)
		return 0
	endif
	//---------------------------------- if cursor was found -------------------------------------
	currCsr = StringFromList(setCsr,csr)
	CheckBox $("traceCsr"+currCsr) ,win=$pSingle ,value=1
	string ctrlList = RemoveFromList("traceCsrA;traceCsrB;traceCsrC;traceCsrD;",ControlNameList(pSingle,";","tr*"))+"trApplyCurButton;"
	ModifyControlList ctrlList	,win=$pSingle	,disable=0
	FetchLatestCursorValues(currCsr)															// now get cursor specific values (right side)
	return 0
end

//################################################################################################

static function FetchLatestCursorValues(string csr [,int keepDelta])
	string gName = GetTopGraph()
	string pSingle = "TraceOffsetPanel#SingleTab"
	if (!strlen(gName) || WinType("TraceOffsetPanel")!=7)
		return 0
	endif
	
	int updateOffsets = ParamIsDefault(keepDelta) ? 1 : !keepDelta
	
	string trace = StringByKey("TNAME", CsrInfo($csr,gName))									// get the info from the trace-associated axis
	string color = StringByKey("rgb(x)",TraceInfo(gName,trace,0), "=")							// show the color of the selected trace
	variable xRange, yRange; string xUnit = "", yUnit = ""
	GetAxisRangeFromTrace(xRange, yRange, xUnit, yUnit, gName, trace)
	
	variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, shift
	GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
	variable scaleNorm = vcsr($csr, gName)*yMul
	variable xShiftCsr = round( (hcsr($csr, gName)*xMul+xOff) *10^6)/10^6						// round to prevent floating point errors
	variable yShiftCsr = round( (vcsr($csr, gName)*yMul+yOff) *10^6)/10^6
	variable yScOrder  = GetValuesOrderOfMagnitude(yMul/1.5)
	variable xScOrder  = GetValuesOrderOfMagnitude(xMul/1.5) 
	
	TitleBox SingleTargetTitle	,win=$pSingle	,title="\f01Trace: \K"+color+ReplaceString("'",trace,"")+"\K(0,0,0)\f00",fColor=(0,0,0)
	Button trApplyCurButton 	,win=$pSingle	,disable=2*(WaveDims(TraceNameToWaveRef(gName,trace))!=1)		// this button only works for 1D traces
	SetVariable trSetYOffset	,win=$pSingle	,value=_NUM:yOff		,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep}
	SetVariable trShiftYOffset	,win=$pSingle	,value=_NUM:yShiftCsr	,format="%g"+yUnit	,limits={-inf,inf,yRange*kOffsetIncrStep*10}
	SetVariable trScaleYOffset	,win=$pSingle	,value=_NUM:yMul		,format="%g"		,limits={-inf,inf,kMultipIncrStep*10^yScOrder}
	SetVariable trSetXOffset	,win=$pSingle	,value=_NUM:xOff		,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep}
	SetVariable trShiftXOffset	,win=$pSingle	,value=_NUM:xShiftCsr	,format="%g"+xUnit	,limits={-inf,inf,xRange*kOffsetIncrStep*10}
	SetVariable trScaleXOffset	,win=$pSingle	,value=_NUM:xMul		,format="%g"		,limits={-inf,inf,kMultipIncrStep*10^xScOrder}
	SetVariable trScaleNorm		,win=$pSingle	,value=_NUM:scaleNorm	,format="%g"+yUnit
	
	SetVariable trOffxDelta		,win=$pSingle							,format="%g"			// first reset unit
	SetVariable trOffyDelta		,win=$pSingle							,format="%g"
	
	string unit = ""
	shift = TraceTool_FetchXShift(TraceNameToWaveRef(gName,trace),unit)
if (!numtype(shift) && updateOffsets)
	SetVariable trOffxDelta		,win=$pSingle	,value=_NUM:shift		,format="%g"+SelectString(strlen(unit),""," "+unit)
endif
	unit = ""
	shift = TraceTool_FetchYShift(TraceNameToWaveRef(gName,trace),unit)
if (!numtype(shift) && updateOffsets)
	SetVariable trOffyDelta		,win=$pSingle	,value=_NUM:shift		,format="%g"+SelectString(strlen(unit),""," "+unit)
endif
	
	SetWindow TraceOffsetPanel, userdata(trace)=trace
	SetWindow TraceOffsetPanel, userdata(cursor)=csr
	return 0
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S GetMajorAxisFromTraceList(string gName, string traces, string which)
	int i,j,items = ItemsInList(traces)
	Make/Free/T/N=(items) Axes = StringByKey(UpperStr(which)+"AXIS", TraceInfo(gName,StringFromList(p,traces),0))
	string axis = Axes[0]
	if (DimSize(Axes,0) > 1)
		FindDuplicates/Free/RT=Axes_clean/DT=Axes_dup Axes
		if (DimSize(Axes_clean,0) > 1)															// need to find the major axis
			Make/Free/I/N=(DimSize(Axes_clean,0)) tempCount
			for (i = 0; i < DimSize(Axes_clean,0); i++)
				for (j = 0; j < DimSize(Axes_dup,0); j++)
					tempCount[i] += StringMatch(Axes_clean[i],Axes_dup[j])						// count the number of duplicates
				endfor
			endfor
			WaveStats/Q tempCount
			return Axes_clean[V_maxRowLoc]														// the maximum amount of occurrences will be the main axis
		else
			return Axes_clean[0]
		endif
	else
		return Axes[0]
	endif
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function GetAxisRangeFromTrace(variable &xRange, variable &yRange, string &xUnit, string &yUnit, string gName, string trace)
	GetAxisRange(xRange, yRange, xUnit, yUnit, gName, StringByKey("XAXIS", TraceInfo(gName,trace,0)), StringByKey("YAXIS", TraceInfo(gName,trace,0)))
end

static function GetAxisRangeFromTraceList(variable &xRange, variable &yRange, string &xUnit, string &yUnit, string gName, string traces)
	GetAxisRange(xRange, yRange, xUnit, yUnit, gName, GetMajorAxisFromTraceList(gName, traces, "X"), GetMajorAxisFromTraceList(gName, traces, "Y"))
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function GetAxisRange(variable &xRange, variable &yRange, string &xUnit, string &yUnit, string gName, string xAxis, string yAxis)
	string info
	GetAxis/W=$gName/Q $xAxis
	xRange = abs(V_max - V_min)
	xUnit  = StringByKey("UNITS", AxisInfo(gName,xAxis))
	xUnit  = SelectString(strlen(xUnit),""," "+xUnit)
	GetAxis/W=$gName/Q $yAxis
	yRange = abs(V_max - V_min)
	yUnit  = StringByKey("UNITS", AxisInfo(gName,yAxis))
	yUnit  = SelectString(strlen(yUnit),""," "+yUnit)
	sprintf info, "%.*g\r", 1, yRange;	yRange = str2num(info)									// rounding
	sprintf info, "%.*g\r", 1, xRange;	xRange = str2num(info)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function GetValuesOrderOfMagnitude(variable val)
	string numChar = "", eChar = ""; sprintf numChar, "%.1E", val
	SplitString/E=("[0-9]+.[0-9]([eE][-+]?[0-9]+)?") numChar, eChar
	variable OrderOfMagnitude = str2num(eChar[1,strlen(eChar)])
	return numtype(OrderOfMagnitude)!=0 ? 0 : OrderOfMagnitude
end

//################################## variable helper function ####################################

function TraceTool_SingleTraceVars(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		s.blockReentry=1
		return ExecuteSingleTraceEdits(s.ctrlName, s.dval)
	endif
end
//------------------------------------------------------------------------------------------------
static function ExecuteSingleTraceEdits(string ctrlName, variable dval)
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string csr = GetUserData("TraceOffsetPanel", "", "cursor")									// get currently used cursor from panel user data
	if (strlen(csr) == 0)
		return 0
	endif
	string trace = StringByKey("TNAME", CsrInfo($csr,gName))
	if (strlen(trace) == 0)
		return 0
	endif
	
	string xUnit = "", yUnit = ""
	variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, xRange = 0, yRange = 0
	GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
	GetAxisRangeFromTrace(xRange, yRange, xUnit, yUnit, gName, trace)
	
	StrSwitch(ctrlName)
		case "trSetYOffset":
			dval = abs(dval)/abs(yRange)<10e-8 ? 0 : dval	
			ModifyGraph/W=$gName offset($trace)={,dval}
		break
		case "trSetXOffset":
			dval = abs(dval)/abs(xRange)<10e-8 ? 0 : dval
			ModifyGraph/W=$gName offset($trace)={dval,}
		break
		case "trShiftYOffset":
			ModifyGraph/W=$gName offset($trace)={,(dval-vcsr($csr, gName)*yMul)}
		break
		case "trShiftXOffset":
			ModifyGraph/W=$gName offset($trace)={(dval-hcsr($csr, gName)*xMul),}
		break
		case "trScaleYOffset":
			dval = abs(dval)<10e-8 ? sign(dval)*10e-8 : dval									// prevent input from getting very small
			dval = dval==1 ? 0 : dval
			ModifyGraph/W=$gName muloffset($trace)={,dval}
		break
		case "trScaleXOffset":
			dval = abs(dval)<10e-8 ? sign(dval)*10e-8 : dval
			dval = dval==1 ? 0 : dval
			ModifyGraph/W=$gName muloffset($trace)={dval,}
		break
		case "trScaleNorm":
			if (vcsr($csr, gName) != 0)
				ModifyGraph/W=$gName muloffset($trace)={,dval/vcsr($csr, gName)}
			endif
		break
	EndSwitch	
	FetchLatestCursorValues(csr)
	AdjustProfileAxisScaling()
	return 0
end

//################################################################################################

function TraceTool_MultiTraceVars(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		s.blockReentry = 1
		s.userdata = ExecuteMultiTraceEdits(s.ctrlName, s.userdata, s.dval)
	endif
	return 0	
end	
// -----------------------------------------------------------------------------------------------
static function/S ExecuteMultiTraceEdits(string ctrlName, string userData, variable dval)
	string gName = GetTopGraph(), traces = "", returnStr = userdata
	if (!strlen(gName))
		return returnStr
	endif
	variable i, items = FetchTraces(gName,traces)
	if (items == 0)
		return returnStr
	endif
	
	string trace = StringFromList(items-1,traces)
	variable YnotX = StringMatch(ctrlName,"*YOffset")
	variable prevVal, curXOff = 0, curYOff = 0, xOff = 0, yOff = 0, xMul = 0, yMul = 0
	GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
	
	StrSwitch(ctrlName)
		case "vaSetYOffset":
		case "vaSetXOffset":
			ControlInfo/W=TraceOffsetPanel#MultiTab ChooseAdd;	int addMode = v_Value
			prevVal = str2num(userdata);	returnStr = num2str(dval)							// get previous offset from and write current value to control user data
			for (i = 0; i < items; i++)
				trace = StringFromList(items-1 - i,traces)
				if (addMode)
					GetAllTraceOffsets(gName, trace, curXOff, curYOff, xMul, yMul)
					if (YnotX)
						ModifyGraph/W=$gName offset($trace)={,i*(dval-prevVal)+curYOff}
					else
						ModifyGraph/W=$gName offset($trace)={ i*(dval-prevVal)+curXOff,}
					endif
				else
					if (YnotX)
						ModifyGraph/W=$gName offset($trace)={,i*dval+yOff}
					else
						ModifyGraph/W=$gName offset($trace)={ i*dval+xOff,}
					endif
				endif
			endfor
		break
		case "vaScaleYOffset":
		case "vaScaleXOffset":
			prevVal = str2num(userdata);	returnStr = num2str(dval)							// get previous multiplier from and write current value to control user data
			prevVal = prevVal != 0 ? prevVal : 1												// make sure multiplier is not zero
			if ((GetKeyState(0)&2^0) != 0)														// ctrl / cmd pressed = absolute mode
				yOff = 0; xOff = 0
			endif
			for (i = 0; i < items; i++)
				trace = StringFromList(items-1 - i,traces)
				GetAllTraceOffsets(gName, trace, curXOff, curYOff, xMul, yMul)
				if (YnotX)
					ModifyGraph/W=$gName offset($trace)={,(curYOff-yOff)*dval/prevVal+yOff}
				else
					ModifyGraph/W=$gName offset($trace)={ (curXOff-xOff)*dval/prevVal+xOff,}
				endif
			endfor
		break
		case "vaShiftYOffset":
		case "vaShiftXOffset":
			for (i = 0; i < items; i++)
				trace = StringFromList(items-1 - i,traces)
				GetAllTraceOffsets(gName, trace, curXOff, curYOff, xMul, yMul)
				if (YnotX)
					ModifyGraph/W=$gName offset($trace)={,(curYOff-yOff)+dval}
				else
					ModifyGraph/W=$gName offset($trace)={ (curXOff-xOff)+dval,}
				endif
			endfor
		break
	EndSwitch
	
	string csr = GetUserData("TraceOffsetPanel", "", "cursor")
	if (strlen(csr))
		FetchLatestCursorValues(csr)
	endif
	AdjustProfileAxisScaling()
	return returnStr
end

// ###############################################################################################

function TraceTool_TraceFilterProc(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventMod & 16 && s.eventCode == 9)													// add right-click menu to directly insert values
		string gName = GetTopGraph()
		string traces = "(No traces found;", currStr
		
		ControlInfo/W=TraceOffsetPanel#MultiTab cHiddenTraces;	int noHidden = V_flag ? !V_Value : 0
		if (strlen(gName))
			traces = "(Add trace:;" + TraceNameList(gName,";",1+4*noHidden)
		endif
		
		PopupContextualMenu traces
		if (V_flag > 1)
			ControlInfo/W=TraceOffsetPanel#MultiTab vTraceFilter
			currStr = SelectString(strlen(S_Value),"",RemoveEnding(S_Value,";")+";")
			currStr = SelectString(CmpStr(currStr,"*;"),"",currStr)
			SetVariable vTraceFilter ,win=TraceOffsetPanel#MultiTab ,value=_STR:currStr+S_Selection+";"
			ControlUpdate/W=TraceOffsetPanel#MultiTab vTraceFilter
		endif
		return 1
	endif
	return 0
end

//++++++++++++++++++++++++++++++ help function to get a trace list ++++++++++++++++++++++++++++++

static function FetchTraces(string win, string &traces)
	int noHidden = 0, omit = 0
	string filter = "", tempList = ""
	if (WinType("TraceOffsetPanel"))															// read settings from panel if open
		ControlInfo/W=TraceOffsetPanel#MultiTab cHiddenTraces
		if (V_flag)
			noHidden = !V_Value
		endif
		ControlInfo/W=TraceOffsetPanel#MultiTab vTraceFilter
		if (V_flag)
			filter = S_Value
		endif
		ControlInfo/W=TraceOffsetPanel#MultiTab cOmitFilter
		if (V_flag)
			omit = V_Value
		endif
	endif
	
	traces = TraceNameList(win,";",1+4*noHidden)
	int i, items = ItemsInList(filter)
	if (strlen(filter) && CmpStr(filter,"*"))													// filter each list item ('*' just means 'include all')
		if (omit)
			for (i = 0; i < items; i++)
				traces = RemoveFromList(ListMatch(traces,StringFromList(i,filter)),traces)
			endfor
		else
			for (i = 0; i < items; i++)
				tempList += ListMatch(traces,StringFromList(i,filter))
			endfor
			traces = tempList
		endif
	endif
	
	traces = RemoveProfileTraceFromList(win, traces)
	traces = RemoveFromListWildCard(ksTraceIgnoreList, traces)
	return ItemsInList(traces)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function FetchAllTraces(string win, string &traces)
	traces = RemoveProfileTraceFromList(win, TraceNameList(win,";",1))
	traces = RemoveFromListWildCard(ksTraceIgnoreList, traces)
	return ItemsInList(traces)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S RemoveProfileTraceFromList(string win, string traces)
	int i, items = ItemsInList(traces)
	string remItem = ""
	for (i = 0; i < items; i++)
		string xAxis = StringByKey("XAXIS",  TraceInfo(win,StringFromList(items-1 - i,traces),0))
		if (StringMatch(xAxis,"Profilebottom") || StringMatch(xAxis,"Profileleft"))
			remItem += StringFromList(items-1 - i,traces)+";"
		endif
	endfor
	return RemoveFromList(remItem,traces)
end

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static function/S RemoveFromListWildCard(string zapListStr, string listStr)						// returns listStr, purged of any items that match an item in ZapListStr.
	string removeStr=""
	int i
	for(i = 0; i < ItemsInList(zapListStr); i++)
		removeStr += ListMatch(listStr, StringFromList(i, zapListStr))							// Wildcards okay! Case insensitive.
	endfor
	return RemoveFromList(removeStr, listStr, ";", 0)
end

//+++++++++++++++++++++++++++++++ help function for Profile Axis +++++++++++++++++++++++++++++++++

static function AdjustProfileAxisScaling()
	string gName = GetTopGraph()
	if (!strlen(gName) || WhichListItem("Profileleft", AxisList(gName)) < 0)
		return 0
	endif

	string traces = ""
	int i, items = FetchAllTraces(gName,traces)
	if (items == 0)
		return 0
	endif
	variable currMax = 0, dataMin = 0, yOff
	for (i = 0; i < items; i++)
		yOff = GetTraceOffset(gName, StringFromList(i,traces), 1)
		if (yOff > currMax)
			currMax = yOff
			dataMin = WaveMin(TraceNameToWaveRef(gName,StringFromList(i,traces)))
		endif
	endfor
	dataMin = numtype(dataMin)!=0 ? 0 : dataMin

	DoUpdate; GetAxis/W=$gName/Q left
	variable scale = (currMax+dataMin-V_min)/(V_Max-V_min)										// calculate the ratio between highest offset and axis height
	ModifyGraph/W=$gName axisEnab(Profileleft)={0,limit(scale, 0.1, 1)}							// prevent scaling to useless values
	return 0
end

//################################### button helper function #####################################

function TraceTool_ButtonFunctions(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		s.blockReentry=1
		TraceButtonEvent(GetTopGraph(), s.ctrlName)
	endif
	return 0
end

//################################################################################################

static function TraceButtonEvent(string gName, string which)
	if (WinType(gName)!=1 || WinType("TraceOffsetPanel")!=7)
		return 0
	endif
	string pSingle = "TraceOffsetPanel#SingleTab"
	string pMulti  = "TraceOffsetPanel#MultiTab"
	string csr = GetUserData("TraceOffsetPanel", "", "cursor"), trace = ""						// get currently used cursor from panel user data
	if (strlen(csr))
		trace = StringByKey("TNAME", CsrInfo($csr,gName))
	endif
	
	variable xOff = 0, yOff = 0, xMul = 0, yMul = 0, setX, setY, cursorPos
	if (strlen(trace) && strlen(csr))
		GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
	endif
	
	string traces = "", axis = "X"
	int i, GoUp = 0, items = FetchAllTraces(gName,traces), retVal = 0
	StrSwitch(which)
		case "CsrActButton":
			if (items == 0)
				break
			endif
			for (i = 0; i < 4; i++)
				string currCsr = StringFromList(i,"A;B;C;D;")
				if (!strlen(CsrInfo($currCsr,gName)))
					trace = StringFromList(0,traces)
					GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
					wave w = TraceNameToWaveRef(gName,trace)
					wave/Z w_x = XWaveRefFromTrace(gName, trace)
					
					GetAxis/W=$gName/Q $StringByKey("XAXIS",TraceInfo(gName,trace,0))
					setX = (V_min+(V_max-V_min)/2 - xOff)/xMul									// center of the current axis corrected for scaling and offset of the current trace
					cursorPos = x2pnt(w,setX)
					if(WaveExists(w_x))
						FindLevel/P/Q w_x, setX;	cursorPos = !v_flag ? V_LevelX : DimSize(w,0)/2
					endif
					cursorPos = limit(cursorPos,0,DimSize(w,0)-1)
					Cursor/W=$gName/P $currCsr, $trace, cursorPos								// set a new cursor in the center of the current axis scale if possible
					
					if (strlen(csr))															// if a cursor was already active, then switch over to the newly placed one
						CheckBox $("traceCsr"+csr)		,win=$pSingle ,value=0
						CheckBox $("traceCsr"+currCsr)	,win=$pSingle ,value=1
					endif
					FetchLatestTraceValues()
					break																		// cursor set => quit
				endif
			endfor
		break
		case "ColorButton":
			ControlInfo/W=$pMulti ColorSelect;		string Table = S_Value
			ControlInfo/W=$pMulti cReverseColor;	variable rev = (GetKeyState(0) & 2^0)!=0
			ColorSpectrumFromTable((V_Value || rev), table)
		break
		case "ReverseButton":
			if (items == 0)
				break
			endif
			string trList = ""
			for (i = 0; i < items; i++)															// reverse trace list
				trList += StringFromList(items-1-i,traces) + ";"
			endfor
			wave/T w_traces = ListToTextWave(traces,";")
			Make/Free/N=(items) xOffsets = GetTraceOffset(gName, w_traces[p], 0)
			Make/Free/N=(items) yOffsets = GetTraceOffset(gName, w_traces[p], 1)
			
			int do2Dsort = 0;	string instanceList = GrepList(traces,"#")
			if (ItemsInList(instanceList))
				do2Dsort = (WaveDims(TraceNameToWaveRef(gName, StringFromList(0, instanceList))) == 2)	// special treatment of 2D waterfall plots
			endif
			
			if (items <= 100)																	// this even works for instances (but is limited to 100 traces)
				Execute/Q/Z "ReorderTraces/W="+gName+" _front_,{"+RemoveEnding(ReplaceString(";",trList,","),",")+"}"
			else
				if (do2Dsort)
					Abort "Error: Cannot reverse 2D data larger that 100 traces."
				endif
				for(i = 0; i < items; i++)
					ReorderTraces/W=$gName _back_, {$StringFromList(i,traces)}
				endfor
			endif
			
			for (i = 0; i < items; i++)															// reapply offsets (in reverse)
				if (do2Dsort)																	// the trace labeling will stay the same)
					ModifyGraph/W=$gName offset($w_traces[items-1-i])={xOffsets[i],yOffsets[i]}
				else
					ModifyGraph/W=$gName offset($w_traces[i])={xOffsets[i],yOffsets[i]}
				endif
			endfor
			FetchLatestTraceValues()
		break
		case "ResetYButton":
			axis = "Y"
		case "ResetXButton":
			items = FetchTraces(gName,traces)
			for (i = 0; i < items; i++)
				if (CmpStr(axis,"Y") == 0)
					ModifyGraph/W=$gName offset($StringFromList(i,traces))={,0}
				else
					ModifyGraph/W=$gName offset($StringFromList(i,traces))={0,}
				endif
			endfor
			SetVariable $("vaSet"+axis+"Offset")	,win=$pMulti	,value=_NUM:0	,userdata="0"
			SetVariable $("vaScale"+axis+"Offset")	,win=$pMulti	,value=_NUM:1	,userdata="1"
			SetVariable $("vaShift"+axis+"Offset")	,win=$pMulti	,value=_NUM:0
			AdjustProfileAxisScaling()
		break
		case "trResetYButton":
			if (strlen(trace) && strlen(csr))
				ModifyGraph/W=$gName offset($trace)={,0} ,muloffset($trace)={,0}
				SetVariable trSetYOffset	,win=$pSingle	,value=_NUM:0
				SetVariable trScaleYOffset	,win=$pSingle	,value=_NUM:1
				SetVariable trShiftYOffset	,win=$pSingle	,value=_NUM:vcsr($csr, gName)
				SetVariable trScaleNorm		,win=$pSingle	,value=_NUM:vcsr($csr, gName)
				AdjustProfileAxisScaling()
			endif
		break
		case "trResetXButton":
			if (strlen(trace) && strlen(csr))
				ModifyGraph/W=$gName offset($trace)={0,} ,muloffset($trace)={0,}
				SetVariable trSetXOffset	,win=$pSingle	,value=_NUM:0
				SetVariable trScaleXOffset	,win=$pSingle	,value=_NUM:1
				SetVariable trShiftXOffset	,win=$pSingle	,value=_NUM:hcsr($csr, gName)
				AdjustProfileAxisScaling()
			endif
		break
		case "trGoToMaxButton":
			GoUp = 1
		case "trGoToMinButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			retVal = 1
			
			wave w = CsrWaveRef($csr,gName)
			WaveStats/Q/P w
			setX = GoUp ? V_maxRowLoc : V_minRowLoc
			setY = GoUp ? V_maxColLoc : V_minColLoc
			if (strlen(ImageInfo(gName,trace,0)))
				Cursor/P/W=$gName/I $csr, $trace, setX, setY									// move cursor to image position
			else
				Cursor/P/W=$gName	$csr, $trace, setX											// move cursor to trace position
			endif
			FetchLatestCursorValues(csr)
		break
		case "trNextTraceButton":
			GoUp = 1
		case "trPrevTraceButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			retVal = 1
			
			int currTrace = WhichListItem(trace, traces)
			string nextTrace = ""
			if (GoUp)
				if (currTrace < items-1)
					nextTrace = StringFromList(currTrace+1,traces)
				endif
			else
				if (currTrace > 0)
					nextTrace = StringFromList(currTrace-1,traces)
				endif
			endif
			if (strlen(ImageInfo(gName,trace,0)) || !strlen(nextTrace))							// already at end or image plot
				break
			endif
			
			setX = hcsr($csr, gName) * xMul + xOff												// correct for current trace's offsets
			GetAllTraceOffsets(gName, nextTrace, xOff, yOff, xMul, yMul)
			setX = (setX - xOff) / xMul															// correct for next trace's offsets
			
			wave w = TraceNameToWaveRef(gName,nextTrace)
			wave/Z w_x = XWaveRefFromTrace(gName, nextTrace)
			cursorPos = x2pnt(w,setX)
			if(WaveExists(w_x))
				FindLevel/P/Q w_x, setX;	cursorPos = !v_flag ? V_LevelX : DimSize(w,0)/2
			endif
			
			Cursor/W=$gName/P $csr, $nextTrace, cursorPos										// move cursor to next trace
			FetchLatestCursorValues(csr)
		break
		case "trAddXButton":
			GoUp = 1
		case "trRemXButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			
			GoUp = GoUp == 0 ? -1 : GoUp
			ControlInfo/W=$pSingle trOffxDelta
			setX = V_Value
			if (setX < 0)
				ModifyGraph/W=$gName muloffset($trace)={-GoUp*abs(xMul),}
			else
				xMul = xMul == -1 ? 0 : xMul
				ModifyGraph/W=$gName muloffset($trace)={abs(xMul),}
			endif
			ModifyGraph/W=$gName offset($trace)={(GoUp*abs(setX)+xOff),}
			FetchLatestCursorValues(csr, keepDelta=1)
		break
		case "trAddYButton":
			GoUp = 1
		case "trRemYButton":
			if (!strlen(trace) || !strlen(csr))
				break
			endif
			
			GoUp = GoUp == 0 ? -1 : GoUp
			ControlInfo/W=$pSingle trOffyDelta
			setY = V_Value
			if (setY < 0)
				ModifyGraph/W=$gName muloffset($trace)={,-GoUp*abs(yMul)}
			else
				yMul = yMul == -1 ? 0 : yMul
				ModifyGraph/W=$gName muloffset($trace)={,abs(yMul)}
			endif
			ModifyGraph/W=$gName offset($trace)={,(GoUp*abs(setY)+yOff)}
			AdjustProfileAxisScaling()
			FetchLatestCursorValues(csr, keepDelta=1)
		break
		case "trNormTo1Button":
			if (strlen(trace) && strlen(csr))
				if (vcsr($csr, gName) != 0)
					ModifyGraph/W=$gName muloffset($trace)={,1/vcsr($csr, gName)}
					FetchLatestCursorValues(csr)
				endif
			endif
		break
		case "trApplyCurButton":
			if (strlen(trace))																	// falls through with only the current trace selected
				traces = trace
				items  = 1
			else
				break
			endif
		case "vaApplyAllButton":
			for (i = 0; i < items; i++)
				trace = StringFromList(i,traces)
				GetAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
				
				wave work = TraceNameToWaveRef(gName, trace)									// apply all scales and offsets to 1D data
				wave/Z w_x = XWaveRefFromTrace(gName, trace)
				if (WaveDims(work) != 1)														// just in case there is 2D data mixed in
					continue
				endif
				work *= yMul
				work += yOff
				if(WaveExists(w_x))
					w_x *= xMul
					w_x += xOff
				else
					SetScale/P x, DimOffset(work,0) * xMul + xOff, DimDelta(work,0) * xMul, WaveUnits(work,0), work
				endif
				ModifyGraph/W=$gName offset($trace)={0,0} ,muloffset($trace)={0,0}
				
				string lognote = ""																// write a simple log
				lognote += SelectString((xMul!=1),"","\rx-scale=" +num2str(xMul))
				lognote += SelectString((xOff!=0),"","\rx-offset="+num2str(xOff))
				lognote += SelectString((yMul!=1),"","\ry-scale=" +num2str(yMul))
				lognote += SelectString((yOff!=0),"","\ry-offset="+num2str(yOff))
				if (strlen(lognote))
					lognote[0] = "\rscaling and offsets applied:"
					Note work, lognote
				endif
			endfor
			FetchLatestTraceValues()
		break
	EndSwitch
	return retVal
end

//#################################### popup helper function #####################################

function TraceTool_ProfileSelect(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif
	
	int noSelect = StringMatch(s.popStr,"none")
	variable axFract = 0.73
	variable axRange = noSelect ? 1 : axFract-0.03
	
	SetActiveSubwindow $gName																	// make sure to do stuff in the graph window
	string axes = AxisList(gName)
	if (FindListItem("Profileleft", axes) != -1)												// see if the additional Profile axis is attached
		RemoveFromGraph/Z/W=$gName $("AddTraceProfile")											// remove the old trace (if present)
	endif
	if (WhichListItem("bottom",axes) > -1)														// make the bottom axis shorter
		ModifyGraph/W=$gName axisEnab(bottom)={0,axRange}
	endif
	if (WhichListItem("top",axes) > -1)
		ModifyGraph/W=$gName axisEnab(top)={0,axRange}
	endif
	if (noSelect)																				// do nothing
		return 0
	endif
	
	AppendToGraph/W=$gName/C=(0,15872,65280)/B=Profilebottom/L=Profileleft/VERT $s.popStr/TN=AddTraceProfile	// append the Profile trace
	ModifyGraph/W=$gName axisEnab(Profilebottom)={axFract+0.03,1}	
	ModifyGraph/W=$gName freePos(Profileleft)   ={axFract+0.03,kwFraction}						// move the new axes to the right position
	
	ModifyGraph/W=$gName freePos(Profilebottom)=0, mirror(Profilebottom)=2
	ModifyGraph/W=$gName tkLblRot(Profileleft)=90, tlOffset(Profileleft)=-2
	ModifyGraph/W=$gName minor(Profileleft)=1, tick(Profileleft)=2, standoff=0
	ModifyGraph/W=$gName ZisZ(Profilebottom)=1,	ZisZ(Profileleft)=1
	ModifyGraph/W=$gName mirror(Profilebottom)=2,notation(Profilebottom)=1
	AdjustProfileAxisScaling()
	return 0
end

//################################################################################################

function ColorSpectrumFromTable(int rev, string table)
	string traces = "", gName = GetTopGraph()
	int i, itm = FetchTraces(gName,traces)
	if (!itm)																					// may be an image
		string image = StringFromList(0,ImageNameList(gName,";"))
		if (strlen(image))
			ModifyImage/W=$gName $image ctab={,,$table,rev}
		endif
		return 0
	endif
	ColorTab2Wave $table; wave M_colors
	int colornum = DimSize(M_Colors,0)-1
	for (i = 0; i < itm; i++)
		variable row = itm==1 ? 0 : (i/(itm-1))*colornum
		variable idx = rev==0 ? i : (itm-1)-i
		ModifyGraph/W=$gName/Z rgb($StringFromList(idx,traces)) = (M_Colors[row][0], M_Colors[row][1], M_Colors[row][2])
	endfor
	KillWaves/Z M_colors
	return 0
end

//________________________________________________________________________________________________
//
//										GRAPH SIZE PANEL
//________________________________________________________________________________________________

static StrConstant kGraphSizePanelTitle = "Graph Size & Styling"

function BuildGraphSizePanel()
	string pName = "GraphSizePanel"
	if (WinType(pName))
		DoWindow/F $pName
		return 0
	endif
	
	variable pW = 350, pH = 650, pM = 10, pC = (pW-60)/2, pMR = pM+60
	variable TFTResX, TFTResY																	// get the display resolution to align window better (center of full screen)
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,",")) - Str2Num(StringFromList(0,TFTInfo,",")))*(72/ScreenResolution)  
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,",")) - Str2Num(StringFromList(1,TFTInfo,",")))*(72/ScreenResolution)  
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	
	string option = StrVarOrDefault("root:"+ksGraphSizeSetting,"anchor:RT;background:1;frame:3;unit:1;")								// load saved settings
	string anchor = StringByKey("anchor" ,option)		;anchor = SelectString(strlen(anchor),"RT",anchor)
	variable back = NumberByKey("background" ,option)	;back = numtype(back) ? 1 : max(back,1)
	variable frame= NumberByKey("frame" ,option)		;frame= numtype(frame)? 3 : max(frame,1)
	variable unit = NumberByKey("unit" ,option)			;unit = numtype(unit) ? 1 : max(unit,1)
	
	NewPanel/K=1/W=((TFTResX-pW)/2,(TFTResY-pH)/2,(TFTResX+pW)/2,(TFTResY+pH)/2)/N=$pName as kGraphSizePanelTitle
	ModifyPanel /W=$pName ,fixedSize=1 ,noEdit=1
	SetWindow $pName hook(UpdateHook)=SizePanelUpdateFunc
	SetWindow $pName userdata(procVersion)=num2str(kVersion)									// save procedure version
	
	SetDrawLayer UserBack
	SetDrawEnv linethick=2,			fillpat= 0;	DrawRect pM+5,	pM,		pW-pM-5, 	pM+190		// full rectangle
	SetDrawEnv linethick=2, dash=3,	fillpat= 0;	DrawRect pM+45,	pM+30,	pW-pM-45,	pM+160		// dashed rectangle
	
	SetDrawEnv linethick=2,linefgc=(16380,28400,65535),arrow=3,arrowfat=0.80;	DrawLine pM+47,   pM+62,  pW-pM-47, pM+62	// blue arrows
	SetDrawEnv linethick=2,linefgc=(16380,28400,65535),arrow=3,arrowfat=0.80;	DrawLine pC+95,   pM+32,  pC+95,    pM+157
	SetDrawEnv linethick=2,linefgc=( 8000,39320,19940),arrow=3,arrowfat=0.80;	DrawLine pM+9,    pM+142, pW-pM-9,  pM+142	// green arrows
	SetDrawEnv linethick=2,linefgc=( 8000,39320,19940),arrow=3,arrowfat=0.80;	DrawLine pC-30,   pM+2,   pC-30,    pM+187
	
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine pC-15,   pM+163, pC-15,    pM+187	// purple arrows
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine pC-15,   pM+3,   pC-15,    pM+27	// vert top
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine pM+8,    pM+85,  pM+43,    pM+85	// hor left
	SetDrawEnv linethick=2,linefgc=(36870,14755,58980),arrow=3,arrowfat=0.80;	DrawLine pW-pM-43,pM+85,  pW-pM-8,  pM+85	// hor right
	
	TitleBox TplotW				,pos={pC+8  ,45}	,size={48,16}	,title=" plot W: "	,anchor=MT ,labelBack=(16380,28400,65535)
	TitleBox TplotH				,pos={pC+73 ,85}	,size={45,16}	,title=" plot H: "	,anchor=MT ,labelBack=(16380,28400,65535)
	TitleBox TgraphW			,pos={pC-60 ,85}	,size={55,16}	,title=" graph H: "	,anchor=MT ,labelBack=(8000,39320,19940)
	TitleBox TgraphH			,pos={pC+3	,125}	,size={58,16}	,title=" graph W: "	,anchor=MT ,labelBack=(8000,39320,19940)
	ModifyControlList ControlNameList(pName,";","T*")	,frame=0	,fStyle=1	,fixedSize=1	,fColor=(65535,65535,65535)

	SetVariable VTopMargin		,pos={pC,	 16}	,value=_NUM:0
	SetVariable VPlotWidth		,pos={pC,	 63}	,value=_NUM:0	
	SetVariable VLeftMargin		,pos={pM-5,	 103}	,value=_NUM:0
	SetVariable VRightMargin	,pos={pW-pMR+3,103}	,value=_NUM:0
	SetVariable VGraphHeight	,pos={pC-65, 103}	,value=_NUM:0
	SetVariable VPlotHeight		,pos={pC+65, 103}	,value=_NUM:0
	SetVariable VBottomMargin	,pos={pC,	 177}	,value=_NUM:0
	SetVariable VGraphWidth		,pos={pC,	 143}	,value=_NUM:0
	SetVariable VMagnification	,pos={pW-pMR-12,177},value=_NUM:0
	SetVariable VGraphFont		,pos={pM+25, 177}	,value=_NUM:0
	ModifyControlList ControlNameList(pName,";","V*") ,limits={0,2000,1} ,size={60,18} ,bodyWidth=55 ,proc=Gsize_VarControl
	SetVariable VMagnification	,limits={0,8,0.25} 	,bodyWidth=60	,title="magnify:"	,format="%g x"
	SetVariable VGraphFont		,limits={0,100,0}	,bodyWidth=40	,title="font:"								,help={"Sets the global graph font. If shift+enter is pressed instead the global axes font is changed. Insert 0 to reset to 'auto'"}
	
	PopupMenu FrameColorSelect	,pos={pM+10, 16}	,size={80,22}	,title="colors:"
	PopupMenu PlotColorSelect	,pos={pM+10, 45}	,size={80,22}	,title=""
	
	PopupMenu SizeUnit			,pos={pW-pMR-38,16}	,size={85,23}	,title="unit: "
	TitleBox SwapTitle			,pos={pC+20, 85}	,size={60,23}	,title="\JCswap:"				,frame=0
	CheckBox SwapAxes			,pos={pC+27, 105}	,size={23,23}	,title=""						,value=0	,proc=Gsize_SwapGraph
	CheckBox cPrint				,pos={pM+100, 210}	,size={23,20}	,title="Print Commands"			,value=0	,help={"Prints modify commands into history."}
	CheckBox KeepAutoMode		,pos={pW-120,210}	,size={23,20}	,title="Don't Lock Sizes"		,value=1	,help={"Will reset the mode to Auto after changing the size."}
	Button ResetAllButton		,pos={pM+5,  208}	,size={85,20}	,title="reset sizes"			,fstyle=1	,help={"Resets all size controls to automatic mode."}

	variable g = 230, d = 26
	GroupBox GrAxisToggle		,pos={pM-2,  g+9}	,size={334,128}	,title="Set Axis Tick & Label Margins:"		,fStyle=1	,frame=0		,fColor=(36870,14755,58980)
	TabControl tickCtrlTabs 	,pos={pM+2,	 g+d*1} ,size={327,107}	,tabLabel(0)="default axes" ,tabLabel(1)="custom axes"	,focusring=0	,proc=Gsize_TabControl	,disable=2
	Button ToggleAxisMode		,pos={pW-90, g+11}	,size={75,20}	,title="\Zr090switch mode"					,help={"Switch to a different mode to change other properties of the axis."}		
	
	TitleBox SetAxisTile		,pos={pM+3,  g+d*2}	,size={40,23}	,title="Axis:"					,frame=0	,anchor=RT
	CheckBox cAxis_left			,pos={57,    g+d*2}	,size={23,23}	,title="Left"					,value=0
	CheckBox cAxis_bottom		,pos={129,   g+d*2}	,size={23,23}	,title="Bottom"					,value=0
	CheckBox cAxis_right		,pos={201,   g+d*2}	,size={23,23}	,title="Right"					,value=0
	CheckBox cAxis_top			,pos={272,   g+d*2}	,size={23,23}	,title="Top"					,value=0
	
	g -= 2
	SetVariable tlAxis_left		,pos={57,    g+d*3}	,value=_NUM:0	,title="Tick:"
	SetVariable tlAxis_bottom	,pos={129,   g+d*3}	,value=_NUM:0
	SetVariable tlAxis_right	,pos={201,   g+d*3}	,value=_NUM:0
	SetVariable tlAxis_top		,pos={272,   g+d*3}	,value=_NUM:0
	
	SetVariable alAxis_left		,pos={57,    g+d*4}	,value=_NUM:0	,title="Label:"
	SetVariable alAxis_bottom	,pos={129,   g+d*4}	,value=_NUM:0
	SetVariable alAxis_right	,pos={201,   g+d*4}	,value=_NUM:0
	SetVariable alAxis_top		,pos={272,   g+d*4}	,value=_NUM:0
	
	g += 140
	GroupBox GrStyleSetting		,pos={pM-2,  g+4}		,size={334,107}	,title="Set & Copy Sizes / Styles:"	,fStyle=1	,frame=0	,fColor=(8000,39320,19940)
	TitleBox CopyStyleTile		,pos={pM+13, g+d*1+3}	,size={40,23}	,title="\f01Copy:"					,fStyle=0	,frame=0
	Button CopySizeButton		,pos={pM+47, g+d*1}		,size={45,22}	,title="Sizes"						,help={"Copies all graph-size related values."}
	Button CopyStyleButton		,pos={pM+102,g+d*1}		,size={45,22}	,title="Style"						,help={"Copies the style of the top graph (hold shift to save style to the clipboard)."}
	CheckBox CopyToClipboard	,pos={pM+155,g+d*1+3}	,size={40,22}	,title="to Clip" 		,value=0	,help={"Copies to the clipboard."}
	CheckBox CopyAxes			,pos={pM+215,g+d*1+3}	,size={60,22}	,title="with Axis Range",value=1	,help={"Copy or apply the axis scaling as well."}
	
	PopupMenu GrStyle			,pos={pM+10, g+d*2+3}	,size={187,22}	,title="Style:"
	Button ApplyButton			,pos={pM+205,g+d*2+2}	,size={52,22}	,title="\f01Apply"					,help={"Applies selected style or size to the top graph."}
	Button RevertButton			,pos={pM+268,g+d*2+2}	,size={52,22}	,title="Revert"						,help={"Reverts to last state before Apply was pressed."}
	
	PopupMenu AspectSelect		,pos={pM+10, g+d*3+5}	,size={105,22}	,title="Aspect:"
	PopupMenu FormatSelect		,pos={pM+205,g+d*3+5}	,size={115,22}	,title="   Apply Format"			,help={"Select from a set of predefined formats to apply to the current graph."}
	CheckBox AspectMode			,pos={pM+125,g+d*3+7}	,size={60,22}	,title="as Width"		,value=0	,help={"Set the aspect ratio as width instead of the height."}
	
	g += 114
	GroupBox GrAnnotationTools	,pos={pM-2,  g+4}		,size={334,158}	,title="Set & Add Legend:"			,fStyle=1	,frame=0	,fColor=(10,34800,52430)
	TitleBox AnnoAnchorTile		,pos={pM+12, g+d*2-4}	,size={40,23}	,title="Anchor:"					,fStyle=0	,frame=0
	GroupBox GrAnchorBox		,pos={pM+10, g+d*3-15}	,size={60,60}	,title=""
	CheckBox cAnchor_LT			,pos={pM+15, g+d*3-9}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_LC			,pos={pM+15, g+d*3+7}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_LB			,pos={pM+15, g+d*3+23}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_MT			,pos={pM+31, g+d*3-9}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_MC			,pos={pM+31, g+d*3+7}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_MB			,pos={pM+31, g+d*3+23}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_RT			,pos={pM+47, g+d*3-9}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_RC			,pos={pM+47, g+d*3+7}	,size={23,23}	,title=""	,mode=1		,value=0
	CheckBox cAnchor_RB			,pos={pM+47, g+d*3+23}	,size={23,23}	,title=""	,mode=1		,value=0
	
	TitleBox AnnoSelectTile		,pos={pM+8,	 g+d*1+2}	,size={40,23}	,title="Select:"		,fStyle=1	,frame=0
	PopupMenu AnnotationSelect	,pos={pM+50, g+d*1-1}	,size={22,22}	,title=""							,help={"Selects from one of the available annotations in the graph."}
	SetVariable AnnoRename		,pos={pM+55, g+d*1}		,size={180,22}	,title="Rename:"					,help={"Renames the selected annotation of the target graph."}
	Button KillAnnoButton		,pos={255,   g+d*1-2}	,size={75,22}	,title="Remove"						,help={"Kills the selected annotation."}
	
	PopupMenu AnnoFrameSelect	,pos={pM+55, g+d*2}		,size={180,22}	,title="Frame:"
	PopupMenu AnnoBackgrSelect	,pos={pM+55, g+d*3}		,size={180,22}	,title="Backgr.:"
	PopupMenu AnnoInfo1Select	,pos={pM+55, g+d*4}		,size={120,22}	,title="Add Info:"		,fStyle=2	,help={"Select trace information to dynamically display in the legend."}
	PopupMenu AnnoInfo2Select	,pos={pM+115,g+d*4}		,size={120,22}	,title="&:"				,fStyle=2	,help={"Select additional trace information to display in the legend."}
	
	Button CreateAnnoButton		,pos={255,   g+d*2}		,size={75,22}	,title="Add New"		,fStyle=1	,help={"Adds a new legend to the top graph."}
	CheckBox cAnno_ReplaceOld	,pos={265,   g+d*3+2}	,size={22,22}	,title="Replace"		,value=0	,help={"If active, any new annotation replaces the selected existing one."}
	CheckBox cAnno_DoTextbox	,pos={265,   g+d*4-2}	,size={22,22}	,title="Textbox"		,value=0	,help={"If active, creates a text-box style legend instead of a dynamic legend."}
	
	DrawLine pM+15, g+d*5-2, pW-pM-15, g+d*5-2
	TitleBox AnnoFontTile		,pos={pM+8,	 g+d*5.2+2}	,size={40,23}	,title="Text:"			,fStyle=1	,frame=0
	CheckBox cAnno_TraceColor	,pos={pM+45, g+d*5.2+2}	,size={22,22}	,title="Color:"			,value=0	,help={"Colors any description text in the same color as the trace."}	,side=1
	SetVariable AnnoColumns		,pos={pM+53, g+d*5.2}	,size={120,22}	,title="Cols.:"						,help={"Set to > 1 to arrange the trace legend in multiple columns."}
	SetVariable AnnoFontZoom	,pos={pW-178,g+d*5.2}	,size={100,22}	,title="Size:"						,help={"Set the relative size of fonts for the trace legend."}
	CheckBox cAnno_TextBold		,pos={pW-69, g+d*5.2+2}	,size={60,22}	,title="b"  ,fStyle=1	,value=0	,help={"Toggle trace-info text to bold."}
	CheckBox cAnno_TextItalic	,pos={pW-40, g+d*5.2+2}	,size={60,22}	,title="i"  ,fStyle=2	,value=0	,help={"Toggle trace-info text to italic."}
	
	PopupMenu GrStyle			,mode=1		,value=Gsize_FetchGraphMacroList()									,bodyWidth=150
	PopupMenu SizeUnit			,mode=unit	,value=#"\"points;inches;cm;\""			,proc=Gsize_PresetPopup		,bodyWidth=60
	PopupMenu AspectSelect		,mode=1		,value=#("\""+kAspectSelectList+"\"")								,bodyWidth=68
	PopupMenu FormatSelect		,mode=0		,value=#("\""+kFormatSelectList+"\"")								,bodyWidth=115
	PopupMenu AnnoFrameSelect	,mode=frame	,value=#"\"None;UnderLine;Box;\""									,bodyWidth=103
	PopupMenu AnnoBackgrSelect	,mode=back	,value=#"\"Opaque;Transparent;Graph Color;Window Color;\""			,bodyWidth=103
	PopupMenu AnnoInfo1Select	,mode=1		,value=#"\"  -;Δx;Δy;x*;y*;\""										,bodyWidth=43
	PopupMenu AnnoInfo2Select	,mode=1		,value=#"\"  -;Δx;Δy;x*;y*;\""										,bodyWidth=43
	PopupMenu AnnotationSelect	,mode=0		,value=Gsize_FetchGraphAnnotationList()								,bodyWidth=20
	PopupMenu FrameColorSelect	,mode=1		,value= "*COLORPOP*"	,popColor=(65535,65535,65535)				,bodyWidth=35
	PopupMenu PlotColorSelect	,mode=1		,value= "*COLORPOP*"	,popColor=(65535,65535,65535)				,bodyWidth=35
	
	SetVariable AnnoColumns									,value=_NUM:1	,bodyWidth=40	,limits={1,10,1}	,proc=Gsize_AnnoVarChange
	SetVariable AnnoFontZoom				,format="%g%"	,value=_NUM:100	,bodyWidth=53	,limits={1,999,25}	,proc=Gsize_AnnoVarChange
	SetVariable AnnoRename									,value=_STR:""	,bodyWidth=103						,proc=Gsize_AnnoDoRename
	ModifyControlList ControlNameList(pName,";","tlAxis_*")	,size={55,18}	,bodyWidth=55	,limits={-50,50,1}	,proc=Gsize_VarControl
	ModifyControlList ControlNameList(pName,";","alAxis_*")	,size={55,18}	,bodyWidth=55	,limits={-50,50,1}	,proc=Gsize_VarControl
	ModifyControlList ControlNameList(pName,";","*Button")	,proc=Gsize_CopyPaste
	ModifyControlList ControlNameList(pName,";","*Select")	,proc=Gsize_PresetPopup
	ModifyControlList ControlNameList(pName,";","cAxis_*")	,proc=Gsize_ToggleAxis
	ModifyControlList ControlNameList(pName,";","cAnchor_*"),proc=Gsize_ToggleAnchor
	ModifyControlList ControlNameList(pName,";","cAnno_*")	,proc=Gsize_ToggleAnnotation
	Button ToggleAxisMode									,proc=Gsize_ToggleModeButton
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pName,";","*") ,fsize=10
	endif
	
	// +++++++++++++++++++++++++++++++ apply settings ++++++++++++++++++++++++++++++++++++++++
	string ctrlList = RemoveFromList("VMagnification;VGraphFont;",ControlNameList(pName,";","V*"))
	variable div  = unit == 1 ? 1 : (unit == 2 ? 72 : 72/2.54)								// in points, inches or cm
	variable step = unit == 1 ? 1 : (unit == 2 ? 0.014 : 0.036)
	ModifyControlList ctrlList ,limits={0,8000/div,step}									// official max size is 8000 points
	CheckBox $("cAnchor_"+anchor) value = 1
	
	string gName = GetTopMainGraph()
	if (strlen(gName))
		AutoPositionWindow/M=0/R=$gName $pName
	endif
	
	Gsize_FetchGraphSizeValues()
	return 0
end
//++++++++++++++++++++++++++++++++ panel update helper function ++++++++++++++++++++++++++++++++++
function SizePanelUpdateFunc(STRUCT WMWinHookStruct &s)
	if (s.eventCode == 0)
		Gsize_FetchGraphSizeValues()
		return 1
	endif
	return 0
end

//################################################################################################

function/S Gsize_FetchGraphMacroList()
	string list = MacroList("!BoxStyle",";","KIND:1,SUBTYPE:GraphStyle")
	string name = GetUserData("GraphSizePanel", "", "CopyStyleName")
	string copy = GetUserData("GraphSizePanel", "", "CopyStyle")
	string size = GetUserData("GraphSizePanel", "", "CopySize")
	if (strlen(copy))
		list += "Copied" + SelectString(strlen(name),""," Style: "+name)+";"
	endif
	if (strlen(size))
		list += "Copied" + SelectString(strlen(name),""," Size: "+name)+";"
	endif
	return "BoxStyle;"+list+"From Clipboard;"													// BoxStyle is always at the front
end

//################################################################################################

function/S Gsize_FetchGraphAnnotationList()
	string name = GetTopGraph(), list = ""
	if (WinType(name) == 1)
		list = AnnotationList(name)
	endif
	return SelectString(strlen(list),"no annotation",list)
end

//################################################################################################

function Gsize_TabControl(STRUCT WMTabControlAction &s) : TabControl
	if (s.eventCode == 2)
		Gsize_FetchGraphSizeValues()
	endif
end

//################################################################################################

function Gsize_SwapGraph(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif
	ModifyGraph/W=$gName/Z swapXY=s.checked
	ControlInfo/W=$(s.win) cPrint
	if (V_Value && V_flag == 2)
		print "•ModifyGraph swapXY="+num2str(s.checked)
	endif
	Gsize_FetchGraphSizeValues()
	return 0
end
//################################################################################################

function Gsize_ToggleAnchor(STRUCT WMCheckboxAction& s) : CheckBoxControl						// selects and saves the current annotation anchor
	if (s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph()
	string list = RemoveFromList(s.ctrlName,ControlNameList(s.win,";","cAnchor_*")); int i
	for (i = 0; i < ItemsInList(list); i++)
		CheckBox $StringFromList(i,list) ,win=$(s.win) ,value=0
	endfor
	string anchor; SplitString/E="cAnchor_(.*)" s.ctrlName, anchor
	string settings = StrVarOrDefault("root:"+ksGraphSizeSetting,"")
	string/G root:$ksGraphSizeSetting = ReplaceStringByKey("anchor",settings,anchor)
	
	ControlInfo/W=$(s.win) AnnoRename	;string name = S_Value
	ControlInfo/W=$(s.win) cPrint		;int doPrint = V_Value
	if (!strlen(gName) || !strlen(name))
		return 0
	endif
	
	string cmd = StringByKey("TYPE",AnnotationInfo(gName,name))+"/W="+gName+"/C/N="+name+"/A="+anchor+"/X=5.00/Y=5.00"
	Execute/Q/Z cmd
	if (doPrint)
		Print "•"+ReplaceString("/W="+gName,cmd,"")
	endif
	Gsize_FetchGraphSizeValues()
	return 0
end

//################################################################################################

function Gsize_ToggleAnnotation(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph(), cmd = ""
	if (!strlen(gName))
		return 0
	endif
	
	ControlInfo/W=$(s.win) AnnoRename	;string name = S_Value
	ControlInfo/W=$(s.win) cPrint		;int doPrint = V_Value
	if (!strlen(gName) || !strlen(name))
		return 0
	endif
	
	StrSwitch(s.ctrlName)
		case "cAnno_TraceColor":
			cmd = "QuickLegendStyle(graphName=\""+gName+"\", annoName=\""+name+"\", color="+num2str(s.checked ? 2 : 1)+")"
		break
		case "cAnno_TextBold":
		case "cAnno_TextItalic":
			int fontStyle = 0
			ControlInfo/W=$(s.win) cAnno_TextBold;		fontStyle+= V_Value
			ControlInfo/W=$(s.win) cAnno_TextItalic;	fontStyle+= V_Value*2
			cmd = "QuickLegendStyle(graphName=\""+gName+"\", annoName=\""+name+"\", fontStyle="+num2str(fontStyle)+")"
		break
		case "cAnno_DoTextbox":
			string type = StringByKey("TYPE",AnnotationInfo(gName,name)), list
			int isImage = !CmpStr(type,"ColorScale") || (strlen(StringFromList(0,ImageNameList(gName,";"))) && !FetchAllTraces(gName,list))
			int isValid = !CmpStr(type,"Legend") || !CmpStr(type,"TextBox") || isImage
			if (isValid)
				cmd = SelectString(s.checked,SelectString(isImage,"Legend","ColorScale"),"TextBox")+"/W="+gName+"/C/N="+name
			endif
		break
	EndSwitch
	
	if (strlen(cmd))
		Execute/Q/Z cmd
		if (doPrint)
			Print "•"+ReplaceString("/W="+gName,cmd,"")
		endif
	endif
	Gsize_FetchGraphSizeValues()
	return 0
end

//################################################################################################

function Gsize_ToggleAxis(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif
	
	string axis = ReplaceString("cAxis_",s.ctrlName,"")
	ControlInfo/W=$(s.win) tickCtrlTabs															// new in ver. 2.20
	if (V_flag == 8 && V_Value == 1)
		axis = GetUserData(s.win, s.ctrlName, "axisName")
	endif
	
	ToggleAxisProperty(s.win, gName, axis, !s.checked)
	Gsize_FetchGraphSizeValues()
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function ToggleAxisProperty(string pName, string gName, string axName, variable isOff)
	if (FindListItem(axName,AxisList(gName))<0)
		return 0
	endif
	
	ControlInfo/W=$pName ToggleAxisMode;	variable mode = 0
	if (V_flag == 1)
		mode = str2num(GetUserData(pName,"ToggleAxisMode", "axisMode"))							// new in ver. 2.20
		mode = numtype(mode) ? 0 : mode
	endif
	
	string str
	if (mode)
		ModifyGraph/W=$gName grid($axName)=2*!isOff
		str = "ModifyGraph grid("+axName+")="+num2str(2*!isOff)
	else
		ModifyGraph/W=$gName tick($axName)=3*isOff, noLabel($axName)=1*isOff
		str = "ModifyGraph tick("+axName+")="+num2str(3*isOff)+", noLabel("+axName+")="+num2str(1*isOff)
	endif
	
	ControlInfo/W=$pName cPrint
	if (V_Value)
		Print "•"+str
	endif
end


//################################################################################################

function Gsize_ToggleModeButton(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	string str = GetUserData(s.win, s.ctrlName, "axisMode")
	int mode = strlen(str) ? str2num(str) : 0;	mode = !mode
	Button ToggleAxisMode ,win=$(s.win) ,userData(axisMode)=num2str(mode)
	Switch (mode)
		case 0:
			GroupBox GrAxisToggle ,win=$(s.win) ,title="Set Axis Tick & Label Margins:"		,fColor=(36870,14755,58980)
			TitleBox SetAxisTile  ,win=$(s.win) ,title="Axis:"
		break
		case 1:
			GroupBox GrAxisToggle ,win=$(s.win) ,title="Set Axis Tick Size & Label Center:"	,fColor=(51400,26000,13100)
			TitleBox SetAxisTile  ,win=$(s.win) ,title="Grid:"
		break
	EndSwitch
	Gsize_FetchGraphSizeValues()
end

//################################################################################################

function Gsize_CopyPaste(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif
	string topLevelgName = StringFromList(0, gName, "#")
	string list = "", settings = "", cmd = "", name = ""
	int i,j, mode = 0
	
	ControlInfo/W=$(s.win) cPrint;	int doPrint = (V_Value && V_flag == 2)
	ControlInfo/W=$(s.win) CopyAxes;int doAxes  = (V_Value && V_flag == 2)						// include axes if checkbox is active
	
	string image = StringFromList(0,ImageNameList(gName,";"))
	int isImage = (strlen(image) && !FetchAllTraces(gName,list))								// do a color scale in case of an image plot
	
	StrSwitch(s.ctrlName)
		case "ResetAllButton":
			ModifyGraph/W=$gName/Z width=0, height=0, margin=0
			ModifyGraph/W=$gName/Z gfSize=0, fSize=0, lblMargin=0, tlOffset=0
			ModifyGraph/W=$gName/Z btLen=0, lblPos=0, lblLatPos=0
//			ModifyGraph/W=$gName/Z tick=0, noLabel=0, swapXY=0, expand=0
		break
		case "CopySizeButton":	
			mode = 1
		case "CopyStyleButton":																	// copies style of selected graph
			cmd = WinRecreation(gName,1)
			string childList = ChildWindowList(gName), childStyle = ""							// need to clean the style macro of any subwindow styles (won't work anyway)
			for (i = 0; i < ItemsInList(childList); i++)
				childStyle = WinRecreation(gName+"#"+StringFromList(i,childList),1)
				childStyle = RemoveListItem(0,childStyle, "\r")									// remove "proc" line
				childStyle = RemoveListItem(0,childStyle, "\r")									// remove "pauseupdate" line
				childStyle = RemoveListItem(ItemsInList(childStyle,"\r") - 1,childStyle, "\r")	// remove "endmacro" line
				cmd = ReplaceString(childStyle,cmd,"")											// remove child window style
			endfor
			cmd = SpectraGraphTools#Gsize_cleanupMacro(cmd)
			
			// +++++++++++++++++++++++++++++++ copy and prepare sizes ++++++++++++++++++++++++++++
			if (mode)																			// copy only size related settings
				cmd = GrepList(cmd,"(tick|btLen|noLabel|tlOffset|lblMargin|lblPos|lblLatPos|fSize|swapXY|expand)",0,"\r")
				GetWindow $gName gsize
				variable graphL = V_left, graphT = V_top, graphR = V_right, graphB = V_bottom
				GetWindow $gName psize
				variable plotW = (V_right - V_left),		plotH = (V_bottom- V_top)
				variable margL = round(V_left- graphL),		margT = round(V_top - graphT)
				variable margR = round(graphR - V_right),	margB = round(graphB - V_bottom)
				cmd += "ModifyGraph/Z height="+num2str(plotH)+", width=" +num2str(plotW)+"\r"	// add designated size and margin specifiers
				cmd += "ModifyGraph/Z margin(left)="  +num2str(margL)+", margin(top)="   +num2str(margT)+"\r"
				cmd += "ModifyGraph/Z margin(right)=" +num2str(margR)+", margin(bottom)="+num2str(margB)+"\r"
			endif
			
			// +++++++++++++++++++++++++++++++++ save style to panel +++++++++++++++++++++++++++++
			if (!doAxes)
				cmd = GrepList(cmd, "^(SetAxis)", 1, "\r")
			endif
			ControlInfo/W=$(s.win) CopyToClipboard	
			if ((GetKeyState(0) & 2^2) || (V_flag == 2 && V_Value))								// shift key saves to clipboard
				PutScrapText cmd
				PopupMenu GrStyle ,win=$(s.win) ,popMatch="From Clipboard"
			else
				SetWindow $(s.win), userdata($SelectString(mode,"CopyStyle","CopySize"))=ReplaceString("\r",cmd,";")		// save into panel's user data
				SetWindow $(s.win), userdata(CopyStyleName)=gName
				PopupMenu GrStyle ,win=$(s.win) ,popMatch=SelectString(mode,"Copied Style*","Copied Size*")
			endif
		break
		case "ApplyButton":																		// applies selected graph style
			ControlInfo/W=$(s.win) GrStyle;	cmd = S_Value
			string lastStyle = WinRecreation(topLevelgName,0)
			int noRevert = (StringMatch(lastStyle,"*TransformMirror*") || StringMatch(lastStyle,"*AxisTransform*"))	// graphs with a transformed axis cannot be recreated
			if (noRevert)
				DoAlert 1, "Reverting the graph will not be possible. Apply anyway?"
				if (V_flag != 1)
					break
				endif
			endif
			
			int doStyle = StringMatch(cmd,"Copied*") ? (StringMatch(cmd,"Copied Size*") ? 2 : 1) : (StringMatch(cmd,"From Clipboard*") ? 3 : 0)

			ControlInfo/W=$(s.win) SwapAxes;	int doSwap = V_Value
			ControlInfo/W=$(s.win) KeepAutoMode;int keepAuto = V_Value
			Switch (doStyle)
				case 1:	// copied style
					cmd = GetUserData(s.win, "", "CopyStyle")
				break
				case 2:	// copied size
					cmd = GetUserData(s.win, "", "CopySize")
				break
				case 3:	// clipboard style
					cmd = GetScrapText()
					if (!StringMatch(cmd,"*ModifyGraph*"))
						Abort "It seems there are no valid style commands in the clipboard.\r\rTry to first copy a style (e.g., by using Copy Style while holding Shift)."
					endif
					cmd = ReplaceString("\r", SpectraGraphTools#Gsize_cleanupMacro(cmd), ";")
				break
			EndSwitch
			
			DoWindow/F $topLevelgName															// make sure the window is in front for the macro to work on
			if (strlen(cmd))
				ModifyGraph/W=$gName/Z btLen=0, lblPos=0, lblLatPos=0, lblMargin=0, tlOffset=0	// reset
				ModifyGraph/Z/W=$gName width=0, height=0, margin=0, gfSize=0, fSize=0
				if (!doAxes)
					cmd = GrepList(cmd, "^(SetAxis)", 1)
				endif
				if (doStyle)
					if (StringMatch(cmd,"swapXY=1") != doSwap)									// if the swap states between the copied style and the target graphs do not match ...
						if (doSwap)
							cmd = ReplaceString("height={Aspect",cmd,"width={Aspect")			// ... exchange the applied direction
						else
							cmd = ReplaceString("width={Aspect",cmd,"height={Aspect")
						endif
					endif
					for (i = 0; i < ItemsInList(cmd); i++)
						Execute/Q/Z StringFromList(i,cmd)
						if (doPrint && V_flag == 0)
							Print "•"+StringFromList(i,cmd)
						endif
					endfor
				else
					Execute/Q cmd+"()"
					if (doPrint)
						Print "•"+cmd+"()"
					endif
				endif
			endif
			DoWindow/F $(s.win)
			
			if (!noRevert)
				SetWindow $(s.win), userdata(LastgName)=topLevelgName							// save into panel's user data
				SetWindow $(s.win), userdata(LastStyle)=lastStyle
			endif
			
			if (keepAuto)
				DoUpdate/W=$gName
				ModifyGraph/W=$gName height= 0
				ModifyGraph/W=$gName width = 0
			endif
		break
		case "RevertButton":																	// copies top-graph's style
			cmd = GetUserData(s.win, "", "LastStyle")											// check for existence of backup style
			if (strlen(cmd) > 0)
				KillWindow $topLevelgName
				string/G ExeCmd = cmd
				Execute/P/Q "Execute ExeCmd"
				Execute/P/Q "KillStrings/Z ExeCmd"
			endif
		break
		case "KillAnnoButton":
			ControlInfo/W=$(s.win) AnnoRename
			if (strlen(S_Value))
				cmd = StringByKey("TYPE",AnnotationInfo(gName,S_Value))+"/W="+gName+"/K/N="+S_Value
				Execute/Q/Z cmd
				if (doPrint && !V_flag)
					Print "•"+cmd
				endif
			endif
		break
		case "CreateAnnoButton":
			string anchor		
			list = ControlNameList(s.win, ";", "cAnchor_*")
			for (i = 0; i < ItemsInList(list); i++)
				ControlInfo/W=$(s.win) $StringFromList(i,list)
				if (V_Value)
					SplitString/E="cAnchor_(.*)" StringFromList(i,list), anchor
					break
				endif
			endfor
			
			int fontStyle = 0
			anchor = SelectString(strlen(anchor),"RT",anchor)
			ControlInfo/W=$(s.win) AnnoFrameSelect;		int frame = V_Value-1
			ControlInfo/W=$(s.win) AnnoBackgrSelect;	int backg = V_Value-1
			ControlInfo/W=$(s.win) AnnoInfo1Select;		int mode1 = V_Value-2
			ControlInfo/W=$(s.win) AnnoInfo2Select;		int mode2 = V_Value-2
			ControlInfo/W=$(s.win) AnnoColumns;			int column= V_Value
			ControlInfo/W=$(s.win) AnnoFontZoom;		int ftZoom= V_Value
			ControlInfo/W=$(s.win) cAnno_TraceColor;	int color = V_Value*2
			ControlInfo/W=$(s.win) cAnno_ReplaceOld;	int doKill= V_Value
			ControlInfo/W=$(s.win) cAnno_DoTextbox;		int doBox = V_Value
			ControlInfo/W=$(s.win) cAnno_TextBold;		fontStyle+= V_Value
			ControlInfo/W=$(s.win) cAnno_TextItalic;	fontStyle+= V_Value*2
			
			variable digit1 = str2num(GetUserData(s.win,"AnnoInfo1Select", "digits"))			// live info: significant digits
			variable digit2 = str2num(GetUserData(s.win,"AnnoInfo2Select", "digits"))			// live info: significant digits
			digit1 = numtype(digit1) ? 5 : digit1
			digit2 = numtype(digit2) ? 5 : digit2
			string lastColor = GetUserData(s.win,"AnnoBackgrSelect", "lastColor")
			string backColor = SelectString(backg==0, num2str(backg), SelectString(strlen(lastColor),"0",lastColor))
			
			if (doKill)																			// first, kill the previous legend if requested
				ControlInfo/W=$(s.win) AnnoRename
				if (strlen(S_Value))
					Execute/Q/Z StringByKey("TYPE",AnnotationInfo(gName,S_Value))+"/W="+gName+"/K/N="+S_Value
				endif
				if (!V_flag)
					name = S_Value
				endif
			endif
			
			cmd  = SelectString(isImage,"Legend","ColorScale")+"/W="+gName+SelectString(strlen(name),"","/N="+name)
			cmd += "/F="+num2str(frame)+"/B="+backColor+"/A="+anchor + SelectString(isImage,""," image="+image)
			if (isImage)																		// append image name
				cmd += ";AppendText/W="+gName+SelectString(strlen(name),"","/N="+name)+" \""+image+"\""
			endif
			Execute/Q/Z cmd
			if (V_flag)
				break
			endif
			if (doPrint)
				Print "•"+cmd
			endif
			
			if (!strlen(name))
				list = AnnotationList(gName)
				name = StringFromList(ItemsInList(list)-1,list)
			endif
			
			if (doBox)
				TextBox/C/N=$name/W=$gName
			endif
			
			QuickLegendStyle(graphName=gName, annoName=name, color=color, fontStyle=fontStyle, fontZoom=ftZoom, columns=column, liveInfo=buildLiveInfoFormat(mode1,mode2,digit1,digit2))	// extra options: style legend
			SetVariable AnnoRename ,win=$(s.win) ,value=_STR:name ,userData(lastName)=name
		break
	EndSwitch
	Gsize_FetchGraphSizeValues()
	return 0
end

//################################################################################################

static function/S Gsize_cleanupMacro(string str)
	str = RemoveFromList(ListMatch(str,"*Proc*","\r"), str, "\r")								// remove "proc" line
	str = RemoveFromList(ListMatch(str,"*PauseUpdate*","\r"), str, "\r")						// remove "pauseupdate" line
	str = RemoveFromList(ListMatch(str,"*EndMacro*","\r"), str, "\r")							// remove "endmacro" line
	str = ReplaceString("\t", str, "")
	str = GrepList(str, "^(DefineGuide|ControlBar|SetWindow|Cursor|ListBox|CheckBox|PopupMenu|ValDisplay|SetVariable|Button|TitleBox|Slider|GroupBox|NewPanel|ModifyPanel|RenameWindow|SetActiveSubwindow|ShowTools)", 1, "\r") // remove control items
	str = GrepList(str, "^(ModifyGraph/Z rgb|ModifyGraph/Z muloffset|ModifyGraph/Z offset|ShowInfo)", 1, "\r")
	return str
end

//################################################################################################

function Gsize_AnnoVarChange(struct WMSetVariableAction &s) : SetVariableControl				// renames existing legend
	if (s.eventCode != 1 && s.eventCode != 2)
		return 0
	endif
	
	string gName = GetTopGraph(), cmd = ""
	ControlInfo/W=$(s.win) AnnoRename	;string name = S_Value
	ControlInfo/W=$(s.win) cPrint		;int doPrint = V_Value
	if (!strlen(gName) || !strlen(name))
		return 0
	endif
	
	StrSwitch(s.ctrlName)
		case "AnnoColumns":
			cmd = "QuickLegendStyle(graphName=\""+gName+"\", annoName=\""+name+"\", columns="+num2str(s.dval)+")"
		break
		case "AnnoFontZoom":
			cmd = "QuickLegendStyle(graphName=\""+gName+"\", annoName=\""+name+"\", fontZoom="+num2str(s.dval)+")"
		break
	EndSwitch
	
	if (strlen(cmd))
		Execute/Q/Z cmd
		if (doPrint)
			Print "•"+ReplaceString("/W="+gName,cmd,"")
		endif
	endif
	Gsize_FetchGraphSizeValues()
	return 0
end

//################################################################################################

function Gsize_AnnoDoRename(struct WMSetVariableAction &s) : SetVariableControl				// renames existing legend
	if (s.eventCode != 1 && s.eventCode != 2)
		return 0
	endif
	string gName = GetTopGraph()
	string pName = "GraphSizePanel"
	if (WinType(gName) != 1)
		return 0
	endif
	string anno_list = AnnotationList(gName)
	string last_name = GetUserData(pName, "AnnoRename", "lastName")
	if (WhichListItem(last_name,anno_list)>=0 && strlen(s.sval) && CmpStr(last_name,s.sval))
		if (WhichListItem(s.sval,anno_list)>=0)
			DoAlert/T="Name Duplicate" 0, "An annotation with that name already exists. Cannot rename!"
			return 0
		endif
		string cmd = StringByKey("TYPE",AnnotationInfo(gName,last_name))+"/W="+gName+"/C/N="+last_name+"/R="+s.sval
		Execute/Q/Z cmd
		ControlInfo/W=$(s.win) cPrint
		if (V_Value)
			Print "•"+ReplaceString("/W="+gName,cmd,"")
		endif
		SetVariable AnnoRename ,win=$pName ,userData(lastName)=s.sval
	endif
	Gsize_FetchGraphSizeValues()
end

//################################################################################################

function Gsize_VarControl(struct WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		ExecuteVarChange(s.ctrlName, s.dval)
	endif
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function ExecuteVarChange(string ctrlName, variable val)
	string gName = GetTopGraph(), axis
	string pName = "GraphSizePanel"
	if (!strlen(gName))
		return 0
	endif
	
	ControlInfo/W=$pName SizeUnit; variable modifier = V_Value == 1 ? 1 : (V_Value == 2 ? 72 : 72/2.54)		// in points, inches or cm
	ControlInfo/W=$pName KeepAutoMode;	int keepAuto = V_Value												// reset to Auto scale mode after size change
	ControlInfo/W=$pName cPrint;		int doPrint  = V_flag == 2 ? V_Value :0
	
	SplitString/E="Axis_(.*)" ctrlName, axis
	if (strlen(axis))
		ControlInfo/W=$pName tickCtrlTabs														// new in ver. 2.20
		if (V_flag == 8 && V_Value == 1)
			axis = GetUserData(pName, "cAxis_"+axis, "axisName")
		endif
		ControlInfo/W=$pName ToggleAxisMode
		variable ax_mode = 0, lblMode = 0
		if (V_flag == 1)
			ax_mode = str2num(GetUserData(pName,"ToggleAxisMode", "axisMode"))					// new in ver. 2.20
			ax_mode = numtype(ax_mode) ? 0 : ax_mode
		endif
		lblMode = str2num(GetUserData(pName, ctrlName, "labelMode"))
		lblMode = numtype(lblMode) ? 1 : lblMode
	endif
	
	variable shftKey = ((GetKeyState(0) & 2^2) != 0)
	string cmd = ""
	StrSwitch(ctrlName)
		case "VLeftMargin":
			cmd = "margin(left)="+num2str(val*modifier)
		break
		case "VBottomMargin":
			cmd = "margin(bottom)="+num2str(val*modifier)
		break
		case "VRightMargin":
			cmd = "margin(right)="+num2str(val*modifier)
		break
		case "VTopMargin":
			cmd = "margin(top)="+num2str(val*modifier)
		break
		case "VGraphHeight":
			ControlInfo/W=$pName VTopMargin;	variable marginT = V_Value
			ControlInfo/W=$pName VBottomMargin;	variable marginB = V_Value
			variable plotH = val-marginT-marginB
			if (val == 0)
				cmd = "margin(top)=0, margin(bottom)=0,  height=0"
			elseif (plotH > 0)
				cmd = "height="+num2str(plotH*modifier)
			endif
		break
		case "VGraphWidth":
			ControlInfo/W=$pName VLeftMargin;	variable marginL = V_Value
			ControlInfo/W=$pName VRightMargin;	variable marginR = V_Value
			variable plotW = val-marginL-marginR
			if (val == 0)
				cmd = "margin(left)=0, margin(right)=0,  width=0"
			elseif (plotW > 0)
				cmd = "width="+num2str(plotW*modifier)
			endif
		break
		case "VPlotHeight":
			cmd = "height="+num2str(val*modifier)
		break
		case "VPlotWidth":
			cmd = "width="+num2str(val*modifier)
		break
		case "VMagnification":
			val = val==1 ? 0 : val																// reset to none
			cmd = "expand="+num2str(val)
		break
		case "VGraphFont":
			val = limit(val,-1,99)
			cmd = SelectString(shftKey,"gfSize=","fSize=")+num2str(val)
		break
		case "tlAxis_left":
		case "tlAxis_bottom":
		case "tlAxis_right":
		case "tlAxis_top":
			val = ax_mode && val<0 ? 0 : val 
			cmd = SelectString(ax_mode,"tlOffset(","btLen(")+axis+")="+num2str(val)
		break
		case "alAxis_left":
		case "alAxis_bottom":
		case "alAxis_right":
		case "alAxis_top":																		// modify either the margin or the position
			cmd = SelectString(ax_mode,SelectString(lblMode>2,"lblMargin","lblPos"),"lblLatPos")+"("+axis+")="+num2str(val)
		break
	EndSwitch
	
	if (strlen(cmd))
		Execute/Q/Z "ModifyGraph/W="+gName+" "+cmd
		if (doPrint)
			Print "•ModifyGraph "+cmd
		endif
	endif
	
	if (keepAuto)
		DoUpdate/W=$gName
		if (StringMatch(ctrlName,"*Height"))
			ModifyGraph/W=$gName height = 0
		elseif (StringMatch(ctrlName,"*Width"))
			ModifyGraph/W=$gName width = 0
		endif
	endif
	Gsize_FetchGraphSizeValues()
	return 0
end

//################################################################################################

function Gsize_PresetPopup(STRUCT WMPopupAction &s) : PopupMenuControl
	string gName = GetTopGraph()
	if (s.eventCode != 2)
		return 0
	endif
	
	string name = "", cmd = "", settings = StrVarOrDefault("root:"+ksGraphSizeSetting,"")
	ControlInfo/W=$(s.win) cPrint		;int doPrint = (V_Value && V_flag == 2)
	ControlInfo/W=$(s.win) AnnoRename
	if (v_flag == 5)
		name = S_Value
	endif
	
	StrSwitch(s.ctrlName)
		case "AspectSelect":
			string ratio = s.popStr
			if (!CmpStr(s.popStr,"-"))
				break
			endif
			if (!CmpStr(s.popStr,"Other"))
				Variable ratio_a = 1, ratio_b = 1
				Prompt ratio_a, "Numerator (a): "
				Prompt ratio_b, "Denominator (b): "
				DoPrompt "Enter ratio in the format a:b", ratio_a, ratio_b
				if (V_flag || !ratio_a || !ratio_b)
					break
				endif
				ratio = num2str(ratio_a) + ":" + num2str(ratio_b)
			endif
			ChangeAspectRatio(ratio)
		break
		case "FormatSelect":
			ChangeGraphFormat(s.popStr)
		break
		case "PlotColorSelect":
		case "FrameColorSelect":
			ControlInfo/W=$(s.win) $s.ctrlName
			sprintf cmd, "ModifyGraph/Z/W=%s %s=(%d,%d,%d,%d)", gName, SelectString(StringMatch(s.ctrlName,"Frame*"),"gbRGB","wbRGB"), V_Red, V_Green, V_Blue, V_Alpha
		break
		case "SizeUnit":
			string list = RemoveFromList("VMagnification;VGraphFont;",ControlNameList(s.win,";","V*"))
			variable div  = s.popNum == 1 ? 1 : (s.popNum == 2 ? 72 : 72/2.54)					// in points, inches or cm
			variable step = s.popNum == 1 ? 1 : (s.popNum == 2 ? 0.014 : 0.036)
			ModifyControlList list ,win=$(s.win) ,limits={0,8000/div,step*2}					// official max size is 8000 points
			string/G root:$ksGraphSizeSetting = ReplaceNumberByKey("unit",settings,s.popNum)	// save settings into global variable
		break
		case "AnnoFrameSelect":
			string/G root:$ksGraphSizeSetting = ReplaceNumberByKey("frame",settings,s.popNum)
			if (strlen(gName) && strlen(name))
				cmd = StringByKey("TYPE",AnnotationInfo(gName,name))+"/W="+gName+"/C/N="+name+"/F="+num2str(s.popNum-1)
			endif
		break
		case "AnnoBackgrSelect":
			string/G root:$ksGraphSizeSetting = ReplaceNumberByKey("background",settings,s.popNum)
			string lastColor = GetUserData(s.win,s.ctrlName, "lastColor")
			lastColor = SelectString(strlen(lastColor),"(65535,65535,65535)",lastColor)			// opaque selection always defaults to white
			if (strlen(gName) && strlen(name))
				cmd = StringByKey("TYPE",AnnotationInfo(gName,name))+"/W="+gName+"/C/N="+name+"/B=" + SelectString(!(s.popNum-1), num2str(s.popNum-1), lastColor)
			endif
		break
		case "AnnoInfo1Select":
		case "AnnoInfo2Select":
			if (!strlen(gName) || !strlen(name))
				break
			endif
			ControlInfo/W=$(s.win) AnnoInfo1Select;		int info1 = V_Value-2
			ControlInfo/W=$(s.win) AnnoInfo2Select;		int info2 = V_Value-2
			variable digit1 = str2num(GetUserData(s.win,"AnnoInfo1Select", "digits"))			// live info: significant digits
			variable digit2 = str2num(GetUserData(s.win,"AnnoInfo2Select", "digits"))			// live info: significant digits
			info2 = info1<0 ? info1 : info2														// switch off both if the first one is disabled
			digit1 = numtype(digit1) || info1<0 ? 5 : digit1
			digit2 = numtype(digit2) || info2<0 ? 5 : digit2
			string info = SelectString(info1<0,buildLiveInfoFormat(info1,info2,digit1,digit2),"remove")
			cmd = "QuickLegendStyle(graphName=\""+gName+"\", annoName=\""+name+"\", liveInfo=\""+info+"\")"
		break
		case "AnnotationSelect":
			if (!CmpStr(s.popStr,"no annotation"))
				break
			endif
			SetVariable AnnoRename 		,win=$(s.win) ,value=_STR:s.popStr ,userData(lastName)=s.popStr
			PopupMenu AnnoBackgrSelect	,win=$(s.win) ,mode=1 ,userData(lastColor)=""				// reset color information
		break
	EndSwitch
	
	if (strlen(cmd))
		Execute/Q/Z cmd
		if (doPrint)
			Print "•"+ReplaceString("/W="+gName,cmd,"")
		endif
	endif
	Gsize_FetchGraphSizeValues()
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static StrConstant kAspectSelectList = "Free;1:1;-;4:3;3:2;16:9;-;2:3;3:4;4:5;Other;"

static function ChangeAspectRatio(string which)
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif

	ControlInfo/W=GraphSizePanel AspectMode;int doWidth = (V_Value && V_flag == 2)
	ControlInfo/W=GraphSizePanel SwapAxes;	int isSwap  = (doWidth ? !V_Value : V_Value)
	ControlInfo/W=GraphSizePanel cPrint;	int doPrint = (V_Value && V_flag == 2)
	string a_val = StringFromList(0,which,":")
	string b_val = StringFromList(1,which,":")
	string ratio = SelectString(doWidth, b_val+"/"+a_val, a_val+"/"+b_val)
	if (!strlen(ratio))
		return 0
	endif
	
	string inactiveAxis = SelectString(isSwap,"width","height")
	string activeAxis   = SelectString(isSwap,"height","width")
	string checkSetAxis = SelectString(doWidth,"width","height")
	if (StringMatch(WinRecreation(gName,0), "*"+checkSetAxis+"={Aspect*"))						// make sure to not apply aspect settings in both axes => swapping has no influence on the direction, it is always width
		Execute/Q "ModifyGraph/Z/W="+gName+" "+inactiveAxis+"=0"
		if (doPrint)
			Print "•ModifyGraph "+inactiveAxis+"=0"
		endif
	endif
	
	ratio = "{Aspect,"+ratio+"}"
	if (!CmpStr(which,"Free"))
		ratio = "0"
	endif
	
	Execute/Q "ModifyGraph/Z/W="+gName+" "+activeAxis+"="+ratio+";Gsize_FetchGraphSizeValues();"
	if (doPrint)
		Print "•ModifyGraph "+activeAxis+"="+ratio
	endif
	return 0
end

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

static StrConstant kFormatSelectList = "Reset Format;Paper Half;Paper Full;PPT 16:9 Half;PPT 16:9 Full;PPT 4:3 Half;PPT 4:3 Full;"

static function ChangeGraphFormat(string which)
	string gName = GetTopGraph()
	if (!strlen(gName))
		return 0
	endif
	
	ModifyGraph/W=$gName/Z width=0, height=0, margin=0
	ModifyGraph/W=$gName/Z gfSize=0, fSize=0
	ModifyGraph/W=$gName/Z expand=0, lsize=1
	if (StringMatch(which,"Reset Format") || StringMatch(which,"Free Size"))
		return 0
	endif
	
	int i, isAxis_T=0, isAxis_B=0, isAxis_L=0, isAxis_R=0
	string list = AxisList(gName), axisData, axisType
	for (i = 0; i < ItemsInList(list); i++)														// check which axes are active (increases the margin here)
		axisData = StringByKey("noLabel(x)",AxisInfo(gName,StringFromList(i,list)),"=")
		axisType = StringByKey("AXTYPE",AxisInfo(gName,StringFromList(i,list)))
		int lblMode = str2num(axisData); lblMode = lblMode<0 ? 0 : lblMode
		isAxis_T = !CmpStr(axisType,"top")		? 2-lblMode : isAxis_T
		isAxis_B = !CmpStr(axisType,"bottom")	? 2-lblMode : isAxis_B
		isAxis_L = !CmpStr(axisType,"left")		? 2-lblMode : isAxis_L
		isAxis_R = !CmpStr(axisType,"right")	? 2-lblMode : isAxis_R
	endfor
	
	variable marginT=0, marginB=0, marginL=0, marginR=0, hVal = 200, wVal = 200
	if (StringMatch(which, "Paper*")) 
		ModifyGraph/W=$gName/Z gfSize=8, expand=1.25, lsize=1.5
		marginL = 11 + 9*isAxis_L + 3*(isAxis_B>0 || isAxis_T>0 && !isAxis_L)
		marginR = 11 + 9*isAxis_R + 3*(isAxis_B>0 || isAxis_T>0 && !isAxis_R)
		marginT = 11 + 9*isAxis_T
		marginB = 11 + 9*isAxis_B
		if (StringMatch(which, "*Half"))
			hVal = 216.0; wVal = 244.8
		elseif (StringMatch(which, "*Full"))
			hVal = 230.4; wVal = 496.8
		endif
	endif
	
	if (StringMatch(which, "PPT*")) 
		ModifyGraph/W=$gName/Z gfSize=14, expand=0.75, lsize=1.5
		marginL = 18 + 12*isAxis_L + 4*(isAxis_B>0 || isAxis_T>0 && !isAxis_L)
		marginR = 18 + 12*isAxis_R + 4*(isAxis_B>0 || isAxis_T>0 && !isAxis_R)
		marginT = 18 + 12*isAxis_T
		marginB = 18 + 12*isAxis_B
		if (StringMatch(which, "*16:9*"))
			hVal = 436
			if (StringMatch(which, "*Half"))
				wVal = 425
			elseif (StringMatch(which, "*Full"))
				wVal = 880
			endif
		elseif (StringMatch(which, "*4:3*"))
			hVal = 440
			if (StringMatch(which, "*Half"))
				wVal = 320
			elseif (StringMatch(which, "*Full"))
				wVal = 648
			endif
		endif
	endif
	
	ModifyGraph/W=$gName/Z margin(left) = marginL, margin(right) = marginR
	ModifyGraph/W=$gName/Z margin(top)  = marginT, margin(bottom)= marginB	
	ModifyGraph/W=$gName/Z height = (hVal - round(marginT + marginB))
	ModifyGraph/W=$gName/Z width  = (wVal - round(marginL + marginR))
	
	DoUpdate
	ControlInfo/W=GraphSizePanel KeepAutoMode
	if (V_Value)
		ModifyGraph/W=$gName/Z height = 0
		ModifyGraph/W=$gName/Z width = 0
	endif
	return 0
end

//################################################################################################

function Gsize_FetchGraphSizeValues()
	string gName = GetTopGraph()
	string pName = "GraphSizePanel"
	if (WinType(pName) != 7)
		return 0
	endif
	if (!strlen(gName))
		DoWindow /T $pName, kGraphSizePanelTitle+": No Graph"
		return 0
	endif
	
	variable version, i
	version = str2num(GetUserData(pName,"","procVersion"))
	version = numtype(version) ? 0 : version
	
	DoUpdate/W=$gName
	GetWindow $gName gsize																		// values from GetWindow (sometimes off by a fraction?)
	//print V_left, V_right, V_top, V_bottom		// debug
	variable graphL = V_left, graphR = V_right
	variable graphT = V_top, graphB = V_bottom
	variable graphW = (V_right-V_left)	
	variable graphH = (V_bottom-V_top)
	
	GetWindow $gName psize
	//print V_left, V_right, V_top, V_bottom		// debug
	variable marginL = (V_left-graphL), marginR = (graphR-V_right)
	variable marginT = (V_top-graphT),  marginB = (graphB-V_bottom)
	variable plotW = (V_right - V_left)
	variable plotH = (V_bottom - V_top)
	
	// +++++++++ if available, replace with values from recreation macro (these are exact)++++++++
	string info = WinRecreation(gName,1), childStyle, childList = ChildWindowList(gName)		// need to clean the style macro of any subwindow styles (won't work anyway)
	for (i = 0; i < ItemsInList(childList); i++)
		childStyle = WinRecreation(gName+"#"+StringFromList(i,childList),1)
		childStyle = RemoveListItem(0,childStyle, "\r")											// remove "proc" line
		childStyle = RemoveListItem(0,childStyle, "\r")											// remove "pauseupdate" line
		childStyle = RemoveListItem(ItemsInList(childStyle,"\r")-1,childStyle, "\r")			// remove "endmacro" line
		info = ReplaceString(childStyle,info,"")												// remove child window style
	endfor
	info = GrepList(info,"^\t(ModifyGraph)", 0 , "\r")
	info = GrepList(info,"(?i)(height|width|margin|gfsize|swapXY|RGB)", 0 , "\r")				// extract only relevant keys
	info = ReplaceString("\tModifyGraph/Z ",info,"")
	info = ReplaceString("\tModifyGraph ",info,"")
	info = ReplaceString("\r",info,",")
	
	GetWindow $gName expand;	variable gMagn = !V_Value ? 1 : V_Value							// magnification parameter
	variable gSwap = str2num(StringByKey("swapXY", info, "=", ","));	gSwap = numtype(gSwap) ? 0 : gSwap				// swap axes parameter
	variable WR_marginL	= gSwap ? str2num(StringByKey("margin(bottom)",	info,"=",",")) : str2num(StringByKey("margin(left)",	info,"=",","))
	variable WR_marginT	= gSwap ? str2num(StringByKey("margin(right)",	info,"=",",")) : str2num(StringByKey("margin(top)",		info,"=",","))
	variable WR_marginR	= gSwap ? str2num(StringByKey("margin(top)",	info,"=",",")) : str2num(StringByKey("margin(right)",	info,"=",","))
	variable WR_marginB	= gSwap ? str2num(StringByKey("margin(left)",	info,"=",",")) : str2num(StringByKey("margin(bottom)",	info,"=",","))
	variable WR_plotW	= gSwap ? str2num(StringByKey("height",			info,"=",",")) : str2num(StringByKey("width",			info,"=",","))
	variable WR_plotH	= gSwap ? str2num(StringByKey("width",			info,"=",",")) : str2num(StringByKey("height",			info,"=",","))
	variable WR_graphH	= WR_plotH + WR_marginB + WR_marginT
	variable WR_graphW	= WR_plotW + WR_marginR + WR_marginL
	variable WR_gFont	= str2num(StringByKey("gfSize", info, "=", ",")), gFont = numtype(WR_gFont) ? 10 : WR_gFont
	
	ControlInfo/W=$pName FrameColorSelect															// new in ver. 2.20
	if (abs(V_flag) == 3)
		variable R, G, B
		string sR, sG, sB, sA
		
		R = 65535; G = 65535; B = 65535
		SplitString/E=("wbRGB=\(([0-9]{1,5}),([0-9]{1,5}),([0-9]{1,5}),?([0-9]{1,5})?\)") info, sR, sG, sB, sA
		if (strlen(sR) && strlen(sG) && strlen(sB))
			R = limit(str2num(sR),0,65535)
			G = limit(str2num(sG),0,65535)
			B = limit(str2num(sB),0,65535)
		endif
		if (strlen(sA))
			PopupMenu FrameColorSelect ,win=$pName ,popColor=(R,G,B,limit(str2num(sA),0,65535))
		else
			PopupMenu FrameColorSelect ,win=$pName ,popColor=(R,G,B)
		endif
		
		R = 65535; G = 65535; B = 65535
		SplitString/E=("gbRGB=\(([0-9]{1,5}),([0-9]{1,5}),([0-9]{1,5}),?([0-9]{1,5})?\)") info, sR, sG, sB, sA
		if (strlen(sR) && strlen(sG) && strlen(sB))
			R = limit(str2num(sR),0,65535)
			G = limit(str2num(sG),0,65535)
			B = limit(str2num(sB),0,65535)
		endif
		if (strlen(sA))
			PopupMenu PlotColorSelect ,win=$pName ,popColor=(R,G,B,limit(str2num(sA),0,65535))
		else
			PopupMenu PlotColorSelect ,win=$pName ,popColor=(R,G,B)
		endif
	endif
	
	
	// ++++++++++++++++++++++++++++++++ get aspect ratio info ++++++++++++++++++++++++++++++++++++
	string ratio_h = "", ratio_w = "", list = ""
	SplitString/E=("height={Aspect,([^}]*)") info, ratio_h
	SplitString/E=("width={Aspect,([^}]*)")  info, ratio_w
	variable ratio = getRatioMode(ratio_h, 0)
	if (numtype(ratio))
		ratio = getRatioMode(ratio_w, 1)
	endif
	if (numtype(ratio) && (strlen(ratio_h) || strlen(ratio_w)) && version >= 2.20)
		ratio = WhichListItem("Other", kAspectSelectList)+1										// must be a custom ratio then
	endif
	ratio = ratio < 1 || numtype(ratio) ? 1 : ratio												// backwards compatibility
	
	// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	marginL	= numtype(WR_marginL)==0 ?	WR_marginL	: marginL									// if the recreation values are valid then use these
	marginB	= numtype(WR_marginB)==0 ?	WR_marginB	: marginB
	marginR	= numtype(WR_marginR)==0 ?	WR_marginR	: marginR
	marginT	= numtype(WR_marginT)==0 ?	WR_marginT	: marginT
	plotW	= numtype(WR_plotW)==0 ?	WR_plotW	: plotW
	plotH	= numtype(WR_plotH)==0 ?	WR_plotH	: plotH
	graphH	= numtype(WR_graphH)==0 ?	WR_graphH	: plotH + marginT + marginB
	graphW	= numtype(WR_graphW)==0 ?	WR_graphW	: plotW + marginL + marginR
	
	ControlInfo/W=$pName SizeUnit;
	variable modifier = V_Value==1 ? 1 : (V_Value == 2 ? 72 : 72/2.54)							// in points, inches or cm
	if (modifier != 1)
		marginL	= round((marginL/modifier)*1000)/1000
		marginB	= round((marginB/modifier)*1000)/1000
		marginR	= round((marginR/modifier)*1000)/1000
		marginT	= round((marginT/modifier)*1000)/1000
		graphH	= round((graphH /modifier)*1000)/1000
		graphW	= round((graphW /modifier)*1000)/1000
		plotH	= round((plotH  /modifier)*1000)/1000
		plotW	= round((plotW  /modifier)*1000)/1000
	endif
	// +++++++++++++++++++++++++++++++ apply info to panel controls ++++++++++++++++++++++++++++++
	string undoGName = GetUserData(pName, "", "LastgName")										// check for existence of a undo style for the current graph
	DoWindow /T $pName, kGraphSizePanelTitle+": " + gName
	CheckBox SwapAxes			,win=$pName	,value=gSwap
	SetVariable VLeftMargin		,win=$pName	,value=_NUM:marginL	,fstyle=(numtype(WR_marginL)==0)
	SetVariable VBottomMargin	,win=$pName	,value=_NUM:marginB	,fstyle=(numtype(WR_marginB)==0)
	SetVariable VRightMargin	,win=$pName	,value=_NUM:marginR	,fstyle=(numtype(WR_marginR)==0)
	SetVariable VTopMargin		,win=$pName	,value=_NUM:marginT	,fstyle=(numtype(WR_marginT)==0)
	SetVariable VGraphHeight	,win=$pName	,value=_NUM:graphH	,fstyle=(numtype(WR_graphH)==0)
	SetVariable VGraphWidth		,win=$pName	,value=_NUM:graphW	,fstyle=(numtype(WR_graphW)==0)
	SetVariable VPlotWidth		,win=$pName	,value=_NUM:plotW	,fstyle=(numtype(WR_plotW)==0)
	SetVariable VPlotHeight		,win=$pName	,value=_NUM:plotH	,fstyle=(numtype(WR_plotH)==0)
	SetVariable VGraphFont		,win=$pName	,value=_NUM:gFont	,fstyle=(numtype(WR_gFont)==0)
	SetVariable VMagnification	,win=$pName	,value=_NUM:gMagn
	PopupMenu AspectSelect		,win=$pName	,mode=ratio
	Button RevertButton			,win=$pName	,disable=2*(!StringMatch(gName,undoGName))			// disable button if undo backup does not match to current graph
	
	// +++++++++++++++++++++++++++++++++ tick and label controls +++++++++++++++++++++++++++++++++
	string axes = RemoveFromList("left;bottom;right;top;",AxisList(gName))
	ControlInfo/W=$pName tickCtrlTabs;	int tab = V_flag==8 ? V_Value : 0						// new in ver. 2.20
	if (V_flag == 8)
		TabControl tickCtrlTabs ,win=$pName ,value = (!ItemsInList(axes) ? 0 : tab) ,disable = 2*(!ItemsInList(axes))
	endif
	toggleAxisControls(gName, pName, tab)
	
	// +++++++++++++++++++++++++++++++++++ annotation controls +++++++++++++++++++++++++++++++++++
	ControlInfo/W=$pName AnnoRename																// new in ver. 2.20
	if (V_flag != 5)
		return 0
	endif
	
	string last_name = GetUserData(pName, "AnnoRename", "lastName")
	string anno_list = AnnotationList(gName)
	string anno_name = StringFromList(ItemsInList(anno_list)-1,anno_list)
	if (WhichListItem(last_name,anno_list) >= 0)
		anno_name = last_name
	endif
	SetVariable AnnoRename ,win=$pName ,value=_STR:anno_name ,disable=2*(!strlen(anno_name)) ,userData(lastName)=anno_name
	
	if (strlen(anno_name))
		string data, mode1, mode2, digit1 = "", digit2 = ""
		info = AnnotationInfo(gName, anno_name)
		list = ControlNameList(pName,";","cAnchor_*")
		
		data = StringByKey("A", StringByKey("FLAGS",info), "=", "/")
		for (i = 0; i < ItemsInList(list); i++)
			CheckBox $StringFromList(i,list) ,win=$pName ,value=!CmpStr(StringFromList(i,list),"cAnchor_"+data)
		endfor
		
		data = StringByKey("F", StringByKey("FLAGS",info), "=", "/")
		PopupMenu AnnoFrameSelect	,win=$pName ,mode=str2num(data)+1
		
		data = StringByKey("B", StringByKey("FLAGS",info), "=", "/")
		if (StringMatch(data,"(*"))																// opaque mode has a color set instead /B=(XX,XX,XX)
			PopupMenu AnnoBackgrSelect	,win=$pName ,mode=1 ,userData(lastColor)=data
		else
			PopupMenu AnnoBackgrSelect	,win=$pName ,mode=str2num(data)+1
		endif
		
		data = StringByKey("TYPE",info)
		CheckBox cAnno_DoTextbox	,win=$pName ,value=!CmpStr(data,"Textbox")
		
		info = StringByKey("TEXT",info)															// work only with the text from here on
		SplitString/E=("\\\\f([0-9]{2})") info, data
		modifier = str2num(data); modifier = numtype(modifier) ? 0 : modifier
		CheckBox cAnno_TextBold		,win=$pName ,value=((modifier & 2^0)!=0)
		CheckBox cAnno_TextItalic	,win=$pName ,value=((modifier & 2^1)!=0)
		
		SplitString/E=("\\\\Zr([0-9]{3})") info, data
		modifier = str2num(data); modifier = numtype(modifier) ? 100 : modifier
		SetVariable AnnoFontZoom	,win=$pName	,value=_NUM:modifier
			
		variable m1_val = 0, m2_val = 0, cols = 1, color = 0
		wave/T temp = ListToTextWave(GrepList(ReplaceString("\\r",info,";"),"\\\\s"),";")		// find the rows with a trace
		if (DimSize(temp,0) > 0)
			SplitString/E=("(?:\\\\{GT_live.*?\((.*?)\)})(?:.*?\\\\{GT_live.*?\((.*?)\)})?") StringFromList(1,temp[0],"\\s"), mode1, mode2
			SplitString/E=("\",\s?([0-9]),?\s?([0-9])?") mode1, mode1, digit1					// extract numbers inside the GT_liveInfo brackets
			SplitString/E=("\",\s?([0-9]),?\s?([0-9])?") mode2, mode2, digit2
			SplitString/E=("\\\\K\(([0-9]{1,5},[0-9]{1,5},[0-9]{1,5})\)") temp[0], data
			color = (strlen(data)>0)
			m1_val = str2num(mode1);  m1_val = numtype(m1_val) ? 0 : m1_val+1
			m2_val = str2num(mode2);  m2_val = numtype(m2_val) ? 0 : m2_val+1
			Make/Free/N=(DimSize(temp,0)) count_wave = ItemsInList(temp[p],"\\s")-1				// count the trace entries per row
			cols = WaveMax(count_wave); cols = cols<1 ? 1 : cols
		endif
		PopupMenu AnnoInfo1Select ,win=$pName ,mode=m1_val+1 ,userData(digits)=SelectString(strlen(digit1),"5",digit1)
		PopupMenu AnnoInfo2Select ,win=$pName ,mode=m2_val+1 ,userData(digits)=SelectString(strlen(digit2),"5",digit2)
		SetVariable AnnoColumns   ,win=$pName ,value=_NUM:cols
		CheckBox cAnno_TraceColor ,win=$pName ,value=(color>0)
	endif
	
	return 0
end

//------------------------------------------------------------------------------------------------

static function toggleAxisControls(string gName, string pName, int tab)
	ControlInfo/W=$pName ToggleAxisMode
	variable i, ax_mode = 0
	if (V_flag==1)
		ax_mode = str2num(GetUserData(pName,"ToggleAxisMode", "axisMode"))						// new in ver. 2.20
		ax_mode = numtype(ax_mode) ? 0 : ax_mode
	endif
	
	string axType, axData, axName, axes = AxisList(gName)
	string norm_axes = "left;bottom;right;top;"
	string axes_used = SelectString(tab,norm_axes, RemoveFromList(norm_axes, axes))				// normal or custom axes
	
	Make/Free/N=(4) axTypes = 0
	for (i = 0; i < ItemsInList(axes); i++)														// first find out if there are free axes any  side
		axData = AxisInfo(gName,StringFromList(i,axes))
		axType = StringByKey("AXTYPE",axData)													// count the number of axes (free axes share the space) and ...
		axTypes[WhichListItem(axType, norm_axes)] += 1+str2num(StringByKey("ISTFREE",axData))	// ... make sure true free axes trigger the selection in any case
	endfor
	
	for (i = 0; i < ItemsInList(norm_axes); i++)
		variable tick_opt = 0, tick_set = 0, lbls_opt = 0, lbls_set = 0, lblMode = 1, isActive = 0, gridIsOn = 0
		string currAxix = StringFromList(i,axes_used), lblType, tckType
		string ctrlName = StringFromList(i,norm_axes)											// standard axes are at the same time ctrl names
		
		axName = SelectString(strlen(currAxix),"None",UpperStr(currAxix[0])+currAxix[1,7])
		axData = AxisInfo(gName,currAxix)
		if (strlen(axData))
			axType  = StringByKey("AXTYPE",axData)
			lblMode = str2num(StringByKey("lblPosMode(x)",axData,"=",";"))
			lblMode = lblMode==0 ? (axTypes[WhichListItem(axType,norm_axes)]>1 ? 3 : 1) : lblMode
			lblType = SelectString(ax_mode,SelectString(lblMode>2,"lblMargin(x)","lblPos(x)"),"lblLatPos(x)")
			tckType = SelectString(ax_mode,"tlOffset(x)","btLen(x)")
			
			tick_opt = str2num(StringByKey(tckType,		axData,"=",";"))
			lbls_opt = str2num(StringByKey(lblType,		axData,"=",";"))
			tick_set = str2num(StringByKey("tick(x)",	axData,"=",";"))
			lbls_set = str2num(StringByKey("noLabel(x)",axData,"=",";"))
			gridIsOn = str2num(StringByKey("grid(x)",	axData,"=",";"))
			isActive = ax_mode ? gridIsOn==2 : !(tick_set==3 && lbls_set==1)
		endif
		

		CheckBox $("cAxis_"+ctrlName) ,title=axName  ,win=$pName ,value=isActive	  ,disable=2*(!strlen(axData)) ,userdata(axisName)=currAxix
		ControlInfo/W=$pName $("tlAxis_"+ctrlName)
		if (V_flag == 5)
			SetVariable $("tlAxis_"+ctrlName)		 ,win=$pName ,value=_NUM:tick_opt ,disable=2*(!strlen(axData))
		endif
		ControlInfo/W=$pName $("alAxis_"+ctrlName)
		if (V_flag == 5)
			SetVariable $("alAxis_"+ctrlName)		 ,win=$pName ,value=_NUM:lbls_opt ,disable=2*(!strlen(axData)) ,userdata(labelMode)=num2str(lblMode)
		endif
	endfor
end

//------------------------------------------------------------------------------------------------

static function getRatioMode(string str, int inv)												// finds the input ratio from the popup list
	variable ratio = str2num(str), num1, num2, i
	if (numtype(ratio) || !strlen(str))
		return NaN
	endif
	for (i = 0; i < ItemsInList(kAspectSelectList); i++)										// compare with aspect presets
		sscanf StringFromList(i,kAspectSelectList), "%f:%f", num1, num2
		num1 = inv ? num1/num2 : num2/num1														// calculate the ratio
		if (!numtype(num1) && abs(ratio-num1)<10^-4)
			return i+1
		endif
	endfor
	return NaN
end

//________________________________________________________________________________________________
//	attaches a quick-scaling menu to the top graph
//	version 4 -  2022/03
//________________________________________________________________________________________________

function AttachQuickScalePanel()
	string win=getTopMainGraph()
	if (strlen(win) == 0)
		return 0
	endif
	GetWindow $win hook(MainWindow)
	if (StringMatch(S_Value, "QuickScale_GraphHook"))
		return 0
	endif
	SetWindow $win hook(MainWindow)=QuickScale_GraphHook, hookevents = 4
	ShowInfo
	
	//++++++++++++++++++++++++++++++ build the panel ++++++++++++++++++++++++++++++++	
	string pPath = win+"#QuickScalePanel"
	NewPanel/HOST=$win/N=QuickScalePanel/EXT=0/W=(0,0,148,525) as "Norm & Scale"
	TitleBox tCursor		,pos={10,10}	,size={90,23}	,title="No Cursor"				,frame=0 	,fstyle=1	,userdata="none"
	Button bAddCursor		,pos={78,8}		,size={60,20}	,title="Add Csr"				,help={"Adds a new cursor to the graph."}
	Button bTracePrev		,pos={ 8,35}	,size={60,20}	,title="Prev Tr"				,help={"Set cursor to previous trace in list."}
	Button bTraceNext		,pos={78,35}	,size={60,20}	,title="Next Tr"				,help={"Set cursor to next trace in list."}
	TitleBox tFindPos		,pos={10,60}	,size={90,23}	,title="Place Cursor at:"		,frame=0
	Button bFindMin			,pos={ 8,80}	,size={60,20}	,title="Min Val"				,help={"Find the minimum data value."}
	Button bFindMax			,pos={78,80}	,size={60,20}	,title="Max Val"				,help={"Find the maximum data value."}
	TitleBox tDelSection	,pos={10,108}	,size={90,23}	,title="Delete:"				,frame=0
	TitleBox tCsrDirection	,pos={50,137}	,size={90,23}	,title="from Csr"				,frame=0
	Button bDeleteLeft		,pos={ 8,135}	,size={30,20}	,title="<"						,help={"Deletes all data points to the left of the cursor."}
	Button bDeleteRight		,pos={108,135}	,size={30,20}	,title=">"						,help={"Deletes all data points to the right of the cursor."}
	Button bImgDeleteTop	,pos={58,110}	,size={30,20}	,title="^"						,help={"Deletes all data points above the cursor."}
	Button bImgDeleteBottom	,pos={58,160}	,size={30,20}	,title="v"						,help={"Deletes all data points below the cursor."}
	DrawLine 8, 187, 138 ,187
	TitleBox tNormalize		,pos={10,195}	,size={90,23}	,title="Normalize at Csr to:"	,frame=0
	SetVariable vNormalize	,pos={ 8,215}	,size={130,23}	,title=""						,help={"Normalizes the spectrum to this value at cursor position."}
	Button bNormBack		,pos={ 8,240}	,size={130,20}	,title="Undo Normalization"		,help={"Undoes the previous normalization."}
	DrawLine 8, 272, 138 ,272
	CheckBox cImageYShift	,pos={92,280}	,size={35,23}	,title="Set Y:"					,fColor=(0,0,65280)
	TitleBox tShift			,pos={10,280}	,size={90,23}	,title="Set X at Csr:"			,frame=0
	SetVariable vShift		,pos={ 8,300}	,size={130,23}	,title=""						,help={"Offsets the selected axis so that the cursor will be at this value."}
	TitleBox tRelShift		,pos={10,322}	,size={90,23}	,title="Relative X Shift:"		,frame=0	
	SetVariable vRelShift	,pos={ 8,342}	,size={130,23}	,title=""						,help={"Offsets the selected axis by this relative value."}
	DrawLine 8, 370, 138 ,370
	TitleBox tPhoton		,pos={10,377}	,size={90,23}	,title="Photon Energy:"			,frame=0
	SetVariable vPhoton		,pos={ 8,397}	,size={130,23}	,title=""						,help={"Set the photon energy."}
	Button bSwitch			,pos={ 8,422}	,size={130,20}	,title="Convert KE <=> BE"		,help={"Switch between kinetic and binding energy using the set photon energy."}
	DrawLine 8, 450, 138 ,450
	TitleBox tArithmetic	,pos={10,457}	,size={90,23}	,title="Arithmetic Operation:"	,frame=0
	SetVariable vArithmetic	,pos={ 8,477}	,size={130,23}	,title=""						,help={"Multiply, divide, add, subtract a value by/to/from the data by this command just like in the command line."}
	CheckBox cImageOneCol	,pos={ 9,500}	,size={130,23}	,title="Current Column"			,help={"Applies arithmetic operation to the selected column only instead of the full image."}	,value=1
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	SetVariable vShift		,value=_NUM:0	,limits={-inf,inf,1}	,format="%g"
	SetVariable vNormalize	,value=_NUM:0	,limits={-inf,inf,1}					,userdata(reset)="0"	,userdata(wavname)="none"
	SetVariable vRelShift	,value=_NUM:0	,limits={-inf,inf,1}	,format="%g"	,userdata(offset)="0"	,userdata(wavname)="none"
	SetVariable vPhoton		,value=_NUM:0	,limits={0,inf,0}		,format="%g"
	SetVariable vArithmetic	,value=_STR:""
	CheckBox cImageYShift	,side=1			,disable=1		,proc=QuickRescale_CheckExecute
	ModifyControlList ControlNameList(pPath,";","b*")		,proc=QuickRescale_ButtonExecute
	ModifyControlList ControlNameList(pPath,";","v*")		,proc=QuickRescale_VariableExecute
	ModifyControlList ControlNameList(pPath,";","!bAddCursor") disable = 2						// disable everything for now (other than the add cursor button)
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pPath,";","*")	,win=$pPath	,fsize=10
	endif
	
	SetWindow $(pPath) hook(SubWindow)=QuickScale_PanelHook										// set an additional hook to handle close events
	SetActiveSubwindow $win
	
	int i
	for (i = 0; i < ItemsInList(kUsedCursorList); i++)											// see if a cursor is set already
		string csr = StringFromList(i,kUsedCursorList)
		if (strlen(CsrInfo($csr,win)))
			string trace = StringByKey("TNAME",CsrInfo($csr,win))
			if (strlen(ImageInfo(win,trace,0)))													// update the cursor (recalculates the control values)
				Cursor/I/P $csr, $trace, pcsr($csr,win), qcsr($csr,win)
			else
				Cursor/P   $csr, $trace, pcsr($csr,win)
			endif
			break
		endif
	endfor
	
	return 0
end

//################################################################################################

function QuickScale_GraphHook(STRUCT WMWinHookStruct &s)
	string pPath = s.WinName+"#QuickScalePanel"
	int i, HookTakeover = 0
	Switch(s.eventCode)
		case 8:		// modified => if trace gets added or removed
		case 7:		// cursor moved
			if (WinType(pPath) != 7)															// something went wrong, since there is no panel => remove hook
				SetWindow $s.WinName hook(MainWindow)=$""
				return 0
			endif
			string csr = s.CursorName, trace = s.TraceName
			if (!strLen(trace))																	// if the trace is not assigned, look for other cursors
				for (i = 0; i < ItemsInList(kUsedCursorList); i++)
					csr = StringFromList(i,kUsedCursorList)
					if (strlen(CsrInfo($csr,s.WinName)))
						trace = StringByKey("TNAME",CsrInfo($csr,s.WinName))
						break
					endif
				endfor
			endif
			
			if (!strLen(trace))																			// still not trace? => abort
				ModifyControlList ControlNameList(pPath,";","!bAddCursor") ,win=$pPath ,disable = 2		// disable everything other than the add cursor button
				TitleBox tCursor		,win=$pPath ,title="No Cursor" ,userdata="none"
				CheckBox cImageYShift	,win=$pPath	,disable=1
				return 0
			endif
			HookTakeover = 1
			
			wave w = CsrWaveRef($csr,s.WinName)
			wave/Z w_x = CsrXWaveRef($csr,s.WinName)
			string wName = NameOfWave(w)
			string imgInfo = ImageInfo(s.WinName,s.TraceName,0)
			int isImg = strlen(imgInfo)>0
			
			ModifyControlList ControlNameList(pPath,";","*")		,win=$pPath ,disable = 0			// re-enable everything
			ModifyControlList ControlNameList(pPath,";","bTrace*")	,win=$pPath ,disable = 2*isImg		// disable trace controls for an image
			ModifyControlList ControlNameList(pPath,";","bImg*")	,win=$pPath ,disable = 2*!isImg		// enable image controls
			
			string hUnit, wOffsets
			ControlInfo/W=$pPath vPhoton
			variable hv = V_Value
			variable xOff = DimOffset(w,0), yOff = DimOffset(w,1), xDelta = DimDelta(w,0), yDelta = DimDelta(w,1)
			if (isImg)
				wave/Z imgXW = $(StringByKey("XWAVEDF",imgInfo)+StringByKey("XWAVE",imgInfo))
				wave/Z imgYW = $(StringByKey("YWAVEDF",imgInfo)+StringByKey("YWAVE",imgInfo))
				if (WaveExists(imgXW))
					xOff   = imgXW[0]
					xDelta = (imgXW[DimSize(imgXW,0)-1] - imgXW[0])/(DimSize(imgXW,0)-1)
				endif
				if (WaveExists(imgYW))
					yOff   = imgYW[0]
					yDelta = (imgYW[DimSize(imgYW,0)-1] - imgYW[0])/(DimSize(imgYW,0)-1)
				endif
			elseif(WaveExists(w_x))
				xOff   = w_x[0]
				xDelta = (w_x[DimSize(w_x,0)-1] - w_x[0])/(DimSize(w_x,0)-1)
			endif
			sprintf wOffsets, "%f;%f;", xOff, yOff
			
			ControlInfo/W=$pPath cImageYShift;	int useY = V_Value && isImg
			variable curPos = useY  ? vcsr($csr, s.WinName) : hcsr($csr, s.WinName)
			variable height = isImg ? zcsr($csr, s.WinName) : vcsr($csr, s.WinName)
			
			string offsetWave = GetUserData(pPath, "vRelShift",  "wavname")
			string renormWave = GetUserData(pPath, "vNormalize", "wavname")
			variable startOff = str2num(StringFromList(useY,GetUserData(pPath, "vRelShift", "offset")))
			
			string wUnit = WaveUnits(w,0); 	wUnit[0] = SelectString(strlen(wUnit),""," ")
			TitleBox tShift			,win=$pPath	,title=SelectString(useY,"Set X at Csr:","Set Y at Csr:")
			TitleBox tRelShift		,win=$pPath	,title=SelectString(useY,"Relative X Shift:","Relative Y Shift:")
			CheckBox cImageYShift 	,win=$pPath	,disable=!isImg
			CheckBox cImageOneCol	,win=$pPath	,disable=!(WaveDims(w)==2)*2
			TitleBox tCursor		,win=$pPath	,title="Cursor "+csr	,userdata=csr									// the cursor position will be saved into the variable control
			SetVariable vShift		,win=$pPath	,value=_NUM:curPos		,limits={-inf,inf,(useY ? yDelta : xDelta)} ,format="%g"+wUnit
			SetVariable vNormalize	,win=$pPath	,value=_NUM:height		,limits={-inf,inf,height/10}
			if(!StringMatch(wName,renormWave))																			// check if the current and the last waves are the same
				SetVariable vNormalize	,win=$pPath	,userdata(reset)="0" ,userdata(wavname)=wName						// if not, switch over to the new wave
			endif
			if(StringMatch(wName,offsetWave))																			// check if the current and the last waves are the same
				SetVariable vRelShift	,win=$pPath	,value=_NUM:(round(((useY ? yOff : xOff)-startOff)*10000)/10000)	// set control to the updated value
			else
				SetVariable vRelShift	,win=$pPath	,value=_NUM:0 ,limits={-inf,inf,(useY ? yDelta : xDelta)} ,userdata(wavname)=wName ,userdata(offset)=wOffsets ,format="%g "+wUnit
			endif
			if (hv == 0 || numtype(hv))
				hv = abs(TraceTool_FetchXShift(w, hUnit));	hv = numtype(hv) ? 0 : hv
				SetVariable vPhoton		,win=$pPath	,value=_NUM:hv ,format="%g "+hUnit
			endif
		break
	EndSwitch
	return HookTakeover
end
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
function QuickScale_PanelHook(STRUCT WMWinHookStruct &s)
	if (s.eventCode == 17)	// kill
		string gName = StringFromList(0, s.WinName, "#")
		HideInfo/W=$gName
		SetWindow $gName hook(MainWindow)=$""													// remove hook from main window
	endif
	return 0
end

//################################################################################################

function QuickRescale_ButtonExecute(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode != 2)
		return 0
	endif
	
	string gName = StringFromList(0, s.win, "#")
	string csr = GetUserData(s.win, "tCursor", "")
	string traceList = "", trace = "", notes = ""
	variable Findx = 0, Findy = 0
	int items = FetchTraces(gName,traceList)
	
	if (CmpStr(csr,"none") != 0 && strlen(csr))
		wave   yw = CsrWaveRef($csr,gName)
		wave/Z xw = CsrXWaveRef($csr,gName)
		trace = StringByKey("TNAME",CsrInfo($csr,gName))
		Findx = pcsr($csr,gName)
		Findy = qcsr($csr,gName)
	endif
	if (!strlen(trace) && CmpStr(s.ctrlName,"bAddCursor") != 0)
		s.ctrlName = ""
	endif
	
	int i, isImage = strlen(ImageInfo(gName,trace,0))
	variable GoUp = 0, xOff = 0, yOff = 0, xMul = 0, yMul = 0
	StrSwitch(s.ctrlName)
		case "bFindMax":
			WaveStats/Q/P yw
			Findx = V_maxRowLoc
			Findy = V_maxColLoc
		break
		case "bFindMin":
			WaveStats/Q/P yw
			Findx = V_minRowLoc
			Findy = V_minColLoc
		break
		case "bImgDeleteTop":
			GoUp = 1
		case "bImgDeleteBottom":
			GetAxis/W=$gName/Q $StringByKey("YAXIS",ImageInfo(gName,trace,0))
			v_min = ScaleToIndex(yw,v_min,1)
			v_max = ScaleToIndex(yw,v_max,1)
			GoUp = v_max < v_min ? !GoUp : GoUp													// if the axis is reversed, then swap button function
			if ((!GoUp && (Findy <= 0)) || (GoUp && (Findy >= DimSize(yw,1)-1)))
				break
			endif
			
			if (GoUp)
				DeletePoints/M=1 Findy+GoUp, inf, yw
			else
				notes = GetUserData(s.win, "vRelShift", "offset")															// get the initial x offset
				sprintf notes, "%s;%f;", StringFromList(0,notes), str2num(StringFromList(1,notes))+DimDelta(yw,1)*Findy		// write new offset
				SetVariable vRelShift ,win=$(s.win) ,userdata(offset)=notes													// set new offset
				
				DeletePoints/M=1 0, Findy, yw
				SetScale/P y, DimOffset(yw,1)+DimDelta(yw,1)*Findy, DimDelta(yw,1), WaveUnits(yw,1), yw
			endif
			Findy = GoUp ? DimSize(yw,1)-1 : 0
		break
		case "bDeleteRight":
			GoUp = 1
		case "bDeleteLeft":
			GetAxis/W=$gName/Q $StringByKey("XAXIS",TraceInfo(gName,trace,0))
			if(WaveExists(xw) && !isImage)
				FindLevel/P/Q xw, v_min;	v_min = !v_flag ? V_LevelX : 0
				FindLevel/P/Q xw, v_max;	v_max = !v_flag ? V_LevelX : DimSize(xw,0)-1
			else
				v_min = ScaleToIndex(yw,v_min,0)
				v_max = ScaleToIndex(yw,v_max,0)
			endif
			GoUp = v_max < v_min ? !GoUp : GoUp													// if the axis is reversed, then swap button function
			if ((!GoUp && (Findx <= 0)) || (GoUp && (Findx >= DimSize(yw,0)-1)))
				break
			endif
			
			if(WaveExists(xw) && !isImage)
				if (GoUp)
					DeletePoints/M=0 Findx+GoUp, inf, yw ,xw
				else
					DeletePoints/M=0 0, Findx, yw ,xw
				endif
			else
				if (GoUp)
					DeletePoints/M=0 Findx+GoUp, inf, yw
				else
					notes = GetUserData(s.win, "vRelShift", "offset")														// get the initial x offset
					sprintf notes, "%f;%s;",str2num(StringFromList(0,notes))+DimDelta(yw,0)*Findx, StringFromList(1,notes)	// write new offset
					SetVariable vRelShift ,win=$(s.win) ,userdata(offset)=notes												// set new offset
					
					DeletePoints/M=0 0, Findx, yw
					SetScale/P x, DimOffset(yw,0)+DimDelta(yw,0)*Findx, DimDelta(yw,0), WaveUnits(yw,0), yw
				endif
			endif
			Findx = GoUp ? DimSize(yw,0)-1 : 0
		break
		case "bSwitch":
			variable start = DimOffset(yw,0)
			if (WaveExists(xw) && !isImage)
				start = xw[0]
			endif
			ControlInfo/W=$(s.win) vPhoton
			variable hv = V_Value, skip = 0			
			if (numtype(hv) == 0 && hv > 0)
				#if (Exists("QuickRescale_UserFunc") == 6)										// uses user code for photon energy handling
					skip = QuickRescale_UserFunc(yw, xw, hv)
				#endif
				if (skip)
					break
				endif
				//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
				notes = GetUserData(s.win, "vRelShift", "offset")								// get the initial x offset
				hv = start < 0 ? hv : -hv														// decide the direction
				if ((start+hv)*start <= 0)
					if (WaveExists(xw) && !isImage)
						xw += hv
					else
						SetScale/P x start+hv, DimDelta(yw,0), WaveUnits(yw,0), yw
					endif
					sprintf notes, "%f;%s;",str2num(StringFromList(0,notes))+hv, StringFromList(1,notes)	// write new offset
					SetVariable vRelShift ,win=$(s.win) ,userdata(offset)=notes					// set new offset
				else
					Print "Photon energy is too small."
				endif
			else
				Print "Photon energy needs to be a positive number."
			endif
		break
		case "bNormBack":
			variable oldMax = str2num(GetUserData(s.win, "vNormalize", "reset")), newMax = Wavemax(yw)
			if (oldMax != 0 && numtype(oldMax) == 0)
				yw *= oldMax/newMax																// normalize the wave
				SetVariable vNormalize ,win=$(s.win) ,userdata(reset)="0"
			endif
		break
		case "bAddCursor":
			for (i = 0; i < ItemsInList(kUsedCursorList); i++)									// find first unused cursor
				csr = StringFromList(i,kUsedCursorList)
				if (!strlen(CsrInfo($csr,gName)))
					break
				endif
			endfor
			
			if (!items)																			// if there are no traces, then this must be an image
				trace = StringFromList(0,ImageNameList(gName, ";"))
				wave yw = ImageNameToWaveRef(gName,trace)
			else
				trace = StringFromList(0,traceList)
				wave yw = TraceNameToWaveRef(gName,trace)
				wave/Z xw = XWaveRefFromTrace(gName,trace)										// xw is only needed for traces
			endif
			
			Findx = DimSize(yw,0)/2																// set a new cursor on the center of the first trace or image
			if (!items)
				Findy = DimSize(yw,1)/2
			else
				GetAxis/W=$gName/Q bottom
				if (!V_flag)
					getAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)					// center of the current axis corrected for scaling and offset of the current trace
					if (WaveExists(xw))
						FindLevel/P/Q xw, (V_min+(V_max-V_min)/2 - xOff)/xMul
						Findx = (!v_flag && v_levelX > 0 && v_levelX < DimSize(xw,0)-1) ? v_levelX : round(DimSize(xw,0)/2)		// make sure to stay within range
					else
						Findx = x2pnt(yw,(V_min+(V_max-V_min)/2 - xOff)/xMul)
					endif
				endif
			endif
		break
		case "bTraceNext":
			GoUp = 1
		case "bTracePrev":
			if (strlen(ImageInfo(gName,trace,0)))
				break
			endif
			getAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)
			Findx = hcsr($csr, gName)*xMul + xOff												// correct for current trace's offsets	
			
			variable curTrace = WhichListItem(trace, traceList)
			if (GoUp)
				if (curTrace < items-1)
					 trace = StringFromList(curTrace+1,traceList)
				endif
			else
				if (curTrace > 0)
					 trace = StringFromList(curTrace-1,traceList)
				endif
			endif
			wave yw = TraceNameToWaveRef(gName, trace)
			wave/Z xw = XWaveRefFromTrace(gName,trace)
			
			getAllTraceOffsets(gName, trace, xOff, yOff, xMul, yMul)							// correct for next trace's offsets
			if (WaveExists(xw))
				FindLevel/P/Q xw,(Findx - xOff)/xMul
				Findx = (!v_flag && v_levelX > 0 && v_levelX < DimSize(xw,0)-1) ? v_levelX : round(DimSize(xw,0)/2)			// make sure to stay within range
			else
				Findx = x2pnt(yw,(Findx - xOff)/xMul)
			endif
		break
	EndSwitch

	if (isImage)
		Cursor/P/I $csr, $trace, Findx, Findy													// move cursor to image position
	else
		Cursor/P   $csr, $trace, Findx															// move cursor to trace position
	endif
	SetActiveSubwindow $gName																	// restore focus to main window
	return 0
end

//################################################################################################

function QuickRescale_VariableExecute(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (!(s.eventCode == 1 || s.eventCode == 2))
		return 0
	endif
	
	string varStr = s.sval
	string gName = StringFromList(0, s.win, "#")
	string csr = GetUserData(s.win, "tCursor", "")
	if (CmpStr(csr,"none") == 0 || !strlen(csr))
		return 0
	endif
	wave   yw = CsrWaveRef($csr, gName)
	wave/Z xw = CsrXWaveRef($csr,gName)
	string trace = StringByKey("TNAME",CsrInfo($csr,gName))
	string imgInfo = ImageInfo(gName,trace,0)
	int isImage = strlen(imgInfo)>0
	
	ControlInfo/W=$(s.win) cImageYShift
	variable useY = V_Value && !V_disable, modVal, iniVal
	if (isImage)																				// the xw is used for shifts
		if (useY)
			wave/Z xw = $(StringByKey("YWAVEDF",imgInfo)+StringByKey("YWAVE",imgInfo))
		else
			wave/Z xw = $(StringByKey("XWAVEDF",imgInfo)+StringByKey("XWAVE",imgInfo))
		endif
	endif
	
	StrSwitch(s.ctrlName)
		case "vNormalize":
			modVal = strlen(ImageInfo(gName,trace,0)) ? zcsr($csr,gName)/s.dval : vcsr($csr,gName)/s.dval	// get the 1D y-value or 2D z-value of the cursor
			iniVal = str2num(StringFromList(useY,GetUserData(s.win, "vNormalize", "reset")))
			if (modVal != 0 && numtype(modVal) == 0)
				if (iniVal == 0)
					SetVariable vNormalize ,win=$(s.win) ,userdata(reset)=num2str(Wavemax(yw))				// the original height will be saved into the variable control
				endif
				yw /= modVal
			endif
		break
		case "vShift":
			modVal = useY ? vcsr($csr,gName) : hcsr($csr,gName)
			if (WaveExists(xw) && !isImage)
				xw -= (modVal-s.dval)
				break
			endif
			if (useY)
				SetScale/P y, DimOffset(yw,1)-(modVal-s.dval), DimDelta(yw,1), WaveUnits(yw,1), yw			// scale relative to cursor position in y direction
			else
				SetScale/P x, DimOffset(yw,0)-(modVal-s.dval), DimDelta(yw,0), WaveUnits(yw,0), yw			// scale relative to cursor position in x direction
			endif
		break
		case "vRelShift":
			iniVal = str2num(StringFromList(useY,GetUserData(s.win, "vRelShift", "offset")))
			if (WaveExists(xw) && !isImage)
				modVal = s.dval - (xw[0] - iniVal)												// calculate the difference to the previous change
				xw += modVal
				break
			endif
			if (useY)
				SetScale/P y, iniVal+s.dval, DimDelta(yw,1), WaveUnits(yw,1), yw		
			else
				SetScale/P x, iniVal+s.dval, DimDelta(yw,0), WaveUnits(yw,0), yw
			endif
		break
		case "vArithmetic":
			ControlInfo/W=$(s.win) cImageOneCol;	variable singleCol = V_Value && !V_disable
			DFREF saveDFR = GetDataFolderDFR()
			SetDataFolder GetWavesDataFolderDFR(yw)
			
			string MultiDim = "", columnStr
			if (WaveDims(yw)==2 && singleCol)													// working on 2D waves
				variable col = qcsr($csr,gName)
				if (!strlen(ImageInfo(gName,trace,0)))											// waterfall plot
					SplitString/E=("\]\[(.*)\]") StringByKey("YRANGE",TraceInfo(gName,trace,0)), columnStr		// find the displayed column from TraceInfo
					col = str2num(columnStr)
				endif
				if (numtype(col) == 0)
					MultiDim = "[]["+num2str(col)+"]"
				endif
			endif
			
			string Op = varStr[0]
			if ((CmpStr(Op, "+") == 0) || (CmpStr(Op, "-") == 0) || (CmpStr(Op, "*") == 0) || (CmpStr(Op, "/") == 0))
				string command = PossiblyQuoteName(NameOfWave(yw)) + MultiDim + Op + "=" + varStr[1,inf]			// construct a multiply command string
				Execute/Q/Z command
				if (V_flag)
					Print "There was an error with this command: " + GetErrMessage(V_flag,2)
				endif
			else
				Print "You need to start your expression by a arithmetic operator +,-,* or /."
			endif
			SetDataFolder saveDFR
			break
	EndSwitch
	
	if (isImage)																				// update the cursor (recalculates the control values)
		Cursor/I/P $csr, $trace, pcsr($csr,gName), qcsr($csr,gName)								// move cursor to image position
	else
		Cursor/P $csr, $trace, pcsr($csr,gName)													// move cursor to trace position
	endif
	SetActiveSubwindow $gName																	// restore focus to main window
	return 0
end

//################################################################################################

function QuickRescale_CheckExecute(STRUCT WMCheckboxAction &s) : CheckBoxControl
	if(s.eventCode == 2)
		string gName = StringFromList(0, s.win, "#")
		string csr = GetUserData(s.win, "tCursor", "")											// extract from sub window
		if (CmpStr(csr,"none") != 0 && strlen(csr))
			string image = CsrWave($csr, gName)													// wave reference from cursor
			if (strlen(ImageInfo(gName,image,0)))												// update the cursor (recalculates the control values)
				Cursor/I/P $csr, $image, pcsr($csr, gName), qcsr($csr, gName)
			endif
		endif
	endif
	return 0
end